/* cseed.c
   Interface routines for using the Seed session editor from the C language.
*/

#include "seed.h"

#include <unistd.h>
#include <stdio.h>
#include <string.h>


void program()
/* Pretends to be the Pascal main program. */
{
  printf("program() called.\n");
  exit(3);
}


int seed_stdout ()
/* Make standard output go to Seed.
   Return zero on success, -1 on failure.
   Note, the amount of output that can be handled between two inputs
   may be limited by the pipe capacity (4 kb?).
 */
{
  int p[2];

  if (pipe(p) == -1) {
    perror("seed_stdout, pipe");
    return -1;
  }
  if (dup2(p[1],STDOUT_FILENO) != STDOUT_FILENO) {
    perror("seed_stdout, dup2");
    close(p[0]); close(p[1]);
    return -1;
  }
  if (p[1] != STDOUT_FILENO) close(p[1]);
  if (ed_writefrom(p[0]) == 0) {
    fprintf(stderr,"seed_stdout, ed_writefrom fails.\n");
    close(p[0]);
    return -1;
  }
  /* No buffering should be used. Block buffering is
   * obviously no good, and line buffering is not quite
   * good enough, because Seed input is not thru stdin.
   */
  setbuf(stdout,NULL);
  return 0;
}


lex_wordtype *lex_word_c (char *str)
/* Create a Lex-package "word" string from a C string.
   The result is a Pascal "varying" string.
   Uses a static area, so copy result if you need
   several at a time. */
{
  static lex_wordtype W;
  int len,i;
  len = strlen(str);
  if (len > LEX_MAXWORD) len = LEX_MAXWORD;
  W.length = len;
  for (i=0; i<len; i++) { W.text[i] = *str; str++; }
  return &W;
}

lex_wordtype *LW (char *str)
  /* Obsolete name for lex_word_c. */
  { return lex_word_c (str); }

vstring *vstring_c (char *str)
/* Create a Pascal "varying" string from a C string.
   Uses a static area, so copy result if you need
   several at a time. */
{
  static vstring V;
  int len,i;
  len = strlen(str);
  if (len > MAXSTRING) len = MAXSTRING;
  V.length = len;
  for (i=0; i<len; i++) { V.text[i] = *str; str++; }
  return &V;
}

vstring *VS (char *str)
  /* Obsolete name for vstring_c. */
  { return vstring_c (str); }


void seed_init (int argc, char *argv[])
  {
  __PC0__PCSTART(argc, argv);	/* See Sun Pascal 3.0 Reference Manual,
				   Section 7.4, "C calls Pascal". Note
				   that the inclusion of parameters in
				   this call is empirically verified
				   to work but is not documented. */
  
  ed_init  ();
  }

void seed_fini (int status)
  {
  ed_fini ();
  
  __PC0__PCEXIT(status);	 /* See Sun Pascal 3.0 as above. */

  }
