/* Test C access to the Seed lexline module. */

#include <stdio.h>
#include "cseed.h"
#include "seed_lexline.h"

vstring Word_to_Vstring (lex_wordtype W)
/* Converts a Lex word to a variable-length string. */
{
  vstring V;
  int i,len;
  
  len = W.length;
  if (len > MAXSTRING) len = MAXSTRING;
  V.length = len;
  for (i = 0; i < len; i++)
     V.text[i] = W.text[i];
  return V;
}

typedef enum {EndSymbol,QuitSymbol} SymbolType;

void Tokens()

{
  lex_positiontype here;
  char markch[4];

  markch[0]='\'';
  markch[2]='\'';
  markch[3]=0;
  
  lex_store(&here);
  do {
    lex_get();			if (lex_cancel()) goto cancel;
    switch (lex_token.kind) {
    	case nulltoken:
    	  ed_write(*vstring_c("null"));
    	  break;
    	case cancelsignal:
    	  ed_write(*vstring_c("cancel"));
    	  break;
    	case natnumtoken:
    	  ed_write(*vstring_c("natnum "));
    	  ed_wrint(lex_token.value.natnumber);
    	  break;
    	case realnumtoken:
    	  ed_write(*vstring_c("realnum "));
    	  ed_wrreal(lex_token.value.realnumber,15,8);
    	  break;
    	case marktoken:
    	  ed_write(*vstring_c("mark "));
    	  markch[1]=lex_token.value.character;
    	  ed_write(*vstring_c(markch));
    	  break;
    	case symboltoken:
    	  ed_write(*vstring_c("symbol "));
    	  ed_wrint(lex_token.value.symnum);
    	  break;
    	case wordtoken:
    	  ed_write(*vstring_c("word, length "));
    	  ed_wrint(lex_token.value.word.length);
    	  ed_write(*vstring_c(" '"));
    	  ed_write(Word_to_Vstring(lex_token.value.word));
    	  ed_write(*vstring_c("'"));
    	  break;
    	case stringtoken:
    	  ed_write(*vstring_c("string, length "));
    	  ed_wrint(lex_token.value.string.quoted.length);
    	  ed_write(*vstring_c(" between "));
    	  markch[1] = lex_token.value.string.opener;
    	  ed_write(*vstring_c(markch));
    	  ed_write(*vstring_c(" and "));
    	  markch[1] = lex_token.value.string.closer;
    	  ed_write(*vstring_c(markch));
    	  ed_write(*vstring_c(" '"));
    	  ed_write(lex_token.value.string.quoted);
    	  ed_write(*vstring_c("'"));
    	  break;
    	case eoltoken:
    	  ed_write(*vstring_c("eol"));
    	  break;
    	case eolcommenttoken:
    	  ed_write(*vstring_c("eolcomment"));
    	  break;
    	case eoftoken:
    	  ed_write(*vstring_c("eof"));
    	  break;
    }
    ed_writeln();
    lex_null();
    Tokens();			if (lex_cancel()) goto cancel;
    lex_exit();
  cancel: if (lex_cancel()) lex_restore(here);
  }
  while (!lex_done());
}

int main (int argc, char *argv[])
{
  seed_init(argc, argv);
  lex_init();
  if (seed_stdout() == -1) exit(1);
	      
  lex_bodychar (natnumtoken,'+',*vstring_c("_"));
  lex_quotepair('(',')');
  lex_quotepair('{','}');
  lex_firstchar(stringtoken,'=',*vstring_c("^({"));
  lex_blankchar('+',*vstring_c("*"));
  lex_bodychar(symboltoken,'-',*vstring_c("xyz"));
  lex_bodychar(symboltoken,'+',*vstring_c("-"));
  lex_firstchar(eolcommenttoken,'=',*vstring_c("#"));
  lex_bodychar (eolcommenttoken,'=',*vstring_c(" "));

  lex_define(*lex_word_c("end-it") ,EndSymbol);
  lex_define(*lex_word_c("quit-program"),QuitSymbol);
  lex_ignore(eoltoken);
    
  printf("lexltest - enter tokens.\n");
  printf("'end-it' and 'quit-program' are defined symbols.\n");

  Tokens();

  seed_fini(0);

  return 9;

}
