/* seed_edit.h
   Constants, types, routines for edit.p
*/

#if !defined(SEED_EDIT)
#define SEED_EDIT

/* line length limit (Ed_Getln routine) */
#define ED_MAXLINE 200

typedef char ed_linetype[ED_MAXLINE];
typedef int  scriptindex;
typedef char seed_bool;

typedef char msgkind;
#define nomsg		0
#define processmsg	1
#define cancelmsg	2

typedef struct {
	msgkind kind;
	union {
	   /* void */    /* nomsg */
	   struct {
	      scriptindex processfrom;  /* Index of start of text */
	      seed_bool     eoln,eof;
	      char        character;
	   } process;
	   struct {
	      scriptindex cancelto;     /* Index to reverse to */
	   } cancel;
        } msg;
     } ed_msgtype;


typedef char ed_execmode;
#define	noexec	0
#define	manexec	1
#define jobexec	2
#define	incexec	3

/*
  External declarations for routines in edit.p
  Requires: seed_strings.h
*/

extern void ed_init ();
  /* Initialize editor module. */

extern void ed_fini ();
  /* Finalize editor module. */

extern void writeindex (scriptindex index);
  /* Write index on standard output. */

extern void ed_write (vstring txt);
  /* Write output text into the transcript. */

extern void ed_writeln ();
  /* Write an output newline into the transcript. */

extern void ed_wrint (int i);
  /* Write the integer as output into the transcript. */

extern void ed_wrreal (double r, int w, int d);
  /* Write the real number as output into the transcript.
    Uses the same format as write(r:w:d). */

extern seed_bool ed_writefrom (int fd);
  /* Request editor to accept data from the file descriptor 'fd'
    and store it in the transcript as output. The editor will
    do this in the background, while editing proceeds normally.
    However, the position and tagging of the output may not
    be correct, if the user asks the editor to process some
    other part of the script.
    Returns true for success, false for error. At the moment,
    the only error is to try to define more than one file
    descriptor - only one is allowed at present. */

extern void ed_written();
  /* Request editor to get all output data that is now available in the
    file descriptor defined by Ed_WriteFrom, and tag and store it
    in the transcript. An application should call this routine
    after it has written something to the file descriptor, and before
    it gets more input (otherwise, the output would be processed later,
    and tagged with a later input position). */

extern seed_bool ed_writeends ();
  /* Cancels the preceding ed_WriteFrom request. The editor shall
    no longer accept data from the file descriptor there specified.
    Returns true on success, false for error. At the moment, the
    only error occurs if there is no active output file descriptor. */

extern void ed_used (scriptindex pos);
  /* Tell editor the index of the last input symbol interpreted by the
    processor. The processor may have read later input symbols, but
    guarantees that they have no effect on subsequent output, until
    the next call of ed_Used or ed_Tell. */ 

extern void ed_tell (scriptindex last,
                     scriptindex error,
                     scriptindex start,
                     scriptindex errfst,
                     int errlen);
  /* Sends processor state to Editor.
    The index of the last input symbol read is 'last'.
    The index of the input symbol at which an error has been
    detected is 'error' (infinite if no error).
    If the processor is reporting an error, the erroneous
    input symbol is identified by 'errfst' and 'errlen' - the
    former is the index of the first symbol of the token, and
    the latter is the length of the token.
    The index of the input symbol the processor wants to
    read next is in 'start'. Usually, start = last+1. */

extern void ed_get (ed_msgtype *msg);
  /* Get next input from Editor. */

extern void ed_getln (ed_linetype *line,
                      int         *length,
                      ed_msgtype  *msg);
  /* Get inputs from Editor until end of line, end of file, or
    a cancel condition. A line of 'length' (possibly zero) characters
    is returned in 'line', and the terminating condition in 'msg'. */

extern scriptindex ed_infiniteindex ();
  /* Return a position far, far away ... */

extern seed_bool ed_beginmode (ed_execmode newmode,
                               scriptindex newpos);
  /* Start a new, nested execution mode.
    The first script position executed in the new mode is 'newpos'.
    This must be later than the start of the preceding active mode.
    The function returns 'true' if the mode could be started, and
    'false' for errors. */

extern seed_bool ed_endmode();
  /* Terminate the current, nested execution mode.
    Returns 'true', unless the current mode is the original (bottom)
    mode, which should never terminate. */

#endif
