/* Output from p2c, the Pascal-to-C translator */
/* From input file "edit.p" */


/*  A SEED-like session editor for incremental (reversible) script
    execution.

Author: Niklas Holsti
Status:
   91.02.16 Under construction. No scrolling, no re-reading.
   91.02.27 Uses Screen module.
   91.03.10 Uses Windows module (just one transcript buffer, so far).
   91.04.08 Uses Keymap module.
   91.04.14 Multiple transcripts, multiple windows, cut, copy, paste.
   91.04.19 Reject control keys in typing.
            When opening a window, check that two new spots can be defined
            (one for the cursor, one for cleaning).
   91.04.21 Input and output spots separate from cursor spot.
   91.05.06 Joinline separate from DeleteChar.
            Ed_Getln and associated stuff added.
   91.06.17 All Explain routines have file parameter.
            Problems are reported both on screen and in logfile.
            Ed_Getln does not send read-ahead characers to the
            application until the user asks for reprocessing. Otherwise,
            the line-buffered Lex variant (Lexline.p) tries to read the
            whole line and fails if there is no newline.
            Ed_WrReal accepts 'Extended' type parameter.
   91.06.17 Ts_Insert has preRow/Col parameters.
   91.06.21 Ed_WriteLn uses NewlineWs which uses Wd_BreakLine.
   91.06.22 Window routines have line & char counts.
            MakeNewline handles tagged output (does not use InsShow).
            Display of status line made optional.
            CloseOtherWins sets titleRow for sole remaining window.
            Ed_GetLn applies Ts_Skip to move cursor to start of next
            input line.
   91.09.05 readlog added.
            Cancel sets usedPos:=infinite.
            DelShow does not ask Windows to delete zero characters.
            When window cleaner overstrikes a line, it clears the rest
            of the line, but only if the right margin was not overstruck.
            Note, this may have to be refined when "overstrike-mode" is
            implemented.
            Keymap directory name changed to "pSeed".
   91.09.14 Ef_readIn implemented (InsertFile).
   91.09.15 Log_Open included where it might be needed.
            Old where-line is erased when where-line option is turned off.
            If Delete-Char can't find input to delete, it just beeps
            rather than displays a techical Ts_ message.
   91.09.23 Look for keymap file both in current folder and in Sisu:...
        -24 Accepts preRow,preCol from Ts_Newline.
            ProcEnd and ProcItem implemented.
            ExecuteLine avoids moving cursor if the line is already
            executed, or there is an uncorrected error.
   91.09.25 Ed_Tell extended with errfst,errlen to locate error token,
            used in ProcItem.
   91.09.26 ExecPos is not changed when cancel msg is sent. Instead,
            execPos is zeroed when editing phase starts.
   92.04.10 Port to Sun OS.
            Remember if there is a message, so ClearMess updates
            screen only if necessary.
            Window cleaner uses Sc_Wrote.
            Message displays preceded by Sc_Limbo; we do not keep
            track of cursor position for messages.
            Ed_Fini called before halt; uses Sc_Fini.
            Window cleaner applies with-line overstriking only if
            the row-col exists in the text. (This optimization is
            a bit risky.) Ditto, insertion within line.
   92.04.17 Local vars declared 'private' to reduce risk of sym conflicts.
            Window title always in overstrike mode; cursor in limbo after.
            Begin implementation of "manual" and "job" execution modes.
   92.04.19 Ed_WriteFrom and Ed_WriteEnds added, for handling output from
            subprocesses.
   92.11.29 Keymap file default name changed (really must use $HOME etc...)
   92.12.01 Added 'processfrom' field to messages from editor to application,
            and renamed 'index' field of cancel messages to 'cancelto'.
            Set the fields in 'Process' and Ed_Getln.
            UnWritePastLast corrected (seek to pos+2).
   92.12.02 Refresh-screen (Ef_refresh) added.
            When writing from a file descriptor into the transcript, if there is
            an error in reading it, or if there is no data when there should be,
            the write-from-fd function is turned off.
   92.12.09 Errors in write-from-fd were still not handled correctly. Tried
            to fix, but not a good solution.
   92.12.22 Messages do not beep, unless they are errors.
   93.03.29 Render module used, rather than direct write to output.
            Some errors and warnings are still written to errout, rather
            than the render stream -- most should be rendered, instead.
   93.03.31 Writes pending data from write-channel (writefd) when a
            reversal is reported in Ed_Tell, before the write-point is
            moved (output unwritten). For this, added procedures
            DataAvailable and WriteAndCheck.
            Pending write-channel data is also written at the start
            of an Ed_Getln call, so that line-interleaved output
            gets inserted correctly when a multi-line region is
            executed in one go. (This is a hack, and needs redesign.)
   93.04.01 Ed_written defined and exported, since it seems that the
            applications at times need to make sure that write-channel
            data are processed. Otherwise, they output can be tagged with
            too late script positions, and thus deleted too early etc.
   93.05.14 CutRegion and Paste marked as "changed" the position just before
            the affected region - added 1 to pos(cursor) to fix it.
            When incremental mode execution ends, the location of the
            most recent edit is forgotten, to stop cancellation (when
            the incremental mode ends because some text before its start
            is changed). Also, writes will again be performed (skipwrite
            is reset).
            Overlapping execution mode test is less stringent: a new
            mode can now start at the same position as the current mode,
            but not before.
   93.07.29 Ed_WriteFile added.
            Ed_CatchOutput and Ed_ShowOutput added.
            Local procedure Pause renamed to MessPause (conflict with "pause"
            in unistd).
   94.03.20 Port to Solaris 2.2.
   94.05.05 Ed_MsgType contains a string of data instead of one char.
            Ts_Reads used.
            Overlong lines are reported in a different way: the whole
            line is fetched from Ts, and the error message appears
            after the line, rather than embedded.
            Accept char codes >= 128.
            NameFile, GoToTop and GoToBottom editor functions added.
            MaxTipName and MaxFileName added. AskParamString takes
            max string length parameter.
   94.05.20 Key-map file from /usr/local/src/seed/lib/keymap.t.
   94.07.25 Editor becomes re-entrant: "query transcript" for file names etc.
            is another editable transcript (vide emacs minibuffers ?).
            Some reordering of procedures to account for this.
   94.07.27 Add Ef_abort, Ef_complete, Ef_query.
            While in query mode, prevent editing and execution of other
            transcripts (except NoExec transcripts).
            Key map file is named ".seed-keymap" and is sought in
            working directory, $HOME, and /usr/local/src/seed/lib, in that
            order.
   94.07.28 Clean current window immediately upon entering edit mode, to
            display recent output without delay.
            Tracing and logging functions enabled only if "-debug" is
            specified on command line.
   94.08.02 Begin to move parts to a new "Views" module, to contain data and
            routines for the editor context (windows, transcripts, views).
            Diving line: all idea of "current" window remains here, and all
            user interactions (except for debug) remain here.
   94.08.04 In unprocessable transcripts, any process command works as a
            newline.
   94.09.28 Modifiable/Modified status, "view file" function. Start work.
   94.10.07 Unmodifiable transcript rejects "mutator" edit functions.
            Instead of "allowed editing function" sets, use "function
            maps" that map unallowed functions to Ef_undefined. This lets
            us map process-functions to something useful even in NoExec
            documents.
            Declares "original" (unmodified) status for new transcripts etc.
   94.10.23 On exit, check for unsaved transcripts. Either save them or
            ask user for confirmation to throw changes away, depending on
            the kind of exit command given.
            On a partial save (eg. write input only) always asks for a new
            file name. On a complete save, asks for a new file name if none has
            been defined for the transcript yet.

   TO DO:
            - check file names for proper syntax, existing files etc.
*/


#include <p2c/p2c.h>


#define EDIT_G
#include "edit.h"
/* p2c: strings_f.p, line 5: 
 * Warning: Expected IMPLEMENT, found 'external' [227] */
/* p2c: strings_f.p, line 5: 
 * Warning: Expected END, found 'external' [227] */


void _Edit_init()
{

  /* log_f.p
     External declaration for routines in log.p
     Also logfile var.
  */


  /* render_f.p
     External declarations of render.p routines.
  */
  /* Make sure all data intended for screen get there, ie.
     dump buffer. */

  /* Writes (buffers) one byte for the screen.
     The strange calling interface is designed to
     match the "terminfo" routine "tputs". */

  /* Output the data as such. No 'rendering'. */

  /* Output the text, rendering normally unprintable characters in
     some visible way -- once this is implemented.
     Up to 'maxcol' screen columns are used; the number actually
     used is returned in actcol. For printable text, actcol = length(txt),
     or maxcol, whichever is smaller. */

  /* Same as Rn_Text, but renders only txt[first .. first+len-1].
     Also avoids parameter copying, since the text is a 'var'. */

  /* Initialize module. */

  /* Finalize module. */
  /* */
  /* screen_d.p
     Constants and types for screen.p
  */

  /* at most this many rows per screen */
  /* at most this many columns per row */


  /* operation succeeded */
  /* no TERM environment variable */
  /* could not get termcap data */
  /* required termcap capability missing */
  /* screen area out of bounds */
  /* cursor position out of bounds */
  /* scroll region out of bounds */
  /* delete-line outside scroll region */
  /* insert-        -"-                */

  /* A range of screen rows. */
  /* */
  /* screen_f.p
     External declarations of screen.p routines.
  */

  /* Write an explanation of the result. */

  /* Define the row range and width of the screen. The usual terminals
     have 24 rows and 80 columns. Also sets scrolling region to the
     given row range. */

  /* Inquire screen area. */

  /* Move cursor (insert/overstrike point) to the given row and
     column. Row and column numbers start from 1. */

  /* Informs us that a number of columns have been written
     at the cursor, which thus has moved along. */

  /* Informs us that the cursor position is not known. */

  /* Clear the whole screen and put cursor at row 1, col 1. */

  /* Clear from cursor position to end of line (row). Cursor position
     is included. */

  /* Whether insert mode can be entered. */

  /* Enter insert mode. Text output will appear at the cursor position,
     and the cursor and the rest of the text on the line will be pushed
     right. Text at the right margin spills over and is lost.
     No effect if Sc_CanInsert returns false. */

  /* Enter overstrike mode. Text output will appear at the cursor position,
     overwriting the old text. The cursor moves right, but no text moves.
     The overwritten text is lost. */

  /* Whether Sc_DeleteChar works. */

  /* Delete n characters starting at the cursor and move remaining text
     on the line n positions left to fill the gap. The cursor remains
     at the same position. Space characters are put in the right margin.
     No effect if Sc_CanDeleteChar returns false. */

  /* Set scrolling region. Line insertion or deletion is allowed only
     in the row range given, and only this region is affected,
     all other lines are unchanged.
     After Sc_SetArea, scrolling region is whole screen. */

  /* Insert a new blank line at the cursor. Lines from the cursor to the
     line before the end of the scrolling region all move down one step.
     The line at the end of the scrolling region is lost.
     The cursor stays on the same row, but may move to column 1. */

  /* Delete line at cursor. Cursor must be on a line within the
     scrolling region. Lines from the next line to the end of the
     scrolling region move up one step. A blank line is inserted at the
     end of the scrolling region.
     The cursor stays on the same row, but may move to column 1. */

  /* Make sure all data intended for screen get there, ie.
     dump buffer. */

  /* Initialize Screen package.
     Must be called once, before any other routines.
   */

  /* Finalize the Screen package.
     Cleans up the screen. */
  /* files_f.p
     External declarations for routines in files.p
  */
  /* Opens the named file for reading.
     Returns true for success, false for failure.
   */

  /* Opens the named file for (re-)writing. The file may or may
     not exists before; if it exists, it is erased, then opened
     for writing.
     Returns true for success, false for failure.
   */

  /* Creates a new file for writing.
     Returns true for success, false for failure.
     It is a failure if a file with the given name already
     exists, but the caller cannot distinguish this failure
     from other failures (e.g., no write permission to
     the directory).
   */
  /* argument_f.p
     External declarations of routines in argument.p
  */

  /* Checks if "opt" is an argument on the command line.
     If an argument is "--", remaining arguments are not
     checked.
   */

  /* Returns number of arguments. */

  /* Resets the argument scanner to start from the first "file name"
     argument. Must always be called before the first Arg_NextFile. */

  /* Gets the next "file name" (ie., not "-...") argument.
     If there are no more, returns valid = false.
     Remember to call Arg_FirstFile to start with, to reset the argument
     scanner.
   */
  /* editfuns.p
     Editor functions enumerated.

     Changes:
       94.05.04  Ef_namefile, Ef_goTop, Ef_goBot added.
       94.07.27  Ef_abort, Ef_complete, Ef_query added.
       94.09.28  Ef_viewtip added.
       94.10.07  Ef_disabled added.
  */

  /*  A thing the editor can do. */
  /* undefined function (bottom)                 */

  /* TYPING MODES */

  /* use text insertion mode                     */
  /* use text overstrike mode                    */

  /* TYPING and BACKSPACING */

  /* enter a text character                      */
  /* insert a line break at the cursor           */
  /* delete character to the left of the cursor  */
  /* del char; if newline, join lines            */
  /* backspace cursor:                           */
  /*   insert    : delchar                       */
  /*   overstrike: left, text space, left        */
  /* join cursor line to preceding line (append) */

  /* NAVIGATION */

  /* cursor left one column (no wrap)            */
  /* cursor right one column (no wrap)           */
  /* cursor backward one pos (wrap to prev line) */
  /* cursor forward one pos (wrap to next line)  */
  /* cursor left to beginning of line            */
  /* cursor right to end of line                 */
  /* cursor up one line (scroll if needed)       */
  /* cursor down one line (scroll if needed)     */
  /* scroll backward one screenful               */
  /* scroll forward  one screenful               */
  /* go to top (beginning) of file               */
  /* go to bottom (end) of file                  */

  /* SEARCH and REPLACE */

  /* search forward for a text string            */
  /* search for a text string and change it      */

  /* CUT, COPY and PASTE */

  /* set mark at cursor position                 */
  /* cut text from mark to cursor into buffer    */
  /* copy text from mark to cursor into buffer   */
  /* paste buffered text inserting before cursor */

  /* WINDOWS and TRANSCRIPTS */

  /* Open new window by splitting current window */
  /* Close all except current window             */
  /* Go to next window below (wrap at bottom)    */
  /* Make a new transcript, perhaps from file    */
  /* View a file in a buffer (not modifiable)    */
  /* Go to next transcript (text buffer)         */
  /* Refresh screen completely                   */

  /* FILES */

  /* Store (write) transcript in file.           */
  /* Store (write) only input in file.           */
  /* Store (write) transcript as pure text file. */
  /* Load  (read ) file, inserting it at cursor. */
  /* Give the transcript a new (file) name.      */

  /* PROCESS */

  /* suspend editing, process text up to cursor  */
  /* suspend editing, process current line       */
  /* suspend editing, process to end of input    */
  /* append word to prev input, process the line */
  /* (this is the "menu selection" function)     */
  /* abort processing, terminate (sub)session    */
  /* ask processor to complete this input        */
  /* ask processor to show possible completions  */

  /* save text, exit editor                      */
  /* quit editor (must confirm if text unsaved)  */

  /* TRACING ETC. */

  /* increment log level (debugging)             */
  /* decrement log level (debugging)             */
  /* list window data on log file                */
  /* list transcript data on log file            */
  /* check transcript spots for true row, col &c */
  /* display status line with spot locations &c  */

  /* function now allowed in this window/doc     */
  /* ambiguous or overdefined function (top)     */
  /* keymap_d.p
     Constants and types for keymap.p
  */


  /* very well */
  /* memory full, cannot extend map */
  /* ambiguous (new key string is prefix of old) */
  /* ambiguous (old key string is prefix of new) */
  /* ambiguous and/or undefined */
  /* Not a prefix key string */
  /* */
  /* keymap_f.p
     External declarations for routines in keymap.p
  */













  /* ** Testing routines ** */

  /* Display editor function name map structure. */

  /* Display keystroke map structure. */

  /* Display buffered keystrokes and partial mapping, if any. */

  /* Display the defined name of an editing function. */
  /* transcripts_d.p
     Constants and types for transcripts.p
     Requires: strings_d.p
     NOTE! This version is quite wasteful in storage, since each
     text element is allocated a full "vstring", which may be some
     250 bytes.
  */

  /* a reasonably large maximum line length */

  /* The following constants are intended to be PRIVATE */
  /* Max length of text of element. */
  /* Max number of spots in a transcript. */

  /* Spot table full */
  /* Memory full */
  /* Transcript element full */
  /* Transcript element edit not allowed */
  /* Not a text element */
  /* No element (nil ref) */
  /* Null element (zero length) */
  /* Newline missing */
  /* Newline present, should not be */
  /* Two spots in wrong order */
  /* Spot should be at EOT */
  /* Element cannot be split */
  /* Not at input item */
  /* End of input (at EOT) */
  /* Valid output, not dewritten */
  /* Cannot insert text before output */
  /* Input file format wrong */
  /* Invalid or unimplemented operation */

  /* How to store the transcript in a file. */
  /* .. input part only */
  /* .. input and output as pure text */
  /* .. input and output, tagged as such. */

  /* User typed it */
  /* Program wrote it */
  /* Bot/Eot element */




  /* The following types are intended to be PRIVATE */
  /* Beginning of transcript */
  /* Text, input or output */
  /* Other data, input or output */
  /* End of transcript */

  /* An element of a transcript. */
  /* form a doubly linked list */
  /* input, output, neither (eot) */
  /* whether element ends a line */
  /* causal input symbol (outputs only) */
  /*varying [MaxElemLength] of char*/
  /*Element*/

  /* An insertion point in a transcript */
  /* the element in which the spot lies */
  /* spot is after elem^.theText[offset] */
  /* transcript row and column of char after spot */
  /* whether col is known */
  /* number of input symbols (characters, newlines,
                          etc.) that lie to the left of the spot */
  /*SpotOb*/

  /* A transcript. */
  /* header element   */
  /* trailer element  */
  /* spot data array        */
  /* these spots are in use */
  /* highest-numbered spot or 0 */
  /* whether content modified since creation or
                        last defined "original" state (changes to spots do
                        not count) */
  /*Transcript*/
  /* */
  /* transcripts_f.p
     External declarations for routines in transcripts.p
     Requires: strings_d.p
  */

  /* Initialize module. */

  /* Finalize module. */

  /* Write an explanation of the result. */

  /* Initialise a transcript to be empty. The result can be
     ok or memory full. */

  /* At least this many new spots can still be defined in the transcript. */

  /* Initialise a spot variable to refer to the start of the transcript,
     before the first element. */

  /* Delete a spot. To be done for each spot initialised, before the
     variable is deleted (e.g. by block exit). */

  /* Returns row and column number of the character following the
     given spot. Both row and col start from 1. */

  /* Returns the number of input symbols that precede the given spot.
     Input symbols are input characters and input newlines. */

  /* Determines order of two spots in a transcript.
     Returns sign(spot1-spot2), i.e. negative if spot1 precedes
     spot2 in the transcript. */

  /* Whether the spot is at the right end of its line. Note, there is
     not necessarily a newline at the end of the last transcript line,
     so Ts_AtEol can yield true although Ts_Eoln is false. */

  /* Whether the spot is just before the end of the transcript. */

  /* The length, in screen columns, of the line on which the spot
     lies. For the last transcript line, works as if there were
     always a newline at its end. */

  /* Write the line holding this spot of the transcript. This may be
     an output line, an input line, or a mixed input/output line.
     The output is limited to the column range fromCol..toCol,
     counting the first character as column 1.
     If the line is short, the last column actually written may be
     less than toCol.The number of columns (characters) written is
     returned in numCol. */

  /* Move from a spot by the specified number of rows (up: -, down: +)
     and columns (left: -, right: +). Motion by rows stops at first line
     and last line. Motion by columns stops at left or right margin.
     Note that the end of the transcript is considered to be on a line
     of its own if and only if the transcript proper ends in a newline.
     The new position is returned in movedSpot, which must have been
     initialised by Ts_NewSpot (movedSpot can be the same variable as
     fromSpot).
     Input and output lines and columns are both counted equally. */

  /* Move a spot to the specified row and column position.
     Motion by rows stops at first line and last line.
     Motion by columns stops at left or right margin.
     The spot must have been initialised.
     Input and output lines and columns are both counted equally. */

  /* Move spot forward over the next input newline. If there is no
     input newline between the spot and eot, leave the spot at eot
     and return Ts_nonewline. */

  /* Move the spot to the end of the transcript, just after the
     last input or output symbol.
     The spot must have been initialised. */

  /* Insert (input) character ch at the spot position, pushing the
     remaining text on the line right. The spot moves one step right, to
     follow the inserted character.
     If the spot is not in an input region, it is moved forward to before
     the next input symbol.
     The parameters preRow,preCol return the row and column the spot
     had before the character was inserted, but after the spot was moved
     forward (these values are useful for screen updating).  */

  /* Insert a newline at the spot. The line holding the spot breaks into
     two pieces, with the spot at the start of the later piece.
     The later piece is moved past output that resulted from the
     earlier piece (perhaps splitting an output line).
     If the spot is not in an input region, it is moved forward to before
     the next input symbol.
     The parameters preRow,preCol return the row and column the spot
     had before the newline was inserted, but after the spot was moved
     forward (these values are useful for screen updating). */

  /* Delete input character preceding the spot. The spot does not have
     to be placed immediately after the input.
     After the character is deleted, the line to the right of the
     spot is moved left to fill the gap, The spot is left in the position
     of the deleted character. No output is deleted (here).
     The parameters preRow,preCol return the row and column the spot
     had before the character was deleted (these values are useful for
     screen updating).
     If the input symbol preceding the spot is a newline, the operation
     is rejected with result = Ts_wasnewline. Ts_Joinline must be used
     to delete newlines. */

  /* Delete input newline preceding the spot. The spot does not have to
     be placed immediately after the newline.
     After the newline is deleted, the input line to the right
     of the spot is joined (appended) to the input line to the left of
     the spot.
     Any embedded output in the right line is deleted (it is not necessary
     to "unwrite" it first). Output following the deleted newline, but due
     to input preceding it, is preserved, and will follow the joined line.
     If there is such saved output, but no input newline to the right of
     the spot,  one is added to the joined line anyway (effectively, the
     newline preceding the spot is moved to the end of the transcript input).
     The spot is left at the position of the join.
     The transcript is thus changed in two, possibly disjoint places.
     The first place is the joined line, which is indicated by the spot.
     The second place is the region of the right-hand line, where
     both input and (embedded) output data were removed.
     The parameters delRow,delCol return the start of the removed text
     region, and preRow,preCol return the place after the region. */

  /* Cuts the part between spot1 and spot2 from fromTranscript.
     All of spot1, spot2, and toSpot must lie within or adjacent to
     input data, and spot1 must be <= spot2.
     All input and output between spot1 and spot2 is cut.
     The data cut from fromTranscript are inserted before toSpot
     in toTranscript. Input remains input and output remains output.
     If the cut is successful, spot1 and spot2 are both placed in
     the space where the text was cut.
     The parameters preRow,preCol return the row and column spot2
     had after it was moved past output to be unwritten, but before
     any text (input or output) was deleted (these values are useful
     for screen updating). */

  /* Copies the part between spot1 and spot2 from fromTranscript,
     both input and output. spot1 must precede spot2.
     fromTranscript is not changed, but spot1 is moved forward to
     coincide with spot2.
     The data copied from fromTranscript are inserted before
     toSpot in toTranscript. This position must be within or adjacent
     to input data. If 'asinput' is true, output data are
     converted to input data when copied. */

  /* Erase all of transcript. Any spots in the transcript are moved
     to the start (= the end) of the transcript. */

  /* Skip output until an input symbol  (or end of
     transcript) is found. The new position is immediately
     before the first input symbol after theSpot (or at end of
     transcript, in which case result is Ts_endinput). */

  /* True iff the spot is embedded within output, i.e. has output
     both to the left and to the right.
     This function is useful for the context-dependent choice of
     what a key (e.g. the Enter key) does. */

  /* Move from a spot to a new position immediately before the input
     symbol number beforePos, counting the first input symbol as 1. Note
     that beforePos is absolute, not relative to fromSpot; fromSpot
     is expected to lie close to the sought position and be a good place
     to start seeking.
     The new position is returned in movedSpot, which must have been
     initialised by Ts_NewSpot (movedSpot can be the same variable as
     fromSpot). After a successful move, Ts_Pos(movedSpot) = beforePos-1.
     Input symbols are input characters and input newlines.
     Output elements are not counted.
     The seek may fail if beforePos is nonpositive or too large.
     If beforePos is nonpositive, the movedSpot is placed at fromSpot and
     result is Ts_invalid. If beforePos was too large, the movedSpot is
     defined at end of transcript and result is Ts_endinput. */

  /* Reads the character to the right of the spot and places the spot
     just after the symbol read. The parameter pos returns the number
     of the input symbol read, counting the first symbol as 1.
     Reading a line separator (where Ts_Eoln is true) returns a blank,
     and the spot is moved to the start of the next (input) line.
     If there is no input between the given spot and the end of the
     transcript, a result of Ts_endinput is given, ch and pos are
     undefined, and the spot is set at end of transcript. */

  /* Reads input symbols to the right of the spot and places the spot
     just after the last symbol read. The symbols are returned in the
     parameter data. The parameter pos returns the number of the first
     input symbol read, counting the first symbol in the transcript as 1.
     The number of symbols read and returned depends on the structure
     of the transcript, but reading stops at the next line separator,
     if not before.
     If the first symbol read is a line separator (where Ts_Eoln is true)
     then a blank character is returned and the spot is moved to the
     start of the next (input) line.
     If there is no input between the given spot and the end of the
     transcript, a result of Ts_endinput is given, data and pos are
     undefined, and the spot is set at end of transcript. */

  /* Returns character immediately after (to the right of) the spot.
     The character is in c, unless it was a line separator in which
     case c is blank and eol is true.
     The source of the character (or line separator) is also returned.
     An error results if the spot is at the end of the transcript. */

  /* Whether the spot is just before the end of its line. */

  /* Whether the spot is just before the end of an input file
     within the transcript. */

  /* Insert output text in the transcript at the given spot.
     The rest of the text to the right of the spot, is moved right
     to make room. The spot follows it.
     The output is marked as "due to" or "caused by" the input symbol
     number duepos.
     Writing a null string has no effect on the transcript or the spot
     position, but may be rejected (result other than Ts_ok) if it would
     be illegal to write a non-null string at the given spot. */

  /* Insert an output-newline in the transcript at the given spot.
     The text to the right of the spot, and the spot itself, are thus
     moved to the start of a new line.
     The output is marked as "due to" or "caused by" the input symbol
     number duepos. */

  /* Moves the spot back to just after the last output resulting from the
     input symbols up to and including symbol number safePos (output with
     safe <= safePos).
     On the way, all output resulting from input symbols > safePos is
     marked as invalid, which means it will be deleted when that part of the
     input is reprocessed. Output after the given spot is not considered.
     If no output with safe <= safePos exists, the spot moves to beginning of
     the transcript. */

  /* Deletes all output that immediately follows the given spot.
     If the given spot is within input text, the nothing is done.
     done. Otherwise, any output that immediately follows the
     spot is deleted, up to the next input symbol or eot.
     The output must have been marked invalid, otherwise the result
     code is returned as Ts_validoutput.
     Returns with spot before the next input symbol or at eot.
     Returns, in preRow and preCol, the row and column of the new
     position of the spot (e.g. the input symbol after the output)
     before the output was deleted (these values are useful for
     screen updating).
   */

  /* Store transcript in a text file, with optional handling of output.
     The file is assumed to be open. It is only written to, and not closed.
     The positions of the spots are not stored in the file.
     The character sequence '~[' marks the start and '~]' the end of
     an output element. If the output element is tagged with a 'safe'
     position, the position is stored as a decimal integer between the
     '~' and the '['.
     If the character '~' occurs in the text, it is stored as '~~'.
     The total number of lines written is also returned. */

  /* Load transcript from text file, inserting its text at the given spot.
     If the spot is not in an input region, it is moved forward to before
     the next input symbol.
     The file is assumed to be open, and is read until end of file (or a
     format error). The file is not closed.
     If the file does not contain tags that separate input from output, all
     the text is considered input. If there are tags in the file, input and
     output are loaded as input and output respectively. However, leading
     output (output preceding the first input) is discarded, and other
     output is marked invalid, so it will be removed when the input
     is processed by the application.
     The parameters preRow,preCol return the row and column the spot
     had before the character was inserted, but after the spot was moved
     forward (these values are useful for screen updating).
     The total number of lines read from the file is returned, as well
     as a result code. */

  /*in*/
  /* Reports whether the content of the transcript has been modified
     after the transcript was created or since the last defined "original"
     state (see Ts_Original). Both input and output changes are counted,
     but not changes in the set of spots or the positions of spots. */

  /*io*/
  /* Declares that the content of the transcript is now in an "original"
     state. Ts_Modified will return false until there are further
     changes in the content of the transcript. */


  /** Testing routines **/

  /* Displays element structure, starting from the given row and
     continuing for the given number of rows.
     Display is on screen, one line per row, starting from the given
     column. */

  /* Displays the positions of the currently defined spots,
     including the spot number, text of element, offset within
     element, and row and column numbers. */

  /* Check that the transcript and the defined spots are consistent.
     Report problems on file f. */

  /* List element structure and defined spots on the output file f. */

  /* For testing memory overflow situations - memory will be filled
     when numberElems elements have been allocated. */
  /* */
  /* windows_d.p
     Constants and types for windows.p
  */

  /* Maximum number of rows a window can cover */
  /* Can scroll vertically as little as one line */
  /* Horizontal scroll at least 20 cols at a time */

  /* Window operation sucessful. */
  /* Window would have no rows. */
  /* Window would extend off screen. */
  /* No rows would be left in lower part. */
  /* Negative space in split. */
  /* Windows would overlap */
  /* Some rows were unused. */

  /* First screen rows used. */
  /* Last    "      "    "   */
  /* First text row visible. */
  /*  = screenBotRow - screenTopRow + 1 */
  /* Width in columns */
  /* Half-height of "band" of rows that
                             scroll together horizontally, zero for
                             scrolling each row independently. */
  /* First screen row that may need refresh. */
  /* Cached data to speed up some things: */
  /* .. screen row we looked at last. */
  /*Window*/


  /* When a window display is being updated, the updating routine (Wd_Clean)
     returns a Wd_ScreenOps data structure that tells the editor what to
     do to the screen. The operations are to be done in the order they
     are listed below: first delete a row; then insert a row; then
     overwrite a whole row; then insert or delete characters at some column
     of some row; and finally overstrike characters from some column to
     the end of that row. Usually, only some of these operations are
     requested by Wd_Clean.
        The screen row numbers are given relative to the screen as it is
     successively changed by these operations, so they can be used
     directly in Sc_DeleteLine and Sc_InsertLine routines. The text row
     and column numbers are given relative to the current state of the
     text, i.e. after all the editing operations that have led to the
     window needing updating, so they can also be used directly to find
     the rows in the text.
        The updating routine may have to be called repeatedly, each time
     returning some actions, until the window is clean. The window
     data structure is kept consistent throughout, assuming that the
     screen operations returned by each call to Wd_Clean are implemented.
     Thus, the updating process can be interrupted before the window is
     clean - for example, if the user hits a key such as Next-Page that
     makes it unnecessary to clean the window now.
   */

  /* Delete this screen row (if > 0) and
                            scroll lower rows up one position. */
  /* Insert a new blank line at this screen
                            row (if > 0), scrolling the row and
                            lower rows down one position. */
  /* Overstrike all text on .. */
  /* ..this screen row (if > 0) */
  /* ..taking text from this text row */
  /* ..and this text column. */
  /* Within-row operations .. */
  /* ..in this screen row (if > 0) */
  /* ..taking text from this text row */
  /* Insert or delete .. */
  /* ..this number of chars (+/- for ins/del) */
  /* ..starting at this screen column */
  /* ..taking text from this column. */
  /* Overstrike to end of line.. */
  /* ..this number of chars (at most) */
  /* ..starting at this screen column */
  /* ..taking text from this column.  */
  /*Wd_ScreenOps*/
  /* windows_f.p
     External declarations for the routines in windows.p
  */

  /* Write a short explanation of the result code. */

  /* Open a window with given parameters. TopRow and LeftCol are
     the topmost text row and leftmost text column to be visible.
     The top and bottom screen rows, the number of columns, and the
     'band' for horizontal scrolling are also specified.
     If the window as specified would extend beyond the screen limits,
     Wd_offscreen is returned and the window is not opened.
     If some of the screen rows requested are already in use by other
     windows, Wd_overlap is returned and the window is not opened. */

  /*io*/
  /* o*/
  /* Move and/or resize window on screen. The new top and bottom screen
     rows are given. All screen rows covered by the
     convex union of the old and new window screen areas must be
     free (or used by this window), otherwise Wd_overlap is returned.
     (Window reconfigurations where this condition is not met must be
     implemented by deleting and recreating windows.)
     The text position of the reshaped window is chosen so that the
     existing screen rows are not moved, if possible. */

  /*io*/
  /* o*/
  /* o*/
  /* Split window wholeWin into an upper part (returned in wholeWin)
     and a lower part (returned in lowerWin). The upper part is
     given upperRows screen rows, followed by spaceRows that are left
     unused, and the lower window is given the remaining screen rows
     covered by the original window. If upperRows + spaceRows is so
     large that nothing is left for the lower window, Wd_norowsleft
     is returned and the window is not split.
     Both windows continue to show their respective parts of the
     text. Nothing is done to the spaceRows. */

  /*i */
  /* o*/
  /* Close (discard) the window. The corresponding screen rows
     are marked unused. (If some of them were already so marked,
     Wd_unused is returned.) The window variable cannot be used
     until it is reinitialized with Wd_Open. */

  /* Close all windows by marking all screen rows as unused.
     No window variable can be used until it is reinitialized
     with Wd_Open. */

  /*in*/
  /* Whether the window covers the given text row. */

  /*in*/
  /* Whether the indicated row of the window covers the given text column. */

  /*in*/
  /* Number of topmost visible text row in window. */

  /*in*/
  /* Number of leftmost visible text column of the indicated text row
     in the window. If the row is not visible at all, return 0. */

  /* Scroll window vertically to reach the given topmost text line. */

  /* Scroll given text row (and surrounding band) horizontally to reach
     the given leftmost text column.
     Note, this routine may have no effect if the given text row is not
     now on the screen, before any cleaning operations. To ensure that
     it has effect, clean the window first. */

  /*io*/
  /* Scroll window to cover the given text row.
     No effect if window already covers it.
     Least scrolling step is Wd_Vscroll lines, but limited
     also by window size, e.g. a 1-line window can always
     be scrolled in 1-line steps. */

  /*io*/
  /* Scroll given text row (and surrounding band) to cover the given column.
     No effect if row already covers it.
     Least scrolling step is Wd_Hscroll columns, but limited
     also by window width, e.g. a 5-column window can always be
     scrolled in 5-column steps.
     Note, this routine may have no effect if the given text row is
     not on screen now, before any cleaning. To ensure that it has
     effect, clean the window first. */

  /* Insert a number of lines before the given text row. */

  /* Delete a number of lines starting with the given text row. */

  /* Delete a number of characters starting with the indicated
     row-column position.
     Note, the column is really that of the first character deleted,
     and not the column to the right of it. */

  /* Insert a number of characters on the given text row, at the
     given column. The rest of the line moves right. */

  /* Insert a line break in the given text row, at (just before) the
     given column. The rest of the line is inserted below the given
     row, as a line of its own. */

  /* Inform the window that the given text row needs to be redisplayed
     from the given column to end of line. */

  /* Inform the window that it needs to be completely redisplayed.
     As always, the actual display will be done with the help of
     Wd_Clean. */

  /*io*/
  /* o*/
  /* o*/
  /* Update display of window.
     Implements line deletion, insertion, and scrolling operations.
     Returns necessary screen operations in the ops parameter, or
     signals that the display is up to date by returning clean = true.
     May have to be called many times, in the form
          repeat
             Wd_Clean (win,ops,clean);
             if not clean then implement ops on screen;
             if (user interrupt) then exit loop;
          until clean (or user interrupt)
     See comments for Wd_ScreenOps for more information.
   */


  /*** Testing routines ***/

  /*i */
  /* List window "vertical" (i.e. row number) information on screen
     in vertical format starting at the indicated screen row and column. */

  /*i */
  /* List window "horizontal" (i.e. within-row edits) information on screen
     in vertical format starting at the indicated screen row and column. */

  /*i */
  /* List window information on text output file. */



  /* Submodules and own defs: */
  /* edit_d.p
     Constants and types for edit.p

     Changes:
       94.07.27  AbortMsg added.
       94.08.03  Touched, no real changes.
  */

  /* line length limit (Ed_Getln routine) */



  /* transcript input symbol count */

  /* How a transcript is processed, if at all */
  /* dead text file, no processing */
  /* process line by line, no cancel, no status */
  /* process job steps in sequence */
  /* process incrementally, cancel, status known */

  /* no message yet (internal to Edit) */
  /* text to process */
  /* reverse processing */
  /* abort processing */

  /* Index of start of text */
  /* Index to reverse to */
  /*Ed_MsgType*/
  /* views_d.p
     Data structures for "views" that consist of a screen window,
     a transcript, and a cursor position.

     Changes:
        94.08.04 Add modifiable and modified flags.
        94.10.07 Modified flag has different meaning (Transcripts module notes
                 changes, we just mirror the flag to know when to update the
                 window title).
        94.10.23 Add origname to track changes in transcript file name.
  */

  /* requires "edit_d.p" */



  /* Max length of transcript name */
  /* Max length of file name */

  /* Max number nested exec modes */

  /* Log levels */
  /* Pause between each editing/processing operation */
  /* Log window cleaning (updating) operations */
  /* Log window data structures in connection with above */


  /* Defines the start of a certain execution mode */
  /* manual, job, incremental */
  /* first input symbol covered */


  /* A transcript being edited */
  /* buffer name */
  /* file last loaded from or saved to */
  /* file for next save */
  /* the inputs and outputs */
  /* can be edited */
  /* content has changed since last save */
  /* defines a region mark..cursor */
  /* region defined (logically) */
  /* read pointer */
  /* write pointer */
  /* async write file desc, active if >= 0 */
  /* "output" stream file desc, active if >= 0 */
  /* last script sym read */
  /* process up to and incl this script sym */
  /* first changed script sym */
  /* script symbol in error */
  /* start of error token */
  /* length of error token */
  /* next script sym to be read or re-read */
  /* last script sym claimed to be "used" */
  /* ignore writes (cancelling) */
  /* user wants to abort this (sub)session */
  /* linked list of all transcripts */


  /* A window has been used to view ... */
  /* this transcript, */
  /* with the cursor here, */
  /* and perhaps others views, too. */

  /* How a window's title bar is displayed. */
  /* the window has no title bar */
  /* normal, unemphasized (not current) */
  /* bold, emphasized (current window) */

  /* currently viewed transcript and */
  /* cursor in it */
  /* present title bar style */
  /* screen row for window title bar */
  /* other transcript:cursor pairs */

  /* Hey window! We have your number! */

  /* views_f.p
     External declarations for routines in views.p
  */

  /* NOTE -- bad to have public vars, change! */


  /* Initialize this module. */

  /* You guessed it. */

  /* Return a (constant) value that behaves as an infinite ScriptIndex. */

  /* Returns number of windows open at present.
   The window numbers are 1..NumWindows.
  */

  /* Is the query area in use? */

  /* Define a place on a screen status line.
     Status line number num is = 0, 1, 2, .. relative to the first
     line reserved for status.
     Column position is relative to left margin, col = 1, 2, 3.
   */

  /* Creates a new editable transcript. The transcript name (window title)
     and initial execution mode must be specified, also whether the
     transcript is modifiable by the user.
     The transcript is initially empty and has a null filename.
     The new transcript is added to the set of transcripts maintained
     by the editor.
     The transcript is not yet displayed in any windows.
   */

  /* Destroys an editable transcript.
     Removes it from the the list of transcripts first.
   */

  /* Clean up (refresh) window number w until it is clean or
     user presses a key. Parameter 'clean' shows whether the
     window was cleaned.
   */

  /* Move screen cursor to text cursor position of this window.
     Scroll the window if necessary.
     Includes cleaning of this window, but not others.
   */

  /* Define the title bar style for a window. */

  /* Write title bar for this window. */

  /* Write title bar for all windows showing this transcript. */

  /* Divide a window into two halves, if possible.
     The upper half has the same index, the lower half a new one.
   */

  /* Close this window. Leave screen area unused. */

  /* Close all except the current window. Make the current window
     grow to use all of the available screen. */

  /* Make the window w view the transcript etr. */

  /* Return the first transcript. */

  /* Return the next transcript, or nil if none. */

  /* Create a new transcript for view in this window, specifying if
     it will be modifiable or for view only.
     Return success/failure.
   */

  /* Assign the file name of the current view in this window, if no
     name has yet been assigned.
     Use override = true to change a name assigned earlier.
     The transcript name is set to the base name of the file.
   */

  /* Read in a text or transcript from a file into the current view.
     The file ff is assumed to be opened already using the name fname and
     can contain pure text or a transcript.
     The file is then closed. Success is reported in ok.
   */

  /* Control whether transcript in the current view in this window is
     to be modifiable (editable, changeable) or not.
   */

  /* Whether the transcript in the current view in this window has
     been changed (input, output, or both) since its last "original"
     state (eg. creation, load from file). */

  /* Declare that this transcript is now in an "original" state, eg.
     just created and loaded from a file. */

  /* Redraw all windows on screen, including title bars. */

  /* Insert a number of characters at row,col in all windows
     for transcript etr. */

  /* Insert a number of characters at row,col in all windows
     for transcript etr. */

  /* Delete a number of lines starting with 'row' in all windows
     for transcript etr. */

  /* Insert a number of lines at 'row' in all windows
     for transcript etr. */

  /* Freshen (redisplay) line 'row' from column 'col' in all windows
     for transcript etr. */

  /* Make a newline at 'row' column 'col' in all windows for
     transcript etr. */

  /* Update screen after text has been inserted.
     Text (perhaps including newlines) was inserted at the left
     of the given spot. Before the insertion, the spot was at
     the given row and column. The transcript is refEtip. */

  /* Update screen after text has been deleted in a transcript.
     Text (perhaps including newlines) was deleted, starting at
     delRow,delCol and continuing up to but not including preRow,preCol. */

  /* Update screen after text (perhaps including newlines) was deleted
     at the left of the given spot. Before the deletion, the spot was
     at the given row and column. The transcript is refEtip. */

  /* Status line and similar output routines. */

  /* Display (render) a character for status line or other such. */

  /* Display (render) a string for status line or other such. */

  /* Format n as decimal integer. Fill with leading blanks to at least
     w characters. Result in s.
     Note: implementation is not efficient. */

  /* Show error result from Screen module. */

  /* Show error result from Transcript module. */

  /* Show error result from Windows module. */

  /* Show error result from Keymap module. */

  /* Display message. */

  /* Clear message line and note absence of message.
     Nothing done if there was no message to start with. */

  /* Message and pause.
     User can dump the indicated window, increase/decrease log levels,
     and ask program to stop.
     Returns "true" if execution is to continue.
   */

  /* Turn tracing on or off. */

  /* Open a new interaction (window + transcript) to ask the
     user for something, e.g. a file name to store the transcript in.
     The new window is returned in w. Success/failure in ok.
   */

  /* Delete the query interaction. */

  /* System: */
  /* Pascal translation.

     close, read, write renamed as sys_close, sys_read, sys_write,
     to avoid conflict with Pascal statements.

     Last changed 93.2.17.
  */


  /* WARNING: _SC_CLK_TCK and sysconf() are also defined/declared in <time.h>. */

  /* space for argv & envp */
  /* maximum children per process??? */
  /* clock ticks/sec */
  /* number of groups if multple supp. */
  /* max open files per process */
  /* do we have job control */
  /* do we have saved uid/gids */
  /* POSIX version supported */


  /* max links to file/dir */
  /* max line length */
  /* max "packet" to a tty device */
  /* max pathname component length */
  /* max pathname length */
  /* size of a pipe */
  /* trunc or error on >NAME_MAX */
  /* best char to shut off tty c_cc */
  /* can we give away files */
  /* highest value of any _PC_* */



  /*
    * lseek & access args
    */
  /* Set file pointer to "offset" */
  /* Set file pointer to current plus "offset" */
  /* Set file pointer to EOF plus "offset" */


  /* does file exist */
  /* is it executable by caller */
  /* is it writable by caller */
  /* is it readable by caller */


  /* types_p.h
     POSIX and system types
  */


  /* should be unsigned short */
  /* should be unsigned int */
  /* should be unsigned long */
  /* should be unsigned char */

  /* signal mask - may change */
  /* tty line disc modes */
  /* tty control char */
  /* process id */

  /* file mode bits */
  /* links to a file */

  /* units=ticks (typically 60/sec) */
  /* value = secs since epoch */

  /* len param for string funcs */






  /*Pascal!*/
  /*Pascal!*/
  /*Pascal!*/

  /* !_unistd_h */
  /* Pascal translation of parts of /usr/include/sys/fcntl.h */





  /* XXX close on exec request; must match UF_EXCLOSE in user.h */
  /* posix */

  /* fcntl(2) requests */
  /* Duplicate fildes */
  /* Get fildes flags (close on exec) */
  /* Set fildes flags (close on exec) */
  /* Get file flags */
  /* Set file flags */

  /* Set or Clear a record-lock (Non-Blocking) */
  /* Set or Clear a record-lock (Blocking) */
  /* Get record-locking information */

  /* fcntl(2) flags (l_type field of flock structure) */

  /* read lock */
  /* write lock */
  /* remove lock(s) */




  /* in unistd.p */


  /* Declarations for Unix "select" system call. */

  /* File descriptors sets (bitmaps) for use with Unix "select" call. */



  /* bits per word     */
  /* number of words   */
  /* total size (bits) */


  /* Returns word with only bit n set.
     Should use left shift function, but cannot remember
     what it's called in Sun Pascal. */
  long FD_One, x;

  static int _was_initialized = 0;
  if (_was_initialized++)
    return;
  /*FD_One*/
  if (n == 31) {
/* p2c: fd_sets_p.h, line 26: Warning: Symbol 'N' is not defined [221] */
    FD_One = 0x80000000L;   /* only negative value */
/* p2c: fd_sets_p.h, line 27:
 * Warning: Symbol 'FD_ONE' is not defined [221] */
    return;
  }
  x = 1;
/* p2c: fd_sets_p.h, line 28: Warning: Symbol 'X' is not defined [221] */
  for (i = 1; i <= n; i++) {
/* p2c: fd_sets_p.h, line 29: Warning: Symbol 'I' is not defined [221] */
/* p2c: fd_sets_p.h, line 29: Warning: Symbol 'N' is not defined [221] */
    x += x;
  }
  FD_One = x;
}
/* p2c: fd_sets_p.h, line 34: 
 * Warning: Expected a '.', found PROCEDURE [227] */
/* p2c: fd_sets_p.h, line 34: 
 * Warning: Junk at end of input file ignored [277] */
/* p2c: Note: Remember to call _Edit_init() in main program [215] */




/* End. */
