{  A SEED-like session editor for incremental (reversible) script
   execution.

Author: Niklas Holsti
Status:
  91.02.16 Under construction. No scrolling, no re-reading.
  91.02.27 Uses Screen module.
  91.03.10 Uses Windows module (just one transcript buffer, so far).
  91.04.08 Uses Keymap module.
  91.04.14 Multiple transcripts, multiple windows, cut, copy, paste.
  91.04.19 Reject control keys in typing.
           When opening a window, check that two new spots can be defined
           (one for the cursor, one for cleaning).
  91.04.21 Input and output spots separate from cursor spot.
  91.05.06 Joinline separate from DeleteChar.
           Ed_Getln and associated stuff added.
  91.06.17 All Explain routines have file parameter.
           Problems are reported both on screen and in logfile.
           Ed_Getln does not send read-ahead characers to the
           application until the user asks for reprocessing. Otherwise,
           the line-buffered Lex variant (Lexline.p) tries to read the
           whole line and fails if there is no newline.
           Ed_WrReal accepts 'Extended' type parameter.
  91.06.17 Ts_Insert has preRow/Col parameters.
  91.06.21 Ed_WriteLn uses NewlineWs which uses Wd_BreakLine.
  91.06.22 Window routines have line & char counts.
           MakeNewline handles tagged output (does not use InsShow).
           Display of status line made optional.
           CloseOtherWins sets titleRow for sole remaining window.
           Ed_GetLn applies Ts_Skip to move cursor to start of next
           input line.
  91.09.05 readlog added.
           Cancel sets usedPos:=infinite.
           DelShow does not ask Windows to delete zero characters.
           When window cleaner overstrikes a line, it clears the rest
           of the line, but only if the right margin was not overstruck.
           Note, this may have to be refined when "overstrike-mode" is
           implemented.
           Keymap directory name changed to "pSeed".
  91.09.14 Ef_readIn implemented (InsertFile).
  91.09.15 Log_Open included where it might be needed.
           Old where-line is erased when where-line option is turned off.
           If Delete-Char can't find input to delete, it just beeps
           rather than displays a techical Ts_ message.
  91.09.23 Look for keymap file both in current folder and in Sisu:...
       -24 Accepts preRow,preCol from Ts_Newline.
           ProcEnd and ProcItem implemented.
           ExecuteLine avoids moving cursor if the line is already
           executed, or there is an uncorrected error.
  91.09.25 Ed_Tell extended with errfst,errlen to locate error token,
           used in ProcItem.
  91.09.26 ExecPos is not changed when cancel msg is sent. Instead,
           execPos is zeroed when editing phase starts.
  92.04.10 Port to Sun OS.
           Remember if there is a message, so ClearMess updates
           screen only if necessary.
           Window cleaner uses Sc_Wrote.
           Message displays preceded by Sc_Limbo; we do not keep
           track of cursor position for messages.
           Ed_Fini called before halt; uses Sc_Fini.
           Window cleaner applies with-line overstriking only if
           the row-col exists in the text. (This optimization is
           a bit risky.) Ditto, insertion within line.
  92.04.17 Local vars declared 'private' to reduce risk of sym conflicts.
           Window title always in overstrike mode; cursor in limbo after.
           Begin implementation of "manual" and "job" execution modes.
  92.04.19 Ed_WriteFrom and Ed_WriteEnds added, for handling output from
           subprocesses.
  92.11.29 Keymap file default name changed (really must use $HOME etc...)
  92.12.01 Added 'processfrom' field to messages from editor to application,
           and renamed 'index' field of cancel messages to 'cancelto'.
           Set the fields in 'Process' and Ed_Getln.
           UnWritePastLast corrected (seek to pos+2).
  92.12.02 Refresh-screen (Ef_refresh) added.
           When writing from a file descriptor into the transcript, if there is
           an error in reading it, or if there is no data when there should be,
           the write-from-fd function is turned off.
  92.12.09 Errors in write-from-fd were still not handled correctly. Tried
           to fix, but not a good solution.
  92.12.22 Messages do not beep, unless they are errors.
  93.03.29 Render module used, rather than direct write to output.
           Some errors and warnings are still written to errout, rather
           than the render stream -- most should be rendered, instead.
  93.03.31 Writes pending data from write-channel (writefd) when a
           reversal is reported in Ed_Tell, before the write-point is
           moved (output unwritten). For this, added procedures
           DataAvailable and WriteAndCheck.
           Pending write-channel data is also written at the start
           of an Ed_Getln call, so that line-interleaved output
           gets inserted correctly when a multi-line region is
           executed in one go. (This is a hack, and needs redesign.)
  93.04.01 Ed_written defined and exported, since it seems that the
           applications at times need to make sure that write-channel
           data are processed. Otherwise, they output can be tagged with
           too late script positions, and thus deleted too early etc.
  93.05.14 CutRegion and Paste marked as "changed" the position just before
           the affected region - added 1 to pos(cursor) to fix it.
           When incremental mode execution ends, the location of the
           most recent edit is forgotten, to stop cancellation (when
           the incremental mode ends because some text before its start
           is changed). Also, writes will again be performed (skipwrite
           is reset).
           Overlapping execution mode test is less stringent: a new
           mode can now start at the same position as the current mode,
           but not before.
  93.07.29 Ed_WriteFile added.
           Ed_CatchOutput and Ed_ShowOutput added.
           Local procedure Pause renamed to MessPause (conflict with "pause"
           in unistd).
  94.03.20 Port to Solaris 2.2.
  94.05.05 Ed_MsgType contains a string of data instead of one char.
           Ts_Reads used.
           Overlong lines are reported in a different way: the whole
           line is fetched from Ts, and the error message appears
           after the line, rather than embedded.
           Accept char codes >= 128.
           NameFile, GoToTop and GoToBottom editor functions added.
           MaxTipName and MaxFileName added. AskParamString takes
           max string length parameter.
  94.05.20 Key-map file from /usr/local/src/seed/lib/keymap.t.
  94.07.25 Editor becomes re-entrant: "query transcript" for file names etc.
           is another editable transcript (vide emacs minibuffers ?).
           Some reordering of procedures to account for this.
  94.07.27 Add Ef_abort, Ef_complete, Ef_query.
           While in query mode, prevent editing and execution of other
           transcripts (except NoExec transcripts).
           Key map file is named ".seed-keymap" and is sought in
           working directory, $HOME, and /usr/local/src/seed/lib, in that
           order.
  94.07.28 Clean current window immediately upon entering edit mode, to
           display recent output without delay.
           Tracing and logging functions enabled only if "-debug" is
           specified on command line.
  94.08.02 Begin to move parts to a new "Views" module, to contain data and
           routines for the editor context (windows, transcripts, views).
           Diving line: all idea of "current" window remains here, and all
           user interactions (except for debug) remain here.
  94.08.04 In unprocessable transcripts, any process command works as a 
           newline.
  94.09.28 Modifiable/Modified status, "view file" function. Start work.
  94.10.07 Unmodifiable transcript rejects "mutator" edit functions.
           Instead of "allowed editing function" sets, use "function
           maps" that map unallowed functions to Ef_undefined. This lets
           us map process-functions to something useful even in NoExec
           documents.
           Declares "original" (unmodified) status for new transcripts etc.
  94.10.23 On exit, check for unsaved transcripts. Either save them or
           ask user for confirmation to throw changes away, depending on
           the kind of exit command given.
           On a partial save (eg. write input only) always asks for a new
           file name. On a complete save, asks for a new file name if none has
           been defined for the transcript yet.

  TO DO:
           - check file names for proper syntax, existing files etc.
}

module Edit;

{ Independent but required modules: }
#include "strings_d.p"
#include "strings_f.p"
#include "log_f.p"
#include "render_f.p"
#include "screen_d.p"
#include "screen_f.p"
#include "files_f.p"
#include "argument_f.p"
#include "editfuns.p"
#include "keymap_d.p"
#include "keymap_f.p"
#include "transcripts_d.p"
#include "transcripts_f.p"
#include "windows_d.p"
#include "windows_f.p"

{ Submodules and own defs: }
#include "edit_d.p"
#include "views_d.p"
#include "views_f.p"

{ System: }
#include "unistd_p.h"
#include "fcntl_p.h"
#include "select_p.h"



const

  { Log levels }
  spotchecklog = 1;   { Check transcript spot-data consistency }
  readlog      = 1;   { Log all process-input and cancel operations }
  writelog     = 1;   { Log all writes into the transcript }
  unwritelog   = 1;   { Log all unwrite operations }
  edfunlog     = 1;   { Log all editing functions }
  usedlog      = 1;   { Log changes to the 'used' position index }
  wherelog     = 2;   { Log cursor and read/write spot locations }
  tracelog     = 2;   { Trace many procedure calls }
  

type

  EdFunSet = set of EditFun;
  EdFunMap = array[EditFun] of EditFun;

private var

  rcPlace  : ScreenPlace;  { Place to show row,col counters }
  posPlace : ScreenPlace;  { Place to show edit/exec/last etc. indices }

  infiniteindex: ScriptIndex;  { = Vw_InfiniteIndex (constant) }

  { Global constants }
  allEditFuns  : EdFunSet;   { All editor functions. }
  debugFuns    : EdFunSet;   { Those that are for debugging the editor. }
  mutatorFuns  : EdFunSet;   { Those that change the transcript. }
  processFuns  : EdFunSet;   { Those that call for applic. processing. }
  queryFuns    : EdFunSet;   { Those that use query mode. }

  debugmode: boolean;    { Whether debug functions are enabled. }
  autoshow : boolean;
  tracing  : boolean;    { Trace and pause for many editor routines. }
  whereline: boolean;    { Display status line with cursor posn etc. }


  { Current window and transcript info. }
  curwin   : WinIndex;   { edwin[curwin] is the current window }
  curmap   : EdFunMap;   { the current editing function map }

  activeEtip: EtipRef;    { The active transcript (the one that will
                            get input). Currently, this is the only
                            'processable' transcript', and it is either
                            the application (main) transcripts, or the
                            query mode transcript. }

  { Query mode globals }
  querymode: boolean;    { Query mode is when the user is being asked to
                           name a file to load, for example. }
  querywin : WinIndex;   { The query window itself. }
  askerwin : WinIndex;   { The window from which query mode was entered. }
  askeract : EtipRef;    { Which transcript was "active" when the query
                           mode was entered. Note, this is not always
                           the transcript viewed in askerwin. }

  { Cut and copy buffer }
  clipbuf  : Transcript; { The cut/copy buffer. }
  clipeot  : Spot;       { At end of clipbuf, always. }

  { Scratch "result" flags }
  tsresult : Ts_result;
  scresult : Sc_result;
  wdresult : Wd_result;


procedure Ed_Fini;
  { Finalize editor. }
  begin
  Vw_Fini;
  end {Ed_Fini};

procedure WriteIndex (index: ScriptIndex);
  var s: vstring;
  begin
  Vw_DecimalString(index,1,s);
  Vw_WriteStr(s);
  end {WriteIndex};

private
procedure Beep;
  begin
  Vw_WriteChr(chr(7));
  end {Beep};

private
procedure GoPlace (p: ScreenPlace);
  begin
  with p do Sc_Goto(row,col,scresult);
  Vw_ScProblem(scresult);
  end {GoPlace};

private
procedure WritePos (p: ScriptIndex);
  { Display (render) a position, with special case for infinite. }
  var s: vstring;
  begin
  if p = infiniteindex then s:='  '
                          else Vw_DecimalString(p,3,s);
  Vw_WriteStr(s);
  end {WritePos};

private
procedure WriteRC (code: char; theSpot: Spot; var tip: Transcript);
  { Write row and column numbers. }
  var row,col: integer;
      d: vstring;
  begin
  Ts_Where(theSpot,tip,row,col);
  Sc_OverstrikeMode;
  Vw_WriteChr(code);
  Vw_WriteChr('['); Vw_DecimalString(row,2,d); Vw_WriteStr(d);
  Vw_WriteChr(','); Vw_DecimalString(col,2,d); Vw_WriteStr(d); Vw_WriteChr(']');
  if Log_includes(wherelog) then
    writeln(logfile,'...',code,' at text  row ',row:1,' col ',col:1);
  end {WriteRC};

private
procedure Where;
  { Report cursor position and script indices on status line. }
  begin
  with edwin[curwin], refEtip^ do begin
    GoPlace(rcPlace);
    Sc_OverstrikeMode;
    WriteRC('C',cursor ,theTranscript); Vw_WriteChr(' ');
    WriteRC('R',readpt ,theTranscript); Vw_WriteChr(' ');
    WriteRC('W',writept,theTranscript);
    GoPlace(posPlace);
    Vw_WriteStr('Last '   ); WritePos(lastPos );
    Vw_WriteStr('  Start '); WritePos(startPos);
    Vw_WriteStr('  Exec ' ); WritePos(execPos );
    Vw_WriteStr('  Err '  ); WritePos(errorPos);
    Vw_WriteStr('  Edit ' ); WritePos(editPos );
    Sc_ClearEOL;
    end;
  end {Where};

private
procedure NoWhere;
  { Erase the output from the Where routine. }
  begin
  GoPlace(rcPlace);
  Sc_ClearEOL;
  end {NoWhere};

private
procedure KmProb (result: Km_result);
  { Show error result from Keymap module on errout and logfile. }
  begin
  Km_Explain(errout,result);
  Log_Open;
  write(logfile,'Keymap: ');
    Km_Explain(logfile,result);
    writeln(logfile);
  end {KmProb};

procedure exit(c: integer);
  external c;

private
procedure HaltProgram;
  { Cruel abortion. }
  begin
  Log_close('Edit halts program.');
  Ed_Fini;
  flush(errout);
  exit(1);
  end {HaltProgram};

private
procedure MessPause (txt: vstring);
  { Message and pause. }
  begin
  Where;
  if not Vw_MessPause (txt, curwin) then HaltProgram;
  end {MessPause};

private
procedure MapToOne (funs: EdFunSet; tofun: EditFun; var map: EdFunMap);
  { Map all editing functions in "funs" to "tofun". }
  var f: EditFun;
  begin
  for f:= Ef_undefined to Ef_ambiguous do
    if f in funs then map[f]:= tofun;
  end {MapToOne};

private
procedure MapFunctions (w: WinIndex; var map: EdFunMap);
  { Which editing functions are allowed for window w, in
    the current editor state? How should the functions
    be mapped? }
  var
    f  : EditFun;
    dis: EdFunSet;  { functions that will be disabled }
  begin

  { Start off with all functions allowed and mapped to themselves. }
  dis:= [];
  for f:= Ef_undefined to Ef_ambiguous do map[f]:= f;

  { Depending on transcript state...}
  with edwin[w], refEtip^ do begin
    if refEtip = activeEtip then
      { no remaps }
    else if modeStack[modeTop].mode = NoExec then
      { A pure document. Remap eg. Enter key to newline. This map
        may be overriden by "dis" below. }
      MapToOne (processFuns, Ef_newline, map)
    else
      { An executable, but not the active one. Should not modify etc. }
      dis:= dis + mutatorFuns + processFuns;
    if not modifiable then
      dis:= dis + mutatorFuns + processFuns;
    end;

  { Depending on editor state...}
  if     querymode then dis:= dis + queryFuns;
  if not debugmode then dis:= dis + debugFuns;

  { Disable the bad things. }
  MapToOne (dis, Ef_disabled, map);
  end {MapFunctions};


private
procedure MoveToWindow (w: WinIndex);
  { Make window number w the current window.
    The current window will have a "bold" title bar.
    When in the query window, also the window from which query
    mode was entered will have a bold title.
    The query window itself has no title bar.
    Note, this does not set the allowed editing functions, because
    when exiting query mode, this must be done in query mode, but
    the functions must be set in non-query mode.
  }
  var
    noop: boolean;  { moving from query window to asker window or vv }
  begin
  if w <> curwin then begin
    { Normalize old window }
    if not querymode then Vw_Title (curwin, NormalTitle)
    else begin
      noop:=  (curwin = querywin) and (w = askerwin)
           or (curwin = askerwin) and (w = querywin);
      if not noop then
        if curwin = querywin then Vw_Title (askerwin, NormalTitle)
                             else Vw_Title (curwin  , NormalTitle);
      end;
                     
    curwin:= w;

    { Embolden new window }
    if not querymode then Vw_Title (curwin, BoldTitle)
    else
      if not noop then
        if curwin = querywin then Vw_Title (askerwin, BoldTitle)
                             else Vw_Title (curwin  , BoldTitle);

    end;
  end {MoveToWindow};


private
procedure SelectFunctions;
  { Select and map the editing functions for the current window. }
  begin
  MapFunctions(curwin, curmap);
  end {SelectFunctions};


private
procedure SelectWindow (w: WinIndex);
  { Select a new current window and the corresponding editing
    functions.
  }
  begin
  MoveToWindow (w);
  SelectFunctions;
  end {SelectWindow};


private
procedure Changed (Pos: ScriptIndex);
  { Script position Pos has been changed by the editor.
    Note it in editPos and errorPos.
    Unwrite output from Pos and on. }
  begin
  with edwin[curwin].refEtip^ do begin
    if Pos < editPos then begin
      editPos:=Pos;
      if editPos <= errorPos then errorPos := infiniteindex;
      if Log_includes(unwritelog) then
        writeln(logfile,'Ed.Changed unwrite=',Pos-1:1);
      Ts_Unwrite(writept,theTranscript,Pos-1);
      end;
    end;
  end {Changed};

private
function Logfun: boolean;
  { Whether editing functions are logged. }
  begin
  Logfun:=Log_includes(edfunlog);
  end {Logfun};

private
procedure Typing (k: char);
  { Insert character k at cursor. }
  var preRow,preCol: integer;
  begin
  if Logfun then writeln(logfile,'Typing "',k,'"');
  if (ord(k) < 32) or (ord(k) = 127) then begin
    Beep;
    Vw_Message('[Nontyping key rejected]');
    end
  else
  with edwin[curwin], refEtip^ do begin
    Ts_Insert(cursor,theTranscript,k,preRow,preCol,tsresult);
    if tsresult=Ts_ok then begin
      Changed(Ts_Pos(cursor,theTranscript));
      Vw_InsShow(cursor,refEtip,preRow,preCol);
      end
    else
      Vw_TsProblem(tsresult)
    end;
  end {Typing};

private
procedure MakeNewline;
  { Insert newline at cursor. }
  var row,col: integer;
  begin
  if Logfun then writeln(logfile,'MakeNewline');
  with edwin[curwin], refEtip^ do begin
    if tracing then MessPause('Ts_Newline..');
    Ts_Newline(cursor,theTranscript,row,col,tsresult);
    if tsresult = Ts_ok then begin
      Changed(Ts_Pos(cursor,theTranscript));
      Vw_FreshLineWs(refEtip,row,col);
      Ts_Where   (cursor,theTranscript,row,col);
      Vw_InsLineWs  (refEtip,row,1)
      end
    else
      Vw_TsProblem(tsresult);
    end;
  end {MakeNewline};

private
procedure MoveCursor (rows,cols: integer);
  { Move cursor by some number of rows and columns. }
  begin
  if tracing then MessPause('Moving');
  if Logfun then writeln(logfile,'Move ',rows:1,' rows, ',cols:1,' cols');
  with edwin[curwin], refEtip^ do begin
    Ts_Move(cursor,rows,cols,cursor,theTranscript);
    end;
  end {MoveCursor};

private
procedure GoToExtremes (row: integer);
  { Move cursor by some extreme number of rows, to column one. }
  begin
  if tracing then MessPause('GoToExtremes');
  if Logfun then writeln(logfile,'GoToExtremes ',row:1,' row');
  with edwin[curwin], refEtip^ do begin
    Ts_Goto(cursor,theTranscript,row,1);
    end;
  end {GoToExtremes};

private
procedure Joinline;
  { Delete newline preceding cursor, joining the line holding the
    cursor to the preceding line. Also deletes output that is due
    to the deleted newline, and output embedded in the cursor line. }
  var delRow,delCol,preRow,preCol,row,col: integer;
      chind: ScriptIndex;
  begin
  if Logfun then writeln(logfile,'Join line');
  with edwin[curwin], refEtip^ do begin
    Ts_Joinline(cursor,theTranscript,
                delRow,delCol,preRow,preCol,
                tsresult);
    if tsresult = Ts_ok then begin
      Vw_DelShow(refEtip,delRow,delCol,preRow,preCol);
      chind:= Ts_Pos(cursor,theTranscript) + 1;
      Changed(chind);
      Ts_Where(cursor,theTranscript,row,col);
      Vw_FreshLineWs(refEtip,row,col);
      end;
    end;
  Vw_TsProblem(tsresult);
  end {Joinline};

private
procedure DeleteChar (canjoin: boolean);
  { Delete the character to the left of the cursor.
    Also delete any output that immediately follows it.
    If the input symbol the the left of the cursor is a newline,
    then if 'canjoin' is true, a Joinline is done, otherwise
    an error is reported. } 
  var preRow,preCol: integer;
      chind: ScriptIndex;
  begin
  if Logfun then writeln(logfile,'Delete char');

  with edwin[curwin], refEtip^ do begin
    Ts_Delete(cursor,theTranscript,preRow,preCol,tsresult);
    if tsresult = Ts_ok then begin
      Vw_DelShot(cursor,refEtip,preRow,preCol);
      chind:= Ts_Pos(cursor,theTranscript) + 1;
      Changed(chind);
      Ts_Dewrite(cursor,theTranscript,preRow,preCol,tsresult);
      Vw_DelShot(cursor,refEtip,preRow,preCol);
      end
    else
    if (tsresult = Ts_wasnewline) and canjoin then
      Joinline;
    end;

  if tsresult = Ts_notinput then begin
    tsresult:= Ts_ok;  { avoid technical message }
    Beep;              { just honk }
    end
  else
    Vw_TsProblem(tsresult);
  end {DeleteChar};

private
procedure BackSpace;
  { Backspace one character. In InsertMode, this behaves like
    the DeleteChar function. In OverstrikeMode, the cursor is
    moved left one space and a blank is put there, but the rest
    of the line is not moved. }
  begin
  DeleteChar(true);          { ** Note, different in Overstrike-Mode ** }
  end {BackSpace};


private
procedure MarkThisPoint;
  { Set markpt := cursor pt. }
  begin
  if Logfun then writeln(logfile,'Mark this point');
  with edwin[curwin], refEtip^ do begin
    Ts_Move(cursor, 0,0, theMark, theTranscript);
    marked:=true;
    end;
  Vw_Message('[Mark set]');
  end {MarkThisPoint};


private
procedure SwapMark;
  { Swap mark and cursor positions. }
  var pt: Spot;
  begin
  with edwin[curwin], refEtip^ do begin
    pt:=theMark;
        theMark:=cursor;
                 cursor:=pt;
    end;
  end {SwapMark};


private
procedure CutRegion;
  { Cut the marked region from the current transcript and put the
    data in clipbuf. The transcript becomes unmarked.
    Also delete any output that immediately follows the region. }
  var order,preRow,preCol: integer;
  begin
  with edwin[curwin], refEtip^ do
    if not marked then Vw_Message('[Mark is not set]')
    else begin
      order:=Ts_Order(theMark,cursor,theTranscript);
      if order = 0 then Vw_Message('[Empty region]')
      else begin
        if order > 0 then SwapMark;
        Ts_Erase(clipbuf);
        Ts_Cut (theMark,cursor,theTranscript,
                clipeot,clipbuf,preRow,preCol,tsresult);
        if tsresult = Ts_ok then begin
          Vw_DelShot(cursor,refEtip,preRow,preCol);
          Changed(Ts_Pos(cursor,theTranscript)+1);
          Ts_Dewrite(cursor,theTranscript,preRow,preCol,tsresult);
          Vw_DelShot(cursor,refEtip,preRow,preCol);
          if tsresult = Ts_ok then begin
            marked:=false;
            Vw_Message('[Region cut]');
            end
          else
            Vw_TsProblem(tsresult);
          end
        else
          Vw_TsProblem(tsresult);
        end {nonempty region};
      end {marked};
  end {CutRegion};


private
procedure CopyRegion (asInput: boolean);
  { Copy the marked region from the current transcript and put the
    data in clipbuf. The transcript becomes unmarked. }
  var swapped: boolean;
  begin
  with edwin[curwin], refEtip^ do
    if not marked then Vw_Message('[Mark is not set]')
    else begin
      swapped:=Ts_Order(theMark,cursor,theTranscript) > 0;
      if swapped then SwapMark;
      Ts_Erase(clipbuf);
      Ts_Copy (theMark,cursor,theTranscript,
               clipeot,clipbuf,asInput,tsresult);
      Vw_TsProblem(tsresult);
      if swapped then SwapMark;
      marked:=false;
      Vw_Message('[Region copied]');
      end {marked};
  end {CopyRegion};


private
procedure Paste (asInput: boolean);
  { Copy text from clipbuf to cursor position. }
  var clipbot: Spot;
      preRow,preCol,prePos: integer;
      row,col             : integer;
  begin
  with edwin[curwin], refEtip^ do begin
    Ts_Where(cursor,theTranscript,preRow,preCol);
    prePos:=Ts_Pos(cursor,theTranscript);
    Ts_NewSpot(clipbot,clipbuf,tsresult);
    if tsresult <> Ts_ok then   { could not make clipbot spot }
      Vw_TsProblem(tsresult)
    else begin
      Ts_Copy(clipbot,clipeot,clipbuf,
              cursor,theTranscript,asInput,tsresult);
      Vw_TsProblem(tsresult);
      Ts_Where(cursor,theTranscript,row,col);
      if (row <> preRow) or (col <> preCol) then begin
        Changed(prePos+1);
        Vw_InsShow(cursor,refEtip,preRow,preCol);
        end {something pasted};
      Ts_DelSpot(clipbot,clipbuf);
      end;
    end;
  end {Paste};


private
procedure OpenWin;
  { Divide current window into two halves. }
  var lower: WinIndex;
  begin
  if Vw_Split(curwin, lower) then ;
  end {OpenWin};

private
procedure CloseOtherWins;
  { Close all except current window, which grows to fill screen. }
  begin
  Vw_CloseOthers(curwin);
  end {CloseOtherWins};

private
procedure NextWin;
  { Select next lower window, or wrap around to top window. }
  begin
  SelectWindow( (curwin mod Vw_NumWindows) + 1);
  end {NextWin};


private
function SafePos (var {i} etip: EditTranscript): integer;
  { The number of the last read input position, or the number
    the processor claims it has interpreted. }
  begin
  with etip do
    if usedPos < lastPos then SafePos:=usedPos
                         else SafePos:=lastPos;
  end {SafePos};


procedure Ed_Write (txt: vstring);
  var preRow,preCol: integer;
  begin
  with edwin[curwin], refEtip^ do begin
    Ts_Where(writept,theTranscript,preRow,preCol);
    if Log_includes(writelog) then begin
      write(logfile,'Write [',preRow:1,'.',preCol:1,'] "',txt,'"');
      if skipwrite then write(logfile,' - skipped');
      writeln(logfile);
      end;
    if not skipwrite then begin
      Ts_Write(writept,theTranscript,txt,SafePos(refEtip^),tsresult);
      if tsresult = Ts_ok then
        Vw_InsShow(writept,refEtip,preRow,preCol)
      else
        Vw_TsProblem(tsresult);
      end;
    end;
  end {Ed_Write};


procedure Ed_Writeln;
  var row,col: integer;
  begin
  with edwin[curwin], refEtip^ do begin
    Ts_Where  (writept,theTranscript,row,col);
    if Log_includes(writelog) then begin
      write(logfile,'Writeln [',row:1,'.',col:1,']');
      if skipwrite then write(logfile,' - skipped');
      writeln(logfile);
      end;
    if not skipwrite then begin
      Ts_Writeln(writept,theTranscript,SafePos(refEtip^),tsresult);
      if tsresult = Ts_ok then
        Vw_NewLineWs(refEtip,row,col)
      else
        Vw_TsProblem(tsresult);
      end;
    end;
  end {Ed_Writeln};


private
procedure IntegerDecimal (i: integer; var digs: vstring);
  { Integer to decimal string. }
  var a: integer;
  begin
  if i>=0 then a:= i
          else a:=-i;
  digs:='';
  repeat
    digs:=concat(chr((a mod 10)+ord('0')),digs);
    a:=a div 10;
  until a = 0;
  if i<0 then digs:=concat('-',digs);
  end {IntegerDecimal};


procedure Ed_WrInt (i: integer);
  var digs: vstring;
  begin
  IntegerDecimal(i,digs);
  Ed_Write(digs);
  end {WrInt};


var
  writebuf: packed array [1..maxstring+1] of char;  { see Ed_WrReal }

function sprintf(s,format: univ_ptr; 
                 w,d     : integer;
                 r       : real): univ_ptr;
  external c;

procedure Ed_WrReal (r: real; w,d: integer);
  { Note, this implementation uses the C "sprintf" function. }
  var s: vstring;
      u: univ_ptr;
      i: integer;
  begin
  u:=sprintf(addr(writebuf[1]),addr('%*.*f'),w,d,r);
  i:=0;
  repeat i:=i+1
  until (i > maxstring) or else (ord(writebuf[i]) = 0);
  s:=substr(writebuf,1,i-1);
  Ed_Write(s);
  end {Ed_WrReal};


function Ed_WriteFrom (fd: integer): boolean;
  begin
  with edwin[curwin].refEtip^ do begin
    Ed_WriteFrom:= writefd < 0;
    if writefd < 0 then writefd:=fd
    else Vw_Message('Write-channel already active');
    end;
  end {Ed_WriteFrom};


function Ed_WriteEnds: boolean;
  begin
  with edwin[curwin].refEtip^ do begin
    Ed_WriteEnds:= writefd >= 0;
    if writefd >= 0 then writefd:=-1
    else Vw_Message('Write-channel already ended');
    end;
  end {Ed_WriteEnds};


function read (fd: integer; buf: univ_ptr; len: integer): integer;
  external c;

private
function WriteFromDesc (fd: integer): integer;
  { Write one hunk of data into the current transcript
    from the file descriptor (e.g. pipe) fd.
    Return number of bytes read, or < 0 for error. }
  var data: packed array[1..maxstring] of char;
      num : integer;
      i,d : integer;
      code: integer;
  begin
  num:=read(fd,addr(data[1]),sizeof(data));
  if num < 0 then
    Vw_Message('WriteFromDesc: read failed.')
  else
  if num > 0 then begin
    d:=1;   { start of line (or, temporarily, a newline char) }
    i:=1;   { next char to inspect }
    while i <= num do begin
      code:=ord(data[i]);
      if code = 10 then begin   { newline }
        if d < i then Ed_Write(substr(data,d,i-d));
        Ed_WriteLn;
        d:=i+1;  { start of next line? }
        end
      else
        { Replace unprintable chars by '.'.
          This should eventually be replaced by a proper display
          as ^key etc. }
        if                    (code <  32)
        or ((code >= 127) and (code < 161))
        or ( code  = 255)
        then data[i]:='.';
      i:=i+1;
      end;
    if d < i then Ed_Write(substr(data,d,i-d));
    end;

  WriteFromDesc:= num;
{$w-}   { no warnings about 'addr' }
  end {WriteFromDesc};
{$w+}


private
function DataAvailable (fd: integer): boolean;
  { Poll a file descriptor to see if data is available to
    be read. No timeout is used, just a non-blocking poll.
    Special case: no data is ever available from negative fd's. }
  var fdr,fdw,fde: fd_set_t;
      s  : integer;
      tim: timeval;
      tp : timeval_ptr;
  begin
  if fd < 0 then
    DataAvailable:= false
  else begin
    FD_ZERO(fdr); fdw:=fdr; fde:=fdr;
    FD_SET(fd,fdr);
    tim.tv_sec := 0;
    tim.tv_usec:= 0;
    tp:=addr(tim);
    s :=select(fd+1,fdr,fdw,fde,tp);
    if s < 0 then begin
     Vw_Message('DataAvailable: error in "select"');
      DataAvailable:= false;
      end
    else
      DataAvailable:= s > 0;
    end;
{$w-}   { suppress addr() warnings }
  end {DataAvailable};
{$w+}


private
procedure WriteAndCheck (fd: integer);
  { Write from the write-channel fd, check for errors,
    close down the channel if an error occurs. }
  var wrnum: integer;
  begin
  wrnum:=WriteFromDesc(fd);
  if wrnum <= 0 then begin
    if wrnum = 0 then
      Vw_Message('Internal error: no data in "read"');
    if Ed_WriteEnds then;   { UGH - this depends on there being only
                              one write-channel. Actually, 'fd' should
                              be a ref to some kind of write-channel
                              object, and here we should tell THAT object
                              that there is a problem, and to shut down. }
    end;
  end {WriteAndCheck};


procedure Ed_written;
  begin
  with edwin[curwin], refEtip^ do      { assume we mean this Etip }
    while DataAvailable(writefd) do WriteAndCheck(writefd);
  end {Ed_Written};


function Ed_CatchOutput: boolean;
  var tmp,pid: vstring;
      ok : boolean;
  begin
  { Construct a name for the temporary file. }
  IntegerDecimal(getpid,pid);
  tmp:='/tmp/ed_output.' + pid;
  { Create and truncate the file. }
  rewrite(output,tmp);
  { Open it thru another fd, for reading the text back. }
  tmp:=tmp+chr(0);
  with edwin[curwin].refEtip^ do begin
    outputfd:= sys_open(addr(tmp[1]),O_RDONLY,0);
    if outputfd = -1 then begin
      Vw_Message('CatchOutput: sys_open failed.');
      ok:= false;
      end;
    end;
  { Make it temporary. }
  if unlink(addr(tmp[1])) <> 0 then begin
    Vw_Message('CatchOutput: unlink failed.');
    ok:= false;
    end;
  Ed_CatchOutput:= ok;
{$w-}  { Yes, we know that addr is applied to tmp, thank you very much. }
  end {Ed_CatchOutput};
{$w+}


procedure Ed_ShowOutput;
  const maxtmp = 1000000;    { max size of temp file, bytes. }
  var   wrnum: integer;
  begin
  { Force text from Pascal buffers to file system. }
  flush(output);
  { See if outputfd is defined. }
  with edwin[curwin].refEtip^ do
    if outputfd < 0 then
      Vw_Message('ShowOutput: output not caught.')
    else begin
      { Read what there is and move it to the transcript. }
      repeat wrnum:=WriteFromDesc(outputfd)
      until wrnum <= 0;
      { Error? If so, close down. }
      if wrnum < 0 then begin
        if sys_close(outputfd) <> 0 then ;
        outputfd:= -1;
        end
      else
        { Check if file is large. }
        if lseek(outputfd,0,SEEK_CUR) > maxtmp then begin
          { Vw_Message('Renewing output catcher.'); }
          { Delete this file, make new one. }
          if sys_close(outputfd) <> 0 then
            Vw_Message('ShowOutput: sys_close failed.');
          if Ed_CatchOutput then ;
          end;
      end {catching};
  end {Ed_ShowOutput};

    
private
function WaitForData (delay: integer): integer;
  { Wait until some data are available from the keyboard or
    from a write-channel, or until a timeout of 'delay'
    milliseconds (if zero, unbounded wait).
    Returns number of file descriptor (0 = keyboard), or
    -1 for time-out or error. }
  var fdr,fdw,fde: fd_set_t;
      s  : integer;
      tim: timeval;
      tp : timeval_ptr;
  begin
  FD_ZERO(fdr); fdw:=fdr; fde:=fdr;
  FD_SET(0,fdr);
  {** Here, should inspect all active transcripts. **}
  with edwin[curwin].refEtip^ do
    if writefd >= 0 then FD_SET(writefd,fdr);

  if delay <= 0 then tp:=nil
  else begin
    tim.tv_sec := delay div 1000;
    tim.tv_usec:=(delay mod 1000) * 1000;
    tp:=addr(tim);
    end;

  Rn_Flush;
  s:=select(fd_width,fdr,fdw,fde,tp);

  if s < 0 then begin
    Vw_Message('Error in "select"');
    WaitForData:= -1;
    end
  else if s = 0 then begin
    if delay <= 0 then Vw_Message('Timeout in "select"');
    WaitForData:= -1;
    end
  else if FD_ISSET(0,fdr) then  { preference to keystrokes }
    WaitForData:=  0
  else
    with edwin[curwin].refEtip^ do
      if writefd >= 0 then begin
        if FD_ISSET(writefd,fdr) then
          WaitForData:= writefd
        else begin
          Vw_Message('No data in "select"');
          WaitForData:= -1;
          end
        end
      else begin
        Vw_Message('Problem in "select"');
        WaitForData:=-1;
        end;

{$w-}   { suppress addr() warnings }
  end {WaitForData};
{$w+}


procedure Ed_Tell (last,error,start: ScriptIndex;
                   errfst: ScriptIndex;
                   errlen: integer);
  begin
  with edwin[curwin], refEtip^ do begin { assume we are Telling this Etip }
    if last < lastPos then begin
      { Write pending output before writept is moved by unwrite, below }
      while DataAvailable(writefd) do WriteAndCheck(writefd);
      if Log_includes(unwritelog) then
        writeln(logfile,'Ed_Tell unwrite=',last:1);
      Ts_Unwrite(writept,theTranscript,last);
      end;
    lastPos    :=last;
    errorPos   :=error;
    errorFirst :=errfst;
    errorLength:=errlen;
    if errorPos <= execPos then begin
      execPos:=-1;
      Ts_Seek(cursor,errorFirst+errorLength-1,cursor,theTranscript,tsresult);
      if tsresult = Ts_ok then Ts_Move(cursor,0,+1,cursor,theTranscript)
                          else Vw_TsProblem(tsresult);
      end;
    startPos:=start;
    if tracing then MessPause('Ed_Tell');
    end;
  end {Ed_Tell};


{ QUERY MODE }


procedure Ed_Getln (var line  : Ed_LineType;
                    var len   : integer;
                    var msg   : Ed_MsgType );
  forward;


private
function BeginUserQuery: boolean;
  { Suspend ordinary editing and open a new interaction to ask the
    user for something, e.g. a file name to store the transcript in.
  }
  var queryok: boolean;
  begin
  Vw_CreateQuery(querywin, queryok);
  if queryok then begin
    querymode := true;
    askerwin  := curwin;
    askeract  := activeEtip;
    activeEtip:= edwin[querywin].refEtip;
    SelectWindow(querywin);
    end;
  BeginUserQuery:= queryok;
  end {BeginUserQuery};


private
procedure RestartQuery (error: ScriptIndex);
  { Report error in query transcript.
    The erroneous input is defined as everything up to and including
    the "error" position.
    Let user fix and try again.
  }
  begin    
  Ed_Tell ({last}   0, {error}  error, {start} 1,
           {errfst} 1, {errlen} error);
  end {RestartQuery};


private
procedure EndUserQuery;
  { Resume ordinary editing after the user has been queried for
    e.g. a file name.
  }
  begin
  if querymode then begin
    MoveToWindow(askerwin);
    Vw_QuitQuery;
    activeEtip:= askeract;
    querymode := false;
    SelectFunctions;
    end
  else begin
    Vw_Message('EndUserQuery: not in query mode, internal error!');
    Beep;
    end;
  end {EndUserQuery};


private
function GetConfirmation (prompt: vstring): boolean;
  { Ask the user for confirmation re something.
    Returns true is user said yes. Any other answer, or abort,
    returns false.
  }
  label
    1;
  var
    line  : Ed_LineType;   { a line of response from the user }
    length: integer;       { the length of it }
    msg   : Ed_MsgType;    { what terminated the input }
    ans   : vstring;
    ansyes, ansno: boolean;

  begin
  if Logfun then writeln(logfile,'GetConfirmation...');
  if BeginUserQuery then begin
    Ed_write(prompt);
    1: { Start over here in case of any fixable problems. }
      ansyes:= false;
      ansno := false;
      Ed_Getln (line, length, msg);
      if msg.kind = ProcessMsg then begin
        ans:= substr(line,1,length);
        if Logfun then writeln(logfile,'GetConfirmation "',ans,'"');
        ansyes:= ans = 'yes';
        ansno := ans = 'no';
        if not (ansyes or ansno) then begin
          Ed_Write('Please answer yes or no.');
          RestartQuery(length+1);
          goto 1;
          end;
        end {process msg};

    { Either there is a file name, or the query was aborted. }
    EndUserQuery;
    
    GetConfirmation:= ansyes;
    end {querymode}
  else
    GetConfirmation:= false;
  end {GetConfirmation};


private
function GetFileName (prompt: vstring; var fname: vstring): boolean;
  { Ask the user for a file name.
    Returns true, unless the file name query is aborted.
  }
  label
    1;
  var
    line  : Ed_LineType;   { a line of response from the user }
    length: integer;       { the length of it }
    msg   : Ed_MsgType;    { what terminated the input }
    fnok  : boolean;

  begin
  if Logfun then writeln(logfile,'GetFileName...');
  if BeginUserQuery then begin
    Ed_write(prompt);
    1: { Start over here in case of any fixable problems. }
      fnok:= false;
      Ed_Getln (line, length, msg);
      if msg.kind = ProcessMsg then begin
        if length < 1 then begin
          Ed_Write('Please supply a non-blank file name.');
          RestartQuery(1);
          goto 1;
          end;
        fname:= substr(line,1,length);
        if Logfun then writeln(logfile,'GetFileName "',fname,'"');
        fnok:= true;
        end {process msg};

    { Either there is a file name, or the query was aborted. }
    EndUserQuery;
    
    GetFileName:= fnok;
    end {querymode}
  else
    GetFileName:= false;
  end {GetFileName};


private
function GetFileNameOpen (    prompt: vstring;
                          var fname : vstring;
                          var ff    : text   ): boolean;
  { Ask the user for a file name, open the file for reading.
    File should exist, otherwise user must fix or abort.
    Returns true, unless the file name query is aborted.
  }
  label
    1;
  var
    line  : Ed_LineType;   { a line of response from the user }
    length: integer;       { the length of it }
    msg   : Ed_MsgType;    { what terminated the input }
    fnok  : boolean;

  begin
  if Logfun then writeln(logfile,'GetFileNameOpen...');
  if BeginUserQuery then begin
    Ed_write(prompt);
    1: { Start over here in case of any fixable problems. }
      fnok:= false;
      Ed_Getln (line, length, msg);
      if msg.kind = ProcessMsg then begin
        if length < 1 then begin
          Ed_Write('Please supply a non-blank file name.');
          RestartQuery(1);
          goto 1;
          end;
        fname:= substr(line,1,length);
        if Logfun then writeln(logfile,'GetFileNameOpen "',fname,'"');
        fnok:= OpenTextFile(fname, ff);
        if not fnok then begin
          Ed_Write(fname + ': Could not open.');
          RestartQuery(length+1);
          goto 1;
          end;
        end {process msg};

    { Either there is a file name, or the query was aborted. }
    EndUserQuery;
    
    GetFileNameOpen:= fnok;
    end {querymode}
  else
    GetFileNameOpen:= false;
  end {GetFileNameOpen};


private
procedure NameTipFile;
  { Transcript gets new file name. }
  var fname: vstring;
  begin
  if GetFileName('New file name: ', fname) then
    Vw_NameFile (curwin, fname, {override} true);
  end {NameTipFile};


private
function ReadInFile (var ff: text; fname: vstring): boolean;
  { Read in a file here. The file should already be open, with
    the name or path fname. Return success/failure.
  }
  var
    prePos: integer;
    ok    : boolean;
  begin
  with edwin[curwin], refEtip^ do prePos:=Ts_Pos(cursor,theTranscript);
  Vw_ReadFile (curwin, ff, fname, ok);
  with edwin[curwin], refEtip^ do
    if prePos <> Ts_Pos(cursor,theTranscript) then
      Changed(prePos+1);
  ReadInFile:= ok;
  end {ReadInFile};


private
procedure ReadOrNew (fname: vstring);
  { Read in a file with the given name, if it exists, or otherwise
    note that we have a new file and store the name.
  }
  var ff: text;
  begin
  if OpenTextFile(fname, ff) then begin
    if ReadInFile (ff, fname) then begin
      Vw_NameFile (curwin, fname, {override} true);
      Vw_Original (curwin);
      end
    end
  else begin
    Vw_NameFile (curwin, fname, {override} true);
    Vw_Message  ('(New file)');
    Vw_Original (curwin);
    end;
  end {ReadOrNew};


private
procedure NextEtip;
  { View the next transcript in this window.
    If there are no more, read in the next file given on the
    command line; if there are no more of those, wrap around to
    the first transcript.
  }
  var
    nexte: EtipRef;
    fname: vstring;
    valid: boolean;
  begin
  nexte:= Vw_NextTranscript (edwin[curwin].refEtip);
  if nexte <> nil then
    Vw_View (curwin, nexte)
  else begin
    Arg_NextFile (fname, valid);
    if valid then begin
      if Vw_NewTranscript(curwin, {modify=>} true) then
        ReadOrNew (fname)
      end
    else
      Vw_View(curwin, Vw_FirstTranscript);
    end;
  SelectFunctions;
  end {NextEtip}; 


private
procedure NewEtip;
  { Create a new transcript, perhaps from an existing file.
    Show it in the current window.
    Note: at present this does not check if the file is already
          being edited in another (or the same) transcript.
  }
  var fname: vstring;
  begin
  if GetFileName('Find file: ', fname) then
    if Vw_NewTranscript(curwin, {modify} true) then begin
      ReadOrNew(fname);
      SelectFunctions;
      end;
  end {NewEtip};


private
procedure ViewEtip;
  { Create a new transcript to view an existing file.
    Show it in the current window.
    Note: at present this does not check if the file is already
          being edited in another (or the same) transcript.
  }
  var fname: vstring;
      ff   : text;
  begin
  if GetFileNameOpen('View file: ', fname, ff) then
    if Vw_NewTranscript(curwin, {modify} true) then begin
      if ReadInFile (ff, fname) then
        Vw_NameFile (curwin, fname, {override} true);
      Vw_Modifiable (curwin, false);
      Vw_Original   (curwin);
      Vw_WriteTitle (curwin);
      SelectFunctions;
      end;
  end {ViewEtip};


private
procedure InsertFile;
  { Ask user for a file name, then read the file (text or transcript)
    and insert it at the cursor.
  }
  var
    fname : vstring;
    ff    : text;
  begin
  if GetFileNameOpen('Insert file: ', fname, ff) then begin
    { Note that the cursor position may have changed, if the user
      navigated in the original window during the file name query. }
    if ReadInFile (ff, fname) then ;  { transcript name NOT changed }
    end;
  end {InsertFile};


private
procedure StoreEtip (opt: Ts_fileOpt; var done: boolean);
  { Writes the current transcript to file, following the given options.
    For a partial storage, always asks for a new file name; if the
    user aborts, returns "false" in the done parameter.
  }
  var
    ok   : boolean;    { filename is at hand. }
    conf : boolean;    { need to confirm file overwrite. }
    fname: vstring;
    ff   : text;
    lines: integer;
    msg  : vstring;
  begin
  with edwin[curwin], refEtip^ do begin

    done:= false;  { unless we succeed below }

    { Get the filename and decide whether overwrite needs confirmation. }
    if opt = StoreAll then begin
      { If file name is not yet defined, ask for it. }
      ok:= filename <> '';
      if ok then fname:= filename
      else ok:= GetFileName ('File name for transcript: ', fname);
      conf:= fname <> origname;  { confirm if name changed }
      end
    else begin
      { Partial store. Always ask for a new file name. }
      ok  := GetFileName ('Name of file to write: ', fname);
      conf:= true;
      end;

    if not ok then
      { abandon }
    else begin

      { Create or rewrite the file. }
      ok:= NewTextFile (fname, ff);

      if not ok then begin
        if conf then begin
          Beep;
          ok:= GetConfirmation ('The file "'+fname+
                                '" exists - overwrite it? ');
          end
        else
          ok:= true;   { no need to confirm. }
        if ok then begin
          ok:= WriteTextFile (fname, ff);
          if not ok then begin
            Beep;
            Vw_Message('Could not write "'+fname+'"');
            end {cannot create};
          end {confirm overwrite};
        end {not new};

      if ok then begin

        { Finally. Just do it. }
        Ts_Store(theTranscript,ff,opt,lines,tsresult);
        close(ff);
        if tsresult <> Ts_ok then
          Vw_TsProblem(tsresult)
        else begin
          Vw_DecimalString(lines,1,msg);
          Vw_Message('[Wrote '+msg+' lines to "'+fname+'"]');
          if opt = storeAll then begin
            if filename = '' then
              Vw_NameFile(curwin, fname, {override} false);
            Vw_Original(curwin);
            end;
          done:= true;
          end {ts stored};
        end {file created or overwritten};
      end {filename supplied};
    end {with};
  end {StoreEtip};


private
function StoreAndExit: boolean;
  { Store (save)  all transcripts and exit editor.
    If some (modified and unsaved) transcript lacks a file name,
    prompt user for it. User may cancel. in which case we return
    "false".
  }
  var
    etr : EtipRef;    { Look at each transcript in turn. }
    done: boolean;    { While user allows storing them. }
  begin
  etr := Vw_FirstTranscript;
  done:= true;
  
  while done and (etr <> nil) do begin
    Vw_View (curwin, etr);
    if Vw_Modified (curwin) then
      StoreEtip(StoreAll, done);
    if done then etr:= Vw_NextTranscript(etr);
    end;

  StoreAndExit:= done;
  end {StoreAndExit};


private
function AbandonAndExit: boolean;
  { Abandon changes to all transcripts and exit editor.
    If there are some (modified and unsaved) transcripts,
    user is prompted to confirm the action. User may cancel. in
    which case we return "false".
  }
  var
    etr : EtipRef;    { Look at each transcript in turn. }
    done: boolean;    { While user allows storing them. }
  begin
  etr := Vw_FirstTranscript;
  done:= true;
  
  while done and (etr <> nil) do begin
    Vw_View (curwin, etr);
    if Vw_Modified (curwin) then
      done:= GetConfirmation('Modified documents exist. Leave anyway? ');
    if done then etr:= Vw_NextTranscript(etr);
    end;

  AbandonAndExit:= done;
  end {AbandonAndExit};



{ --- script processing functions -- }

private
procedure NoJobYet;
  begin
  Vw_Message('Job mode unimplemented.');
  end {NoJobYet};


private
procedure Unprocessable;
  begin
  MakeNewLine;
  { Vw_Message('This text is not processable.'); }
  end {Unprocessable};

private
procedure PleaseFixError;
  begin
  Vw_Message('[Please correct the error first]')
  end {PleaseFixError};


private
procedure ExecFromBol;
  { Set startPos and lastPos to indicate that processing is to start
    from the beginning of the line holding the cursor.
    Also moves cursor to the start of the line.
    This routine is used in ManualExec mode. }
  begin
  if Log_includes(tracelog) then MessPause('ExecFromBol..');
  with edwin[curwin], refEtip^ do begin
    Ts_Move(cursor,0,-MaxCol,cursor,theTranscript);
    lastPos :=Ts_Pos(cursor,theTranscript);
    startPos:=lastPos+1;
    end;
  end {ExecFromBol};


private
procedure ExecToEol;
  { Set execPos to indicate that we want to process the script
    up to and including the line holding the cursor, newline
    also included. If there is no (input) newline after the cursor,
    make one. }
  var pos: integer;
  begin
  if Log_includes(tracelog) then MessPause('ExecToEol..');
  with edwin[curwin], refEtip^ do begin
    pos:=Ts_Pos(cursor,theTranscript);
    { We do not check against lastPos here, since we may have
      to cancel first, anyway, if ExecuteItem was used. }
    if pos >= errorPos then PleaseFixError
    else begin
      Ts_Skipln(cursor,theTranscript,tsresult);
      if tsresult = Ts_nonewline then MakeNewline;
      if tsresult = Ts_ok then begin
        pos:=Ts_Pos(cursor,theTranscript);
        if pos >= errorPos then PleaseFixError
        else begin
          execPos:=pos;
          if Log_includes(tracelog) then MessPause('ExecPos set..');
          end
        end
      else
        Vw_TsProblem(tsresult);
      end;
    end;
  end {ExecToEol};


private
procedure UnwritePastLast;
  { Unwrite output due to positions from lastPos+1 up and including
    at least Ts_Pos(cursor), and perhaps all the way to the writept, if
    the writept is still farther in the transcript.
    May move the writept, too.
    This routine is used in ManualExec mode, with the cursor after
    the input line that is to be processed. }
  var pos: integer;
  begin
  if Log_includes(tracelog) then MessPause('UnwritePastLast..');
  with edwin[curwin], refEtip^ do begin
    pos:=Ts_Pos(cursor,theTranscript);
    { The "seek" can be omitted, if the writept is already
      far enough - in this case, all output between cursor and
      writept would be unwritten. }
    if Ts_Pos(writept,theTranscript) > pos
    then tsresult:=Ts_ok
    else Ts_Seek( {from}     cursor,
                  {position} pos+2,   { makes pos(writept) = pos+1 > pos }
                  {setting}  writept,
                  theTranscript,tsresult);
    if (tsresult = Ts_ok)
    or (tsresult = Ts_endinput)
    then begin
      if Log_includes(tracelog) then MessPause('Unwriting..');
      if Log_includes(unwritelog) then
        writeln(logfile,'Ed.UnwritePastLast unwrite=',lastPos:1);
      Ts_Unwrite(writept,theTranscript,lastPos);
      if Log_includes(tracelog) then MessPause('..unwritten');
      end
    else Vw_TsProblem(tsresult);
    end;
  end {UnwritePastLast};


private
procedure ExecuteLine;
  { Set execPos to indicate that we want to process the script
    up to and including the line holding the cursor, newline
    also included. Corresponding actions for Man and Job modes. }
  begin
  if Logfun then writeln(logfile,'ExecuteLine');
  with edwin[curwin], refEtip^ do begin
    case modeStack[modeTop].mode of
      NoExec : Unprocessable;
      JobExec: NoJobYet;
      ManExec: begin
               ExecFromBol;
               ExecToEol;
               UnwritePastLast;
               end;
      IncExec: ExecToEol;
      end {case mode};
    end;
  end {ExecuteLine};


private
procedure ExecuteToHere;
  { Set execPos to indicate that we want to process the script
    up to and including the cursor. }
  var pos: integer;
  begin
  with edwin[curwin], refEtip^ do begin
    case modeStack[modeTop].mode of
      NoExec : Unprocessable;
      JobExec: NoJobYet;
      ManExec: ExecuteLine;
      IncExec: begin
               pos:=Ts_Pos(cursor,theTranscript);
               if pos < errorPos then execPos:=pos;
               { else message: please correct error first. }
               end;
      end {case mode};
    end;
  end {ExecuteToHere};


private
procedure ExecToEot (line: boolean);
  { Request processing up to end of transcript.
    The option 'line' selects whether processing is in units
    of lines or characters. }
  var pos: integer;
  begin
  with edwin[curwin], refEtip^ do begin
    if errorPos <> infiniteindex then PleaseFixError
    else begin
      Ts_GotoEnd(cursor,theTranscript);
      pos:=Ts_Pos(cursor,theTranscript);
      if line then begin
        if pos > 0 then Ts_Seek(cursor,pos,cursor,theTranscript,tsresult)
                   else tsresult:=Ts_ok;
        if tsresult = Ts_ok then ExecToEol
                            else Vw_TsProblem(tsresult)
        end
      else
        ExecuteToHere
      end;
    end;
  end {ExecToEot};


private
procedure ExecuteToEnd (line:boolean);
  { Set execPos to indicate that we want to process the script
    up to the end of the transcript.
    The option 'line' selects whether processing is in units of
    lines or characters.
    For ManExec, processing starts at the start of the current
    line. }
  begin
  with edwin[curwin], refEtip^ do begin
    case modeStack[modeTop].mode of
      NoExec : Unprocessable;
      JobExec: NoJobYet;
      ManExec: begin
               ExecFromBol;
               ExecToEot ( {line} true);
               Ts_GotoEnd(writept,theTranscript);
               if Log_includes(unwritelog) then
                 writeln(logfile,'Ed.ExecuteToEnd unwrite=',startPos-1:1);
               Ts_Unwrite(writept,theTranscript,startPos-1);
               end;
      IncExec: ExecToEot(line);
      end {case mode};
    end;
  end {ExecuteToEnd};


private
procedure ExecuteItem (line: boolean);
  { Pick up current word, append to preceding input line, execute.
    If the current state is 'error', the error token is replaced by
    the current work, and then executed.
    This rather complex routine implements menu selection.
    The word is defined as the longest string of non-blank
    characters following the cursor position and all of the same
    source. Blank characters after the cursor are ignored.
    The word ends in a blank, or a newline, or a non-text element,
    or an element from a source that differs from the first
    non-blank character in the word.
    The option 'line' selects whether processing is in units of
    lines or characters. }
  const
    maxWord = 40;   { max length of word }
  var
    c: char;
    eol: boolean;
    source: Ts_dataSource;
    wsrc  : Ts_dataSource;
    word  : vstring;  { string[maxWord] }
    fail,found: boolean;
    appos,errend : ScriptIndex;
    i     : integer;
  begin
  if Logfun then write(logfile,'ExecuteItem');
  with edwin[curwin], refEtip^ do begin

    appos:=Ts_Pos(cursor,theTranscript);

    { Skip blanks. }
    repeat
      Ts_CharAfter(cursor,theTranscript,c,eol,source,tsresult);
      fail := eol or (tsresult <> Ts_ok);
      found:= c <> ' ';
      if not (fail or found) then Ts_Move(cursor,0,1,cursor,theTranscript);
    until fail or found;
    word:=''; wsrc:=source;

    { Collect word. }
    repeat
      if not fail then begin
        if length(word) >= maxWord then fail:=true
        else begin
          word:=concat(word,c);
          Ts_Move(cursor,0,1,cursor,theTranscript);
          Ts_CharAfter(cursor,theTranscript,c,eol,source,tsresult);
          fail := tsresult <> Ts_ok;
          found:= (c = ' ') or (source <> wsrc) or eol;
          end;
        end;
    until fail or found;

    if fail then begin
      if tsresult <> Ts_ok then Vw_TsProblem(tsresult)
      else if word = ''    then Vw_Message('[No word]')
      else                      Vw_Message('[Word too long]')
      end {fail}
    else begin
      if Log_includes(tracelog) then MessPause('Got word..');
      if errorPos = infiniteindex then
        { No error, append to prev line. }
        Ts_Seek (cursor,appos,cursor,theTranscript,tsresult)
      else begin
        { Locate the error token. }
        errend:=errorFirst+errorLength;
        Ts_Seek(cursor,errend,cursor,theTranscript,tsresult);
        { Accept eot if the position is otherwise ok. }
        if tsresult = Ts_endinput then
          if Ts_Pos(cursor,theTranscript) = errend-1 then
            tsresult:=Ts_ok;
        { Delete the token. }
        i:=1;
        while (i <= errorLength)
          and (tsresult = Ts_ok)
        do begin
          DeleteChar({joinline} true);
          i:=i+1;
          end;
        end;
      { Put word in text. }
      i:=1;
      while (i <= length(word))
        and (tsresult = Ts_ok)
      do begin
        Typing(word[i]);
        i:=i+1;
        end;
      if Log_includes(tracelog) then MessPause('Word inserted..');
      if tsresult = Ts_ok then begin
        if line then ExecuteLine
                else ExecuteToHere;
        end;
      end {found};
    end {with};
  if tsresult <> Ts_ok then Vw_TsProblem(tsresult);
  end {ExecuteItem};


private
procedure ExecuteSummat (line: boolean);
  { Execute something, depending on context.
    The option 'line' selects whether processing is in units of
    lines or characters.}
  begin
  with edwin[curwin], refEtip^ do
    if      Ts_AtOutput(cursor,theTranscript) then ExecuteItem(line)
    else if line                              then ExecuteLine
                                              else ExecuteToHere;
  end {ExecuteSummat};


private
procedure AbortProcess;
  { The (sub)session for this transcript will be aborted. }
  begin
  with edwin[curwin], refEtip^ do
    aborting:= true;
  end {AbortProcess};


procedure Ed_Used (pos: integer);
  begin
  with edwin[curwin].refEtip^ do begin
    if Log_includes(usedlog) then
      writeln(logfile,'Ed_Used(',pos:1,'). usedPos=',usedPos:1,
               ' lastPos=',lastPos:1);
    if pos > lastPos then MessPause('Used > last!')
    else
    if  (usedPos <= lastPos)
    and (pos     <  usedPos) then
      MessPause('Used-pos decrease!')
    else
      usedPos:=pos;
    end;
  end {Ed_Used};


private
procedure SpotCheck;
  { Check transcript elements and spots for correct row, col etc.
    Testing only. }
  var spotok: boolean;
  begin
  Log_Open;
  Ts_Check(logfile,edwin[curwin].refEtip^.theTranscript,spotok);
  if not spotok then MessPause('Spot check failed - see log.');
  end {SpotCheck};


private
procedure DoEdFun (theFun   : EditFun;
                   textChar : char;
                   line     : boolean;
                   var quits: boolean);
  { Perform an editing function.
    The option 'line' selects whether processing is in units of
    lines or characters. }

  var
    dummy_done: boolean;

  procedure Unimp;
    begin
    Beep;
    if not querymode then
      Vw_Message('Unimplemented function.');
    end {Unimp};

  procedure Disabled;
    begin
    Beep;
    if not querymode then
      Vw_Message('Function not allowed in this window/document');
    end {Disabled};

  begin
  if tracing then MessPause('DoEdFun..');
  case theFun of
    Ef_undefined   :                       Unimp;
    Ef_insmode     :                       Unimp;
    Ef_overmode    :                       Unimp;
    Ef_text        : Typing(textChar);
    Ef_newline     : MakeNewline;
    Ef_delchar     : DeleteChar(false);
    Ef_delsymb     : DeleteChar(true);
    Ef_backspace   : BackSpace;
    Ef_joinline    : Joinline;
    Ef_left        : MoveCursor( 0,-1);
    Ef_right       : MoveCursor( 0,+1);
    Ef_bakw        :                       Unimp;
    Ef_forw        :                       Unimp;
    Ef_endleft     : MoveCursor( 0,-MaxCol);
    Ef_endright    : MoveCursor( 0,+MaxCol);
    Ef_up          : MoveCursor(-1, 0);
    Ef_down        : MoveCursor(+1, 0);
    Ef_pageup      : with edwin[curwin].theWindow do MoveCursor(-rows, 0);
    Ef_pagedown    : with edwin[curwin].theWindow do MoveCursor(+rows, 0);
    Ef_goTop       : GoToExtremes(0);
    Ef_goBot       : GoToExtremes(maxint);
    Ef_search      :                       Unimp;
    Ef_replace     :                       Unimp;
    Ef_mark        : MarkThisPoint;
    Ef_cut         : CutRegion;
    Ef_copy        : CopyRegion ({asinput=>} true);
    Ef_paste       : Paste      ({asinput=>} true);
    Ef_openwin     : OpenWin;
    Ef_closewin    : CloseOtherWins;
    Ef_nextwin     : NextWin;
    Ef_newtip      : NewEtip;
    Ef_viewtip     : ViewEtip;
    Ef_nexttip     : NextEtip;
    Ef_refresh     : Vw_RefreshScreen;
    Ef_write       : if Vw_Modified (curwin) then
                        StoreEtip (StoreAll, dummy_done)
                     else
                        Vw_Message ('No changes since last save.');
    Ef_writeInput  : StoreEtip(StoreInput  , dummy_done);
    Ef_writeText   : StoreEtip(StoreAsText , dummy_done);
    Ef_readIn      : InsertFile;
    Ef_namefile    : NameTipFile;
    Ef_process     : ExecuteSummat(line);
    Ef_procline    : ExecuteLine;
    Ef_procend     : ExecuteToEnd (line);
    Ef_procitem    : ExecuteItem  (line);
    Ef_abort       : AbortProcess;
    Ef_complete    :                        Unimp;
    Ef_query       :                        Unimp;
    Ef_exit        : quits:= StoreAndExit;
    Ef_quit        : quits:= AbandonAndExit;
    Ef_logmore     : Log_level(+1);
    Ef_logless     : Log_level(-1);
    Ef_logwindow   : begin
                     Log_Open;
                     Wd_ShowWin (logfile,
                                 edwin[curwin].theWindow);
                     end;
    Ef_logscript   : begin
                     Log_Open;
                     Ts_Dump    (logfile,
                                 edwin[curwin].refEtip^.theTranscript);
                     end;
    Ef_spotcheck   : SpotCheck;
    Ef_whereline   : begin
                     whereline:= not whereline;
                     if not whereline then NoWhere;
                     end;
    Ef_ambiguous   :                       Unimp;
    Ef_disabled    : Disabled;
    otherwise                              Unimp;
    end {case editing function};
  tracing:=Log_includes(tracelog);
  end {DoEdFun};


private
procedure CleanOtherWindows;
  { Clean all but current window.
    May be interrupted by e.g. keypress. }
  var
    numwin, w: integer;
    clean    : boolean;
  begin
  w:=1; clean:=true; numwin:= Vw_NumWindows;
  while (w <= numwin)   { something to do }
    and clean           { cleaning not interrupted by e.g. keypress }
  do begin
    if w <> curwin then Vw_CleanWindow(w,clean);
    w:=w+1;
    end;
  end {CleanOtherWindows};


private
function Primed (var {i} etip: EditTranscript): boolean;
  { Whether the transcript is ready to be processed or subjected to
    a cancel action or abortion. }
  begin
  with etip do
    case modeStack[modeTop].mode of
      NoExec : Primed:=false;
      JobExec: Primed:=false;                      { NoJobYet }
      ManExec: Primed:=(   (execPos >  lastPos)    { processing needed }
                        or aborting             );
      IncExec: Primed:=(   (execPos >  lastPos)    {    "         "    }
                        or (editPos <= lastPos)    { cancel       "    }
                        or aborting             );
      end;
  end {Primed};


private
procedure GoCursor;
  { Go to cursor for current window. }
  begin
  Vw_GoCursor(curwin);
  end {GoCursor};


private
procedure Editing (line: boolean);
  { Edit transcript until it is primed for processing or cancelling.
    The option 'line' selects whether processing is in units of lines
    or characters. }
  var
    fd      : integer;  { data available on this fd }
    theFun  : EditFun;
    textChar: char;
    timeout : boolean;
    quits   : boolean;
  begin
  quits:=false;
  if tracing then MessPause('Editing...');
  with edwin[curwin].refEtip^ do begin
    execPos := 0;  {** Testing **}
    aborting:= false;
    end;

  GoCursor;   { Clean up current window to display recent output. }

  {    M a i n   l o o p   }
  while not (quits or Primed(edwin[curwin].refEtip^))
  do begin
    if whereline then Where;
    if Log_Includes(spotchecklog) then SpotCheck;
    { if tracing then MessPause('WaitForData..'); }    { conflicting i/o ? }

    fd:=WaitForData(300); { any data now? }
    if fd < 0 then begin  { timeout }
      CleanOtherWindows;
      GoCursor;
      fd:=WaitForData(0); { wait with infinite patience }
      end;

    if fd = 0 then begin  { keyboard input }
      Km_GetFun(theFun,textChar,1,timeout);
      if not timeout then begin
        if tracing then MessPause('..got EdFun.');
        Vw_ClearMess;
        tsresult:=Ts_ok;
        DoEdFun (curmap[theFun],textChar,line,quits);
        GoCursor;
        end
      end { key struck }
    else
    if fd > 0 then WriteAndCheck(fd);   { write-channel data }
    end {while};

  if tracing then MessPause('..end edit');
  if quits then HaltProgram;
  end {Editing};


private
procedure Process (var msg: Ed_MsgType; one: boolean);
  { Send script symbol startPos for processing.
    If 'one', just one byte (Ts_Read), else as many
    as the transcript offers (Ts_Reads).
    May fail with some tsresult code, in which case
    msg.kind = NoMsg. }
  var ch           : char;
      preRow,preCol: integer;
      readPos      : ScriptIndex;
  begin
  msg.kind:= ProcessMsg;  { unless there is a problem below }
  with edwin[curwin], refEtip^ do begin
    if tracing then MessPause('Seeking');
    Ts_Seek(readpt,startPos,readpt,theTranscript,tsresult);
    if tsresult <> Ts_ok then msg.kind:= NoMsg
    else begin
      if tracing then MessPause('Reading');

      msg.eoln:=Ts_Eoln(readpt,theTranscript);
      msg.eof :=Ts_Eof (readpt,theTranscript);

      if one then begin
        Ts_Read (readpt,theTranscript, ch,       readPos,tsresult);
        msg.data:=ch;
        end
      else
        Ts_Reads(readpt,theTranscript, msg.data, readPos,tsresult);

      if tsresult <> Ts_ok then msg.kind:= NoMsg
      else begin
        if Ts_Order(writept,readpt,theTranscript) < 0 then
          Ts_Move(readpt,0,0,writept,theTranscript);
        if tracing then MessPause('Dewriting');
        Ts_Dewrite(writept,theTranscript,preRow,preCol,tsresult);
        Vw_DelShot(writept,refEtip,preRow,preCol);
        if tsresult <> Ts_ok then msg.kind:= NoMsg
        else begin
          msg.processfrom:=startPos;
          if tracing then MessPause('Process "' + msg.data + '"');
          editPos  := infiniteindex;
          lastPos  := startPos;
          usedPos  := infiniteindex;
          skipwrite:= false;
          startPos := startPos+length(msg.data);
          if Log_Includes(readlog) then begin
            write(logfile,'Process');
            if msg.eoln then write(logfile,' eoln');
            if msg.eof  then write(logfile,' eof');
            writeln(logfile,' "',msg.data,'"');
            end;
          end;
        end;
      end;
    end;
  end {Process};


private
procedure Cancel (var msg: Ed_MsgType);
  { Send a cancel message.
    For Man and Job, take the equivalent action. }
  begin
  with edwin[curwin], refEtip^ do begin
    if tracing then MessPause('Cancel');
    case modeStack[modeTop].mode of
      NoExec,
      JobExec,
      ManExec: begin
               msg.kind := NoMsg;
               lastPos  := editPos-1;
               end;
      IncExec: begin
               msg.kind:= CancelMsg;
               msg.cancelto:= editPos;
               if Log_Includes(readlog) then
                 writeln(logfile,'Cancel to ',msg.cancelto:1);
               end;
      end {case mode};
  { execPos  := 0; }    {** removed for test **}
    usedPos  :=infiniteindex;
    skipwrite:=true;  { until something is read again }
    end;
  end {Cancel};


private
procedure Aborts (var msg: Ed_MsgType);
  { Send an "abort" message, if in Incremental mode. }
  begin
  with edwin[curwin], refEtip^ do begin
    if tracing then MessPause('Aborts');
    case modeStack[modeTop].mode of
      NoExec,
      JobExec,
      ManExec: begin
               Vw_Message('["Abort" does not apply to this mode]');
               msg.kind:= NoMsg;
               end;;
      IncExec: begin
               msg.kind:= AbortMsg;
               if Log_Includes(readlog) then
                 writeln(logfile,'Abort processor.');
               end;
      end {case mode};
    usedPos  :=infiniteindex;
    skipwrite:=true;  { until something is read again }
    end;
  end {Aborts};


procedure Ed_Get (var msg: Ed_MsgType);
  { Get next input from Editor. }
  begin
  with edwin[curwin], refEtip^ do begin
    msg.kind:= NoMsg;
    repeat
      tsresult:=Ts_ok;
      if (editPos <= lastPos) then Cancel(msg);
      if (msg.kind = NoMsg) then
        if (lastPos  <  execPos)
        or (startPos <= lastPos)
        then
          if (startPos < errorPos) then Process(msg,{one=>}true);
      if msg.kind = NoMsg then begin
        { Nothing to process or nothing to cancel or some problem. }
        Vw_TsProblem(tsresult);
        Editing ({line=>} false);
        end;
    until msg.kind <> NoMsg;
    end;
  end {Ed_Get};


procedure Ed_Getln {var line  : Ed_LineType;
                    var len   : integer;
                    var msg   : Ed_MsgType };

  procedure ProcLine;
    { Read to end of line. May fail (tsresult not ok) or may
      report that line is too long (sets errorPos).
      In case of failure, msg.kind = NoMsg.
      This implementation is inefficient. It could be streamlined
      both by using a special form of 'Process' or (better) by
      using a Ts_Readln routine. However, this would mean putting
      some of the output-handling policies in the transcripts
      module. }
    var endl,oflo: boolean;
        starts   : ScriptIndex;  { Pos of start of text in line }
        newlen: integer;
        i     : integer;
        excess: integer;
    begin
    len :=0;
    oflo:=false;
    repeat
      Process(msg,{one=>}false);
      if msg.kind = ProcessMsg then begin
        if len=0 then starts:=msg.processfrom;
        endl:=msg.eoln or msg.eof;
        if not endl then begin
          newlen:=len + length(msg.data);
          oflo:= newlen > Ed_maxLine;
          if not oflo then
            for i:=1 to length(msg.data) do
              line[len+i]:=msg.data[i];
          len:=newlen;
          end;
        end;
    until endl or (msg.kind <> ProcessMsg);

    if msg.kind = ProcessMsg then begin
      if oflo then begin
        Ed_Write('The above line is too long...');
        excess:=len - Ed_maxLine;
        with edwin[curwin].refEtip^ do
          Ed_Tell({last  } lastPos,
                  {error } startPos-excess,
                  {start } startPos,
                  {errfst} startPos-excess, {errlen} excess);
        end;
      if oflo then msg.kind:= NoMsg
              else msg.processfrom:= starts;
      end;
    end {ProcLine};

  begin
  with edwin[curwin], refEtip^ do begin
    { Write pending output, if any, from previous line(s). }
    while DataAvailable(writefd) do WriteAndCheck(writefd);
    msg.kind:= NoMsg;
    repeat
      len :=0;
      tsresult:=Ts_ok;

      if aborting then begin
        Aborts(msg);
        aborting:= false;
        end
      else
      if (editPos <= lastPos) then
        Cancel(msg);

      if msg.kind = NoMsg then
        if (lastPos  <  execPos)
      { or (startPos <= lastPos) }
        then
          if (startPos < errorPos) then begin
            ProcLine;
            if msg.kind = ProcessMsg then begin
              { Move cursor to start of next input line. }
              Ts_Skip(cursor,theTranscript,tsresult);
              if (tsresult <> Ts_ok) and (tsresult <> Ts_endinput)
              then msg.kind:= NoMsg;
              end;
            end;
      if msg.kind = NoMsg then begin
        { Nothing to process or nothing to cancel or some problem. }
        Vw_TsProblem(tsresult);
        Editing ({line=>} true);
        end;
    until msg.kind <> NoMsg;
    end;

  end {Ed_Getln};


function Ed_InfiniteIndex: ScriptIndex;
  begin
  Ed_InfiniteIndex:= infiniteindex;
  end {Ed_InfiniteIndex};


function Ed_BeginMode (newmode: Ed_ExecMode; newpos: ScriptIndex): boolean;
  var omode: Ed_ExecMode;
      opos : ScriptIndex;
  begin
  with edwin[curwin], refEtip^ do begin
    omode:=modeStack[modeTop].mode;
    opos :=modeStack[modeTop].startPos;
    if opos > newpos then begin
      Vw_Message('Overlapping exec mode, rejected');
      Ed_BeginMode:=false;
      end
    else if modeTop >= MaxExecs then begin
      Vw_Message('Too many nested exec modes.');
      Ed_BeginMode:=false;
      end
    else begin
      if omode = newmode then Vw_Message('Duplicate exec mode - warning only');
      modeTop:=modeTop+1;
      with modeStack[modeTop] do begin
        mode    := newmode;
        startPos:= newpos;
        end;
      Vw_WriteTitle(curwin);
      Ed_BeginMode:=true;
      end;
    end;
  end {Ed_BeginMode};


function Ed_EndMode: boolean;
  begin
  with edwin[curwin], refEtip^ do begin
    if modeTop <= 1 then begin
      Vw_Message('Bottom exec mode should never end.');
      Ed_EndMode:=false;
      end
    else begin
      case modeStack[modeTop].mode of
        NoExec : ;
        JobExec: ;
        ManExec: ;
        IncExec: begin
                 editPos:=infiniteindex;  { quit cancelling }
                 skipwrite:=false;
                 end;
        end;
      modeTop:=modeTop-1;
      Vw_WriteTitle(curwin);
      end;
    end;
  end {Ed_EndMode};


function getenv (name: univ_ptr): univ_ptr;    { From UNIX }
  external C;


private
function LookForResourceFile (name: vstring; var f: text): boolean;
  { Tries to open the file for reading, looking first in working dir,
    then in $HOME and finally in /usr/local/src/seed/lib.
  }
  var
    pathname: vstring;  { Hope it's long enough. }
    done    : boolean;
    homeptr : univ_ptr;
    homedir : vstring;
  begin
  done:= OpenTextFile(name, f);
  if not done then begin
    homeptr:= getenv(addr('HOME'));
    if homeptr <> nil then begin
      Cstring(homeptr, homedir);
      pathname:= homedir + '/' + name;
      done:= OpenTextFile(pathname, f);
      end;
    if not done then begin
      pathname:= '/usr/local/src/seed/lib/' + name;
      done:= OpenTextFile(pathname, f);
      end;
    end;
  LookForResourceFile:= done;
  end {LookForResourceFile};


private
function KeyFunDef: boolean;
  { Initialise Keymap module by giving each editing function
    a name and reading a keystroke map file.
    Returns 'false' in case of fatal failure. }
  var
    result  : Km_result;
    undef,ambig: boolean;
    mapf    : text;

  procedure Def (theFun: EditFun; name: vstring);
    begin
    Km_NameFun(theFun,name,result);
    if result <> Km_Ok then begin
      write(errout,'"',name,'" ');
      KmProb(result);
      writeln(errout);
      end;
    end {Def};

  function QuitDef: boolean;
    { Define the default 'quit' function. }
    var quitk   : vstring;
    begin
    quitk:='xx';
    quitk[1]:=chr(ord('X')-64);  { ^X }
    quitk[2]:=chr(ord('C')-64);  { ^C }
    Km_Map(Ef_quit,quitk,result);
    if result <> Km_ok then begin
      write(errout,'Could not define default "quit" code: ');
      KmProb(result);
      writeln(errout);
      end;
    QuitDef:= result = Km_ok;
    end {QuitDef};
    
  begin
  Def(Ef_undefined,'undefined' );
  Def(Ef_insmode  ,'insert'    );
  Def(Ef_overmode ,'overstrike');
  Def(Ef_text     ,'text'      );
  Def(Ef_newline  ,'newline'   );
  Def(Ef_delchar  ,'delchar'   );
  Def(Ef_delsymb  ,'delsymb'   );
  Def(Ef_backspace,'backspace' );
  Def(Ef_joinline ,'joinline'  );
  Def(Ef_left     ,'left'      );
  Def(Ef_right    ,'right'     );
  Def(Ef_endleft  ,'endleft'   );
  Def(Ef_endright ,'endright'  );
  Def(Ef_bakw     ,'bakw'      );
  Def(Ef_forw     ,'forw'      );
  Def(Ef_up       ,'up'        );
  Def(Ef_down     ,'down'      );
  Def(Ef_pageup   ,'pageup'    );
  Def(Ef_pagedown ,'pagedown'  );
  Def(Ef_goTop    ,'go-top'    );
  Def(Ef_goBot    ,'go-bot'    );
  Def(Ef_search   ,'search'    );
  Def(Ef_replace  ,'replace'   );
  Def(Ef_mark     ,'mark'      );
  Def(Ef_cut      ,'cut'       );
  Def(Ef_copy     ,'copy'      );
  Def(Ef_paste    ,'paste'     );
  Def(Ef_openwin  ,'open-window'    );
  Def(Ef_closewin ,'close-windows'  );
  Def(Ef_nextwin  ,'next-window'    );
  Def(Ef_newtip   ,'new-transcript' );
  Def(Ef_viewtip  ,'view-transcript');
  Def(Ef_nexttip  ,'next-transcript');
  Def(Ef_refresh  ,'refresh-screen' );
  Def(Ef_write     ,'write-transcript');
  Def(Ef_writeInput,'write-input'     );
  Def(Ef_writeText ,'write-text'      );
  Def(Ef_readIn    ,'read-insert'     );
  Def(Ef_namefile  ,'name-file'       );
  Def(Ef_process  ,'process-char'   );
  Def(Ef_procline ,'process-line'   );
  Def(Ef_procend  ,'process-end'    );
  Def(Ef_procitem ,'process-item'   );
  Def(Ef_abort    ,'abort-process'  );
  Def(Ef_complete ,'complete-input' );
  Def(Ef_query    ,'query-completions');
  Def(Ef_exit     ,'exit'      );
  Def(Ef_quit     ,'quit'      );
  Def(Ef_logmore  ,'logmore'   );
  Def(Ef_logless  ,'logless'   );
  Def(Ef_logwindow,'logwindow' );
  Def(Ef_logscript,'logscript' );
  Def(Ef_spotcheck,'spotcheck' );
  Def(Ef_whereline,'whereline' );
  Def(Ef_disabled ,'disabled'  );
  Def(Ef_ambiguous,'ambiguous' );

  if tracing then MessPause('Ed.func. names defined.');

  { Km_ShowNames; }

  if not LookForResourceFile ('.seed-keymap', mapf) then begin
    writeln(errout,'Could not find the keymap (.seed-keymap). Quitting.');
    KeyFunDef:=QuitDef;
    end
  else begin
    Km_ReadMap(mapf,undef,ambig,result);
    close(mapf);
    if result <> Km_Ok then begin
      KmProb(result);
      if ambig then write(errout,' (ambiguous)');
      if undef then write(errout,' (undefined)');
      KeyFunDef:=QuitDef;
      writeln(errout);
      MessPause('Press space bar to continue...');
      end
    else
      KeyFunDef:=true;
    if tracing then MessPause('Keymap file read.');
    end;

  end {KeyFunDef};




procedure Ed_Init;
  { Edit initialisation. }
  var
    fname    : vstring;   { First file name on command line }
    validname: boolean;   { if there is one. }

  begin

  Vw_Init;

  Vw_StatusPlace (0,  1, rcPlace);
  Vw_StatusPlace (0, 31, posPlace);

  infiniteindex:= Vw_InfiniteIndex;

  autoshow := false;
  tracing  := false;
  whereline:= false;

  allEditFuns:= [ Ef_undefined .. Ef_ambiguous ];
  debugFuns  := [ Ef_logmore   .. Ef_whereline ];
  mutatorFuns:= [ Ef_text      .. Ef_joinline  ]
              + [ Ef_replace, Ef_cut, Ef_paste,
                  Ef_readIn,
                  Ef_procitem ];
  processFuns:= [ Ef_process .. Ef_query ];
  queryFuns  := [ Ef_search, Ef_replace,
                  Ef_readIn, Ef_namefile,
                  Ef_quit ];

  debugmode:= Arg_Option('-debug');

  querymode:= false;  { askerwin and querywin consequently undefined. }

  { Clip-buffer transcript. }
  Ts_Create(clipbuf,tsresult);
  if tsresult = Ts_ok then Ts_NewSpot(clipeot,clipbuf,tsresult);
  Vw_TsProblem(tsresult);

  curwin:= 1;
  activeEtip:= edwin[curwin].refEtip;
  MapFunctions(curwin, curmap);
  Vw_Title (curwin, BoldTitle);

  if tracing then MessPause('Main window created.');

  if not KeyFunDef then begin
    MessPause('Keymap initialisation failed. Quitting.');
    HaltProgram;
    end;

  { Read in first file on command line, if any. }
  Arg_FirstFile;
  Arg_NextFile (fname, validname);
  if validname then ReadOrNew (fname);

  end {Ed_Init};
