{ editfuns.p
  Editor functions enumerated.

  Changes:
    94.05.04  Ef_namefile, Ef_goTop, Ef_goBot added.
    94.07.27  Ef_abort, Ef_complete, Ef_query added.
    94.09.28  Ef_viewtip added.
    94.10.07  Ef_disabled added.
}

type
  EditFun =      {  A thing the editor can do. }
      (Ef_undefined,   { undefined function (bottom)                 }

                  { TYPING MODES }

       Ef_insmode,     { use text insertion mode                     }
       Ef_overmode,    { use text overstrike mode                    }

               { TYPING and BACKSPACING }

       Ef_text,        { enter a text character                      }
       Ef_newline,     { insert a line break at the cursor           }
       Ef_delchar,     { delete character to the left of the cursor  }
       Ef_delsymb,     { del char; if newline, join lines            }
       Ef_backspace,   { backspace cursor:                           }
                       {   insert    : delchar                       }
                       {   overstrike: left, text space, left        }
       Ef_joinline,    { join cursor line to preceding line (append) }

                   { NAVIGATION }

       Ef_left,        { cursor left one column (no wrap)            }
       Ef_right,       { cursor right one column (no wrap)           }
       Ef_bakw,        { cursor backward one pos (wrap to prev line) }
       Ef_forw,        { cursor forward one pos (wrap to next line)  }
       Ef_endleft,     { cursor left to beginning of line            }
       Ef_endright,    { cursor right to end of line                 }
       Ef_up,          { cursor up one line (scroll if needed)       }
       Ef_down,        { cursor down one line (scroll if needed)     }
       Ef_pageup,      { scroll backward one screenful               }
       Ef_pagedown,    { scroll forward  one screenful               }
       Ef_goTop,       { go to top (beginning) of file               }
       Ef_goBot,       { go to bottom (end) of file                  }

                 { SEARCH and REPLACE }

       Ef_search,      { search forward for a text string            }
       Ef_replace,     { search for a text string and change it      }

                { CUT, COPY and PASTE }

       Ef_mark,        { set mark at cursor position                 }
       Ef_cut,         { cut text from mark to cursor into buffer    }
       Ef_copy,        { copy text from mark to cursor into buffer   }
       Ef_paste,       { paste buffered text inserting before cursor }

              { WINDOWS and TRANSCRIPTS }

       Ef_openwin,     { Open new window by splitting current window }
       Ef_closewin,    { Close all except current window             }
       Ef_nextwin,     { Go to next window below (wrap at bottom)    }
       Ef_newtip,      { Make a new transcript, perhaps from file    }
       Ef_viewtip,     { View a file in a buffer (not modifiable)    } 
       Ef_nexttip,     { Go to next transcript (text buffer)         }
       Ef_refresh,     { Refresh screen completely                   }

                    { FILES }

       Ef_write,       { Store (write) transcript in file.           }
       Ef_writeInput,  { Store (write) only input in file.           }
       Ef_writeText,   { Store (write) transcript as pure text file. }
       Ef_readIn,      { Load  (read ) file, inserting it at cursor. }
       Ef_namefile,    { Give the transcript a new (file) name.      }

                   { PROCESS }

       Ef_process,     { suspend editing, process text up to cursor  }
       Ef_procline,    { suspend editing, process current line       }
       Ef_procend,     { suspend editing, process to end of input    }
       Ef_procitem,    { append word to prev input, process the line }
                       { (this is the "menu selection" function)     }
       Ef_abort,       { abort processing, terminate (sub)session    }
       Ef_complete,    { ask processor to complete this input        }
       Ef_query,       { ask processor to show possible completions  }

       Ef_exit,        { save text, exit editor                      }
       Ef_quit,        { quit editor (must confirm if text unsaved)  }

                 { TRACING ETC. }

       Ef_logmore,     { increment log level (debugging)             }
       Ef_logless,     { decrement log level (debugging)             }
       Ef_logwindow,   { list window data on log file                }
       Ef_logscript,   { list transcript data on log file            }
       Ef_spotcheck,   { check transcript spots for true row, col &c }
       Ef_whereline,   { display status line with spot locations &c  }

       Ef_disabled,    { function now allowed in this window/doc     }
       Ef_ambiguous);  { ambiguous or overdefined function (top)     }
