{ File descriptors sets (bitmaps) for use with Unix "select" call. }

#if !defined(_FD_SETS_P)
#define _FD_SETS_P

#if !defined(FD_SETSIZE)
#define FD_SETSIZE 1024
#endif

const
   fd_setsz    = FD_SETSIZE;                    
   fd_wordsize = 32;                       { bits per word     }
   fd_words    = (fd_setsz+fd_wordsize-1) div fd_wordsize;
                                           { number of words   }
   fd_width    = fd_wordsize * fd_words;   { total size (bits) }

type
   fd_set_t = array [0..fd_words-1] of integer;

function FD_One (n: integer): integer;
  { Returns word with only bit n set.
    Should use left shift function, but cannot remember
    what it's called in Sun Pascal. }
  var i,x: integer;
  begin
  if n = 31 then FD_One:=16#80000000  { only negative value }
  else begin
    x:=1;
    for i:=1 to n do x:=x+x;
    FD_One:=x;
    end;
  end {FD_One};
  
procedure FD_ZERO (var fdset: fd_set_t);
  { Make the set empty. }
  var i: integer;
  begin
  for i:=0 to fd_words-1 do fdset[i]:=0;
  end {FD_ZERO};

procedure FD_SET (fd: integer; var fdset: fd_set_t);
  { Add 'fd' to the set, i.e. make it a member. }
  var word,bit: integer;
  begin
  word:= fd div fd_wordsize;
  bit := fd mod fd_wordsize;
  fdset[word]:=fdset[word] | FD_One(bit);
  end {FD_SET};

procedure FD_CLR (fd: integer; var fdset: fd_set_t);
  { Remove 'fd' from the set, i.e. make it not a member. }
  var word,bit: integer;
  begin
  word:= fd div fd_wordsize;
  bit := fd mod fd_wordsize;
  fdset[word]:=fdset[word] & ~ FD_One(bit);
  end {FD_CLR};

function FD_ISSET (fd: integer; var fdset: fd_set_t): boolean;
  { Whether 'fd' is in the set, i.e. is a member of the set. }
  var word,bit: integer;
  begin
  word:= fd div fd_wordsize;
  bit := fd mod fd_wordsize;
  FD_ISSET := (fdset[word] & FD_One(bit)) <> 0;
  end {FD_ISSET};

#endif
