{ File handling module.

  Provides two different "open" routines for safe reading and
  writing of text files.

  Uses the Sun Pascal ioerr.h and I/O eror trap facilities.

  Author: Niklas Holsti around 1994.07.27

}

module Files;

#include "ioerr.h"
#include "strings_d.p"

private var
  ioerr: IOerror_codes;

private
function ErrorHandler (in code: IOerror_codes;
                       in fileptr: univ_ptr): boolean;
  { This error handler just records the error code in the
    global variable and continues execution. }
  begin
  ioerr:= code;
  ErrorHandler:= true;  { keep going. }
{$w-}   { Yes, I know that fileptrr is not used, thank you very much. }
  end {ErrorHandler};
{$w+}


private
function GoodPathName (pathname: vstring): boolean;
  { Is the pathname superficially good? }
  begin
  if length(pathname) < 1 then     { pathname empty - a "no no" }
    GoodPathName:= false
  else
    GoodPathName:= not (
        (pathname[1] = '^')       { Sun Pascal special case (usee argv[n]) }
     or (pathname[1] = '*')       {  "    "      "       "  (ask for path) }
     or (pathname[1] = '-') );    {  "    "      "       "  (stdin or stdout) }
  end {GoodPathName};


function OpenTextFile (pathname: vstring; var f: text): boolean;
  { Opens the named file for reading.
    Returns true for success, false for failure.
  }
  var
    errcode: integer;
  begin
  if GoodPathName(pathname) then begin
    errcode:= 99;
    open (f, pathname, 'old', errcode);
    if errcode = 0 then reset(f);
    OpenTextFile:= errcode = 0;
    end
  else
    OpenTextFile:= false;
  end {OpenTextFile};


function WriteTextFile (pathname: vstring; var f: text): boolean;
  { Opens the named file for (re-)writing. The file may or may
    not exists before; if it exists, it is erased, then opened
    for writing.
    Returns true for success, false for failure.
  }
  var
    errcode: integer;
  begin
  if GoodPathName(pathname) then begin
    errcode:= 99;
    open (f, pathname, 'unknown', errcode);
    if errcode = 0 then rewrite(f);
    WriteTextFile:= errcode = 0;
    end
  else
    WriteTextFile:= false;
  end {WriteTextFile};


function NewTextFile (pathname: vstring; var f: text): boolean;
  { Creates a new file for writing.
    Returns true for success, false for failure.
    It is a failure if a file with the given name already
    exists, but the caller cannot distinguish this failure
    from other failures (e.g., no write permission to
    the directory).
  }
  var
    errcode: integer;
  begin
  if GoodPathName(pathname) then begin
    errcode:= 99;
    open (f, pathname, 'new', errcode);
    if errcode = 0 then rewrite(f);
    NewTextFile:= errcode = 0;
    end
  else
    NewTextFile:= false;
  end {NewTextFile};
