{ lexline_d.p
  Constants and types for lexline.p

  94.10.24  Omit "tag" field from Lex_PositionType.

}

const
  Lex_maxLine   = Ed_maxLine;    { line length limit (buffered mode) }
  Lex_maxWord   =    80      ;  { word (alphameric string) length limit }
  Lex_maxString = Lex_maxLine;  { string (in quotes) length limit,
                                  excluding the quotes. }

  Lex_Set       = '=';       { Opcodes for char-set operations }
  Lex_Include   = '+';
  Lex_Exclude   = '-';

type

  Lex_LineType  = Ed_LineType;
  Lex_WordType  = varying [Lex_maxWord] of char;
  Lex_RealType  = real;

  Lex_PositionType = record
    last     : ScriptIndex;  { last script position used by program }
    start    : ScriptIndex;  { start of current token (if > last,
                               there is no current token) }
    length   : integer;      { length of current token (if any) }
    character: char;         { read-ahead character }
  { tag      : vstring;   }  { for tracing and debugging. }
    end;

  { Note that a Lex_PositionType.last is the last position used by the
    application program, not the last position read from the editor.
    The last position read is the eol on the line last read. }

  TokenKind = (NullToken,            { no current token (keep this first) }
               CancelSignal,         { a cancel request, no valid token }
               NatNumToken,          { unsigned integer }
               RealNumToken,         { unsigned real number }
               MarkToken,            { some special character }
               SymbolToken,          { a word symbol }
               WordToken,            { a word (not defined as a symbol) }
               StringToken,          { a quoted string }
               EolToken,             { end-of-line }
               EolCommentToken,      { Ada-style rest-of-line comment }
               EofToken);            { end-of-file (not implemented, but
                                       keep this last in the list) }
  
  TokenType = record
    case kind: TokenKind of 
      NullToken      : ();
      CancelSignal   : ();
      NatNumToken    : (natnumber : integer);
      RealNumToken   : (realnumber: Lex_RealType);
      MarkToken      : (character : char   );
      SymbolToken    : (symnum    : integer);
      WordToken      : (word      : Lex_WordType);
      StringToken    : (quoted    : vstring;
                        opener,closer: char);
      EolToken       : ();
      EolCommentToken: ();
      EofToken       : ();
      end { TokenType };


  Lex_CsOpType = char;   { Type of char-set opcode }
