{ Log file.
  Author: Niklas Holsti    around 1991.4.11
  Changes:
     91.09.05  NH  Log file not opened until log-level becomes
                   positive, or Log_open is called.
                   Log_close added.
     92.04.09  NH  Port to Sun Unix.
     92.04.17  NH  Local vars 'private' for less risk of sym conflicts.
     94.08.04  NH  Log file in /tmp, name include process id.
                   Uses unistd_p.h.
}

module Log;

#include "strings_d.p"
#include "unistd_p.h"

var
  logfile  : text;     { public variable }

private var
  loglevel : integer;  { 0     - no log output     }
                       { 1,2,..- increasing detail }
  logopened: boolean;  { whether log file has been opened }

procedure Log_open;
  { Make sure that log file is opened. }
  var
    pid, x  : pid_t;
    logfname: vstring;  { name of log file. }
  begin
  if not logopened then begin
    pid:= getpid;
    logfname:= '';
    x:= pid;
    repeat
      logfname:= chr(ord('0')+(x mod 10)) + logfname;
      x:= x div 10;
    until x = 0;
    logfname:= '/tmp/seedlogfile.' + logfname;
    rewrite(logfile, logfname);
    logopened:=true;
    end;
  end {Log_open};

procedure Log_close (msg: vstring);
  { Write final message and close log file, it it was open. }
  begin
  if logopened then begin
    writeln(logfile,msg);
    writeln(logfile,'Closing log file.');
    close(logfile);
    logopened:=false;
    end;
  end {Log_close};

private
procedure WhatLevel;
  begin
  Log_open;
  writeln(logfile,'Log level := ',loglevel:1);
  end {WhatLevel};

procedure Log_level (change: integer);
  { Increase/decrease log level. }
  begin
  loglevel:=loglevel+change;
  if loglevel < 0 then loglevel:=0;
  if (loglevel > 0) or logopened then WhatLevel;
  end {Log_level};

function  Log_includes (level: integer): boolean;
  { Whether log level includes a given level. }
  begin
  Log_includes:= loglevel >= level;
  end {Log_includes};

procedure Log_Init;
  begin
  loglevel := 0;
  logopened:= false;
  end {LogInit};

procedure Log_Fini;
  begin
  Log_close('Finalized.');
  end {Log_Fini};
