{ scanline_f.p
  External declarations for scanline.p
}

var
  Scan_eol : external char;             { newline symbol from Scan_Char }
  Scan_can : external char;             { cancel (reverse) symbol from Scan_char }

procedure Scan_Store (var safe: Scan_PositionType;
                          tag : vstring);
  { Store a checkpoint of line scanner state in the output parameter.
    The tag is for tracing and debugging purposes. }
  external;

function Scan_Next: ScriptIndex;
  { Position of first script symbol not looked at yet. }
  external;

function Scan_Char (i: ScriptIndex): char;
  { Return the input symbol at the indicated script position. }
  external;

procedure Scan_Used;
  { Report last-used position to editor. }
  external;

procedure Scan_Mark;
  { Mark the current position for possible later use in an error
    message. There is only one mark, so later marks override earlie
    ones. See Scan_Error for further info. }
  external;

procedure Scan_Error;
  { Signals an error at this point in the script.
    The error point is the highest ScriptIndex used in Scan_Char.
    The error substring is starts at the most recent Scan_Mark and
    ends at the error point. }
  external;

procedure Scan_exit;
  { Signals that a grain has completed parsing with success. }
  external;

function Scan_done: boolean;
  { Returns "true" if the current grain should be terminated 
    (with success or cancelled). Returns false if the current grain
    should be reprocessed. }
  external;

function Scan_cancel: boolean;
  { Whether the preceding Scan call "raised" the cancelled "exception".
    Any call that asks for more input can cause cancel.
  }
  external;

procedure Scan_Restore (safe: Scan_PositionType);
  { Tries to restore the state at "safe". If it succeeds,
    resets Scan_cancel and Scan_done to false ("exception handled").
    If it fails, sets them true ("exception reraised"). }
  external;

procedure Scan_Init;
  { Initialize the ScanLine module. }
  external;
