{ screen_f.p
  External declarations of screen.p routines.
}

procedure Sc_Explain (var f: text; result: Sc_result);
  { Write an explanation of the result. }
  external;

procedure Sc_SetArea (    rows  : Sc_rowRange;
                          cols  : integer;
                      var result: Sc_result);
  { Define the row range and width of the screen. The usual terminals
    have 24 rows and 80 columns. Also sets scrolling region to the
    given row range. }
  external;

procedure Sc_GetArea (var rows: Sc_rowRange;
                      var cols: integer);
  { Inquire screen area. }
  external;

procedure Sc_Goto (row,col: integer; var result: Sc_result);
  { Move cursor (insert/overstrike point) to the given row and
    column. Row and column numbers start from 1. }
  external;

procedure Sc_Wrote (cols: integer);
  { Informs us that a number of columns have been written
    at the cursor, which thus has moved along. }
  external;

procedure Sc_Limbo;
  { Informs us that the cursor position is not known. }
  external;

procedure Sc_Clear;
  { Clear the whole screen and put cursor at row 1, col 1. }
  external;

procedure Sc_ClearEOL;
  { Clear from cursor position to end of line (row). Cursor position
    is included. }
  external;

function  Sc_CanInsert: boolean;
  { Whether insert mode can be entered. }
  external;

procedure Sc_InsertMode;
  { Enter insert mode. Text output will appear at the cursor position,
    and the cursor and the rest of the text on the line will be pushed
    right. Text at the right margin spills over and is lost.
    No effect if Sc_CanInsert returns false. }
  external;

procedure Sc_OverstrikeMode;
  { Enter overstrike mode. Text output will appear at the cursor position,
    overwriting the old text. The cursor moves right, but no text moves.
    The overwritten text is lost. }
  external;

function  Sc_CanDeleteChar: boolean;
  { Whether Sc_DeleteChar works. }
  external;

procedure Sc_DeleteChar (n: integer);
  { Delete n characters starting at the cursor and move remaining text
    on the line n positions left to fill the gap. The cursor remains
    at the same position. Space characters are put in the right margin.
    No effect if Sc_CanDeleteChar returns false. }
  external;

procedure Sc_SetScroll (rows: Sc_rowRange; var result: Sc_result);
  { Set scrolling region. Line insertion or deletion is allowed only
    in the row range given, and only this region is affected,
    all other lines are unchanged.
    After Sc_SetArea, scrolling region is whole screen. }
  external;

procedure Sc_InsertLine (var result: Sc_result);
  { Insert a new blank line at the cursor. Lines from the cursor to the
    line before the end of the scrolling region all move down one step.
    The line at the end of the scrolling region is lost.
    The cursor stays on the same row, but may move to column 1. }
  external;
    
procedure Sc_DeleteLine (var result: Sc_result);
  { Delete line at cursor. Cursor must be on a line within the
    scrolling region. Lines from the next line to the end of the
    scrolling region move up one step. A blank line is inserted at the
    end of the scrolling region.
    The cursor stays on the same row, but may move to column 1. }
  external;

procedure Sc_Flush;
  { Make sure all data intended for screen get there, ie.
    dump buffer. }
  external;

procedure Sc_Init;
  { Initialize Screen package.
    Must be called once, before any other routines.
  }
  external;

procedure Sc_Fini;
  { Finalize the Screen package.
    Cleans up the screen. }
  external;
