{ strings.p

  String operations a'la Turbo Pascal.

  Changes:
    1993.04.21  NH  String length changed from 255 to 256 in an attempt
                    to get rid of data (bus) alignment errors when
                    calling Seed routines from C. Change is actually in
                    strings_d.p.
    1994.07.27  NH  Cstring added.
}


module Strings;

#include "strings_d.p"

function Copy (s: vstring; index,count: integer): vstring;
  { Return the substring of s, starting at index and
    extending for count characters, or as many as can
    be found in s[index..].
    If index is > length(s), return null string.
  }
  var f: integer;
  begin
  f:=length(s)-index+1;
  if f < 1 then Copy:=''
  else begin
    if f > count then f:=count;
    Copy:=substr(s,index,f);
    end;
  end {Copy};

procedure Delete (var s: vstring; index,count: integer);
  { Delete count bytes at s[index..] }
  var t: integer;
  begin
  if index <= length(s) then begin
    t:=index+count;
    if t <= length(s) then
      s:= substr(s,1,index-1) + substr(s,t,length(s)-t+1)
    else
      s:= substr(s,1,index-1);
    end;
  end {Delete};

procedure Insert (source: vstring; var s: vstring; index: integer);
  { Insert the source string at v[index..] }
  var f,l,t: integer;
  begin
  t:=maxstring-index+1;   { space remaining }
  f:=length(source);
  if f > t then f:=t;  { this much from source }
  t:=t-f;
  l:=length(s)-index+1;
  if l > t then l:=t;     { this much from end of s }
  if f > 0 then begin
    if l > 0 then
      s:=substr(s,1,index-1) + substr(source,1,f) + substr(s,index,l)
    else
      s:=substr(s,1,index-1) + substr(source,1,f);
    end;
  end {Insert};

procedure Cstring (str: univ_ptr; var s: vstring);
  { str points to a null-terminated "C-style" string.
    This string is copied to the output parameter s.
    At most 'maxstring' characters are copied, any
    overflow is silently discarded.
  }
  type
    cstr = packed array [1..maxstring] of char;
  var
    csp: ^cstr;
    n  : integer;
  begin
  csp:= str;
  s:= '';
  n:= 1;
  repeat
    if csp^[n] <> chr(0) then begin
      s:= s + csp^[n];
      n:= n+1;
      end
    else
      n:= maxstring+1;
  until n > maxstring;
  end {Cstring};
