{ transcripts_d.p
  Constants and types for transcripts.p
  Requires: strings_d.p
  NOTE! This version is quite wasteful in storage, since each
  text element is allocated a full "vstring", which may be some
  250 bytes.
}

const
  MaxCol = 2000;  { a reasonably large maximum line length }

  { The following constants are intended to be PRIVATE }
  MaxElemLength =  maxstring;  { Max length of text of element. }
  MaxSpots      =  10;         { Max number of spots in a transcript. }

type
  Spot       = integer;
  Ts_result = (Ts_Ok,
               Ts_spotsfull,   { Spot table full }
               Ts_memfull,     { Memory full }
               Ts_elemfull,    { Transcript element full }
               Ts_immutable,   { Transcript element edit not allowed }
               Ts_nontext,     { Not a text element }
               Ts_noelem,      { No element (nil ref) }
               Ts_nullelem,    { Null element (zero length) }
               Ts_nonewline,   { Newline missing }
               Ts_wasnewline,  { Newline present, should not be }
               Ts_wrongorder,  { Two spots in wrong order }
               Ts_noteot,      { Spot should be at EOT }
               Ts_atomic,      { Element cannot be split }
               Ts_notinput,    { Not at input item }
               Ts_endinput,    { End of input (at EOT) }
               Ts_validoutput, { Valid output, not dewritten }
               Ts_fixedoutput, { Cannot insert text before output }
               Ts_wrongform,   { Input file format wrong }
               Ts_invalid      { Invalid or unimplemented operation }
              );

  Ts_fileOpt = (               { How to store the transcript in a file. }
               StoreInput,     { .. input part only }
               StoreAsText,    { .. input and output as pure text }
               StoreAll        { .. input and output, tagged as such. }
              );

  Ts_dataSource = (InputSource,   { User typed it }
                   OutputSource,  { Program wrote it }
                   OmSource);     { Bot/Eot element }




  { The following types are intended to be PRIVATE }
  ElementRef    = ^Element;
  ElementKind   = (BotElement,    { Beginning of transcript }
                   TextElement,   { Text, input or output }
                   OtherElement,  { Other data, input or output }
                   EotElement     { End of transcript }
                  );
  ElementSource = Ts_dataSource;

  Element     = record         { An element of a transcript. }
    prev,next: ElementRef;     { form a doubly linked list }
    source   : ElementSource;  { input, output, neither (eot) }
    newline  : boolean;        { whether element ends a line }
    safe     : integer;        { causal input symbol (outputs only) }
    case kind: ElementKind of
      BotElement  : ();
      TextElement : (theText: vstring {varying [MaxElemLength] of char});
      OtherElement: ();
      EotElement  : ();
    end {Element};

  SpotOb = record         { An insertion point in a transcript }
    elem   : ElementRef;  { the element in which the spot lies }
    offset : integer;     { spot is after elem^.theText[offset] }
    row,col: integer;     { transcript row and column of char after spot }
    knowcol: boolean;     { whether col is known }
    inpos  : integer;     { number of input symbols (characters, newlines,
                            etc.) that lie to the left of the spot }
    end {SpotOb};

  Transcript = record     { A transcript. }
    botElem : ElementRef;                    { header element   }
    eotElem : ElementRef;                    { trailer element  }
    spots   : array [1..MaxSpots] of SpotOb; { spot data array        }
    spotlist: set of 1..MaxSpots;            { these spots are in use }
    hiSpot  : integer;                       { highest-numbered spot or 0 }
    modified: boolean;  { whether content modified since creation or
                          last defined "original" state (changes to spots do
                          not count) }
    end {Transcript};

