{ views_d.p
  Data structures for "views" that consist of a screen window,
  a transcript, and a cursor position.

  Changes:
     94.08.04 Add modifiable and modified flags.
     94.10.07 Modified flag has different meaning (Transcripts module notes
              changes, we just mirror the flag to know when to update the
              window title).
     94.10.23 Add origname to track changes in transcript file name.
}

{ requires "edit_d.p" }

const

  MaxWindows   = 10;

  MaxTipName   = 30;  { Max length of transcript name }
  MaxFileName  = 50;  { Max length of file name }

  MaxExecs     = 20;  { Max number nested exec modes }

const
  { Log levels }
  pauselog     = 2;   { Pause between each editing/processing operation }
  cleanlog     = 4;   { Log window cleaning (updating) operations }
  clwinlog     = 4;   { Log window data structures in connection with above }

type

  ExecMode = record   { Defines the start of a certain execution mode }
    mode    : Ed_ExecMode;   { manual, job, incremental }
    startPos: ScriptIndex;   { first input symbol covered }
    end;

  EtipRef = ^EditTranscript;

  EditTranscript = record       { A transcript being edited }
    name         : vstring;     { buffer name }
    origname     : vstring;     { file last loaded from or saved to }
    filename     : vstring;     { file for next save }
    theTranscript: Transcript;  { the inputs and outputs }
    modifiable   : boolean;     { can be edited }
    modified     : boolean;     { content has changed since last save }
    theMark      : Spot;        { defines a region mark..cursor }
    marked       : boolean;     { region defined (logically) }
    readpt       : Spot;        { read pointer }
    writept      : Spot;        { write pointer }
    writefd      : integer;     { async write file desc, active if >= 0 }
    outputfd     : integer;     { "output" stream file desc, active if >= 0 }
    lastPos      : ScriptIndex; { last script sym read }
    execPos      : ScriptIndex; { process up to and incl this script sym }
    editPos      : ScriptIndex; { first changed script sym }
    errorPos     : ScriptIndex; { script symbol in error }
    errorFirst   : ScriptIndex; { start of error token }
    errorLength  : ScriptIndex; { length of error token }
    startPos     : ScriptIndex; { next script sym to be read or re-read }
    usedPos      : ScriptIndex; { last script sym claimed to be "used" }
    skipwrite    : boolean;     { ignore writes (cancelling) }
    aborting     : boolean;     { user wants to abort this (sub)session }
    modeStack    : array [1..MaxExecs] of ExecMode;
    modeTop      : integer;
    nextEtip     : EtipRef;     { linked list of all transcripts }
    end;

  ViewRef  = ^EditView;

  EditView = record        { A window has been used to view ... }
    refEtip : EtipRef;          { this transcript, }
    cursor  : Spot;             { with the cursor here, }
    nextView: ViewRef;          { and perhaps others views, too. }
    end;

  TitleStyle = (           { How a window's title bar is displayed. }
                NoTitle,        { the window has no title bar }
                NormalTitle,    { normal, unemphasized (not current) }
                BoldTitle);     { bold, emphasized (current window) }

  EditWindow = record
    theWindow    : Window;
    refEtip      : EtipRef;     { currently viewed transcript and }
    cursor       : Spot;        { cursor in it }
    title        : TitleStyle;  { present title bar style }
    titleRow     : integer;     { screen row for window title bar }
    views        : ViewRef;     { other transcript:cursor pairs }
    end;

  WinIndex = 1..MaxWindows; { Hey window! We have your number! }

  ScreenPlace = record
    row,col  : integer;
    end;
