{ views_f.p
  External declarations for routines in views.p
}

var  { NOTE -- bad to have public vars, change! }

  edwin: extern array [WinIndex] of EditWindow;

procedure Vw_Init;
  { Initialize this module. }
  external;

procedure Vw_Fini;
  { You guessed it. }
  external;

function Vw_InfiniteIndex: ScriptIndex;
  { Return a (constant) value that behaves as an infinite ScriptIndex. }
  external;

function Vw_NumWindows: integer;
  { Returns number of windows open at present.
   The window numbers are 1..NumWindows.
  }
  external;

function Vw_InQueryMode: boolean;
  { Is the query area in use? }
  external;

procedure Vw_StatusPlace (num, col: integer; var place: ScreenPlace);
  { Define a place on a screen status line.
    Status line number num is = 0, 1, 2, .. relative to the first
    line reserved for status.
    Column position is relative to left margin, col = 1, 2, 3.
  }
  external;

procedure Vw_NewEditableTranscript (    aname  : vstring;
                                        amode  : Ed_ExecMode;
                                        modify : boolean;
                                    var theEtip: EtipRef);
  { Creates a new editable transcript. The transcript name (window title)
    and initial execution mode must be specified, also whether the
    transcript is modifiable by the user.
    The transcript is initially empty and has a null filename.
    The new transcript is added to the set of transcripts maintained
    by the editor.
    The transcript is not yet displayed in any windows.
  }
  external;

procedure Vw_KillEditableTranscript (var theEtip: EtipRef);
  { Destroys an editable transcript.
    Removes it from the the list of transcripts first.
  }
  external;

procedure Vw_CleanWindow (w: WinIndex; var clean: boolean);
  { Clean up (refresh) window number w until it is clean or 
    user presses a key. Parameter 'clean' shows whether the
    window was cleaned.
  }
  external;

procedure Vw_GoCursor (w: WinIndex);
  { Move screen cursor to text cursor position of this window.
    Scroll the window if necessary.
    Includes cleaning of this window, but not others.
  }
  external;

procedure Vw_Title (w: WinIndex; title: TitleStyle);
  { Define the title bar style for a window. }
  external;

procedure Vw_WriteTitle (w: WinIndex);
  { Write title bar for this window. }
  external;

procedure Vw_WriteTitles (theEtip: EtipRef);
  { Write title bar for all windows showing this transcript. }
  external;

function Vw_Split (w: WinIndex; var low: WinIndex): boolean;
  { Divide a window into two halves, if possible.
    The upper half has the same index, the lower half a new one.
  }
  external;

procedure Vw_CloseWindow (w: integer);
  { Close this window. Leave screen area unused. }
  external;

procedure Vw_CloseOthers (var except: WinIndex);
  { Close all except the current window. Make the current window
    grow to use all of the available screen. }
  external;

procedure Vw_View (w: WinIndex; etr: EtipRef);
  { Make the window w view the transcript etr. }
  external;
   
function Vw_FirstTranscript: EtipRef;
  { Return the first transcript. }
  external;

function Vw_NextTranscript (etr: EtipRef): EtipRef;
  { Return the next transcript, or nil if none. }
  external;

function Vw_NewTranscript (w: WinIndex; modify: boolean): boolean;
  { Create a new transcript for view in this window, specifying if
    it will be modifiable or for view only.
    Return success/failure.
  }
  external;

procedure Vw_NameFile (w: WinIndex; fname: vstring; override: boolean);
  { Assign the file name of the current view in this window, if no
    name has yet been assigned.
    Use override = true to change a name assigned earlier.
    The transcript name is set to the base name of the file.
  }
  external;

procedure Vw_ReadFile (w: WinIndex; var ff: text; fname: vstring;
                       var ok: boolean);
  { Read in a text or transcript from a file into the current view.
    The file ff is assumed to be opened already using the name fname and
    can contain pure text or a transcript.
    The file is then closed. Success is reported in ok.
  }
  external;

procedure Vw_Modifiable (w: WinIndex; modify: boolean);
  { Control whether transcript in the current view in this window is
    to be modifiable (editable, changeable) or not.
  }
  external;

function Vw_Modified (w: WinIndex): boolean;
  { Whether the transcript in the current view in this window has
    been changed (input, output, or both) since its last "original"
    state (eg. creation, load from file). }
  external;

procedure Vw_Original (w: WinIndex);
  { Declare that this transcript is now in an "original" state, eg.
    just created and loaded from a file. }
  external;

procedure Vw_RefreshScreen;
  { Redraw all windows on screen, including title bars. }
  external;

procedure Vw_InsCharWs (etr: EtipRef; row,col: integer; num: integer);
  { Insert a number of characters at row,col in all windows
    for transcript etr. }
  external;

procedure Vw_InsCharWs (etr: EtipRef; row,col: integer; num: integer);
  { Insert a number of characters at row,col in all windows
    for transcript etr. }
  external;

procedure Vw_DelLineWs (etr: EtipRef; row: integer; num: integer);
  { Delete a number of lines starting with 'row' in all windows
    for transcript etr. }
  external;

procedure Vw_InsLineWs (etr: EtipRef; row: integer; num: integer);
  { Insert a number of lines at 'row' in all windows
    for transcript etr. }
  external;

procedure Vw_FreshLineWs (etr: EtipRef; row,col: integer);
  { Freshen (redisplay) line 'row' from column 'col' in all windows
    for transcript etr. }
  external;

procedure Vw_NewlineWs (etr: EtipRef; row,col: integer);
  { Make a newline at 'row' column 'col' in all windows for
    transcript etr. }
  external;

procedure Vw_InsShow (theSpot: Spot; refEtip: EtipRef;
                      preRow,preCol: integer);
  { Update screen after text has been inserted.
    Text (perhaps including newlines) was inserted at the left
    of the given spot. Before the insertion, the spot was at
    the given row and column. The transcript is refEtip. }
  external;

procedure Vw_DelShow (refEtip: EtipRef;
                      delRow,delCol,preRow,preCol: integer);
  { Update screen after text has been deleted in a transcript.
    Text (perhaps including newlines) was deleted, starting at
    delRow,delCol and continuing up to but not including preRow,preCol. }
  external;

procedure Vw_DelShot (theSpot: Spot; refEtip: EtipRef;
                      preRow,preCol: integer);
  { Update screen after text (perhaps including newlines) was deleted
    at the left of the given spot. Before the deletion, the spot was
    at the given row and column. The transcript is refEtip. }
  external;

{ Status line and similar output routines. }

procedure Vw_WriteChr (c: char);
  { Display (render) a character for status line or other such. }
  external;

procedure Vw_WriteStr (s: vstring);
  { Display (render) a string for status line or other such. }
  external;

procedure Vw_DecimalString (n: integer; w: integer; var s: vstring);
  { Format n as decimal integer. Fill with leading blanks to at least
    w characters. Result in s.
    Note: implementation is not efficient. }
  external;

procedure Vw_ScProblem (result: Sc_result);
  { Show error result from Screen module. }
  external;

procedure Vw_TsProblem (result: Ts_result);
  { Show error result from Transcript module. }
  external;

procedure Vw_WdProblem (result: Wd_result);
  { Show error result from Windows module. }
  external;

procedure Vw_KmProblem (result: Km_result);
  { Show error result from Keymap module. }
  external;

procedure Vw_Message (txt: vstring);
  { Display message. }
  external;

procedure Vw_ClearMess;
  { Clear message line and note absence of message.
    Nothing done if there was no message to start with. }
  external;

function Vw_MessPause (txt: vstring; w: WinIndex): boolean;
  { Message and pause.
    User can dump the indicated window, increase/decrease log levels,
    and ask program to stop.
    Returns "true" if execution is to continue.
  }
  external;

procedure Vw_Trace (sw: boolean);
  { Turn tracing on or off. }
  external;

procedure Vw_CreateQuery (var w: WinIndex; var ok: boolean);
  { Open a new interaction (window + transcript) to ask the
    user for something, e.g. a file name to store the transcript in.
    The new window is returned in w. Success/failure in ok.
  }
  external;

procedure Vw_QuitQuery;
  { Delete the query interaction. }
  external;
