{ windows_d.p
  Constants and types for windows.p
}

const
  Wd_MaxRows = 100;    { Maximum number of rows a window can cover }
  Wd_Vscroll =   1;    { Can scroll vertically as little as one line }
  Wd_Hscroll =  20;    { Horizontal scroll at least 20 cols at a time }

type
  Wd_result = (
              Wd_ok,           { Window operation sucessful. }
              Wd_norows,       { Window would have no rows. }
              Wd_offscreen,    { Window would extend off screen. }
              Wd_norowsleft,   { No rows would be left in lower part. }
              Wd_negspace,     { Negative space in split. }
              Wd_overlap,      { Windows would overlap }
              Wd_unused        { Some rows were unused. }
              );

  Window = record
     screenTopRow: integer;   { First screen rows used. }
     screenBotRow: integer;   { Last    "      "    "   }
     textTopRow  : integer;   { First text row visible. }
     rows        : integer;   {  = screenBotRow - screenTopRow + 1 }
     cols        : integer;   { Width in columns }
     band        : integer;   { Half-height of "band" of rows that
                                scroll together horizontally, zero for
                                scrolling each row independently. }
     topDirt     : integer;   { First screen row that may need refresh. }
     cached      : record     { Cached data to speed up some things: }
       screenRow : integer;   { .. screen row we looked at last. }
       end;
     end {Window};


  { When a window display is being updated, the updating routine (Wd_Clean)
    returns a Wd_ScreenOps data structure that tells the editor what to
    do to the screen. The operations are to be done in the order they
    are listed below: first delete a row; then insert a row; then
    overwrite a whole row; then insert or delete characters at some column
    of some row; and finally overstrike characters from some column to
    the end of that row. Usually, only some of these operations are
    requested by Wd_Clean.
       The screen row numbers are given relative to the screen as it is
    successively changed by these operations, so they can be used
    directly in Sc_DeleteLine and Sc_InsertLine routines. The text row
    and column numbers are given relative to the current state of the
    text, i.e. after all the editing operations that have led to the
    window needing updating, so they can also be used directly to find
    the rows in the text.
       The updating routine may have to be called repeatedly, each time
    returning some actions, until the window is clean. The window
    data structure is kept consistent throughout, assuming that the
    screen operations returned by each call to Wd_Clean are implemented.
    Thus, the updating process can be interrupted before the window is
    clean - for example, if the user hits a key such as Next-Page that
    makes it unnecessary to clean the window now.
  }

  Wd_ScreenOps = record
      deleteRow   : integer;  { Delete this screen row (if > 0) and
                                scroll lower rows up one position. }
      insertRow   : integer;  { Insert a new blank line at this screen
                                row (if > 0), scrolling the row and
                                lower rows down one position. }
      writeFull   : record    { Overstrike all text on .. }
         screenRow: integer;  { ..this screen row (if > 0) }
         textRow  : integer;  { ..taking text from this text row }
         textCol  : integer;  { ..and this text column. }
         end;
      withinRow   : record    { Within-row operations .. } 
         screenRow: integer;  { ..in this screen row (if > 0) }
         textRow  : integer;  { ..taking text from this text row }
         insdel     : record    { Insert or delete .. }
           chars    : integer;  { ..this number of chars (+/- for ins/del) }
           screenCol: integer;  { ..starting at this screen column }
           textCol  : integer;  { ..taking text from this column. }
           end;
         overstrike : record    { Overstrike to end of line.. }
           chars    : integer;  { ..this number of chars (at most) }
           screenCol: integer;  { ..starting at this screen column }
           textCol  : integer;  { ..taking text from this column.  }
           end;
         end;
      end {Wd_ScreenOps};
