{ edit_f.p
  External declarations for routines in edit.p
  Requires: strings_d.p
}

procedure Ed_Init;
  { Initialize editor module. }
  external;

procedure Ed_Fini;
  { Finalize editor module. }
  external;

procedure WriteIndex (index: ScriptIndex);
  { Write index on standard output. }
  external;

procedure Ed_Write (txt: vstring);
  { Write output text into the transcript. }
  external;

procedure Ed_Writeln;
  { Write an output newline into the transcript. }
  external;

procedure Ed_WrInt (i: integer);
  { Write the integer as output into the transcript. }
  external;

procedure Ed_WrReal (r: real; w,d: integer);
  { Write the real number as output into the transcript.
    Uses the same format as write(r:w:d). }
  external;

function Ed_WriteFrom (fd: integer): boolean;
  { Request editor to accept data from the file descriptor 'fd'
    and store it in the transcript as output. The editor will
    do this in the background, while editing proceeds normally.
    However, the position and tagging of the output may not
    be correct, if the user asks the editor to process some
    other part of the script.
    Returns true for success, false for error. At the moment,
    the only error is to try to define more than one file
    descriptor - only one is allowed at present. }
  external;

procedure Ed_written;
  { Request editor to get all output data that is now available in the
    file descriptor defined by Ed_WriteFrom, and tag and store it
    in the transcript. An application should call this routine
    after it has written something to the file descriptor, and before
    it gets more input (otherwise, the output would be processed later,
    and tagged with a later input position). }
  external;

function Ed_WriteEnds: boolean;
  { Cancels the preceding Ed_WriteFrom request. The editor shall
    no longer accept data from the file descriptor there specified.
    Returns true on success, false for error. At the moment, the
    only error occurs if there is no active output file descriptor. }
  external;

function Ed_CatchOutput: boolean;
  { Request editor to redirect the Pascal standard "output" stream
    to a temporary file, for later inclusion in the transcript
    using Ed_ShowOutput. Returns true for success. }
  external;

procedure Ed_ShowOutput;
  { When an Ed_CatchOutput is in effect, this makes the editor
    update the transcript to include all the "output" text that
    was produced after the call to Ed_CatchOutput, or the
    last call to Ed_ShowOutput. Typically, CatchOutput is
    called once at the start of an application, and ShowOutput
    is called after each token has been processed. }
  external;

procedure Ed_Used (pos: ScriptIndex);
  { Tell editor the index of the last input symbol interpreted by the
    processor. The processor may have read later input symbols, but
    guarantees that they have no effect on subsequent output, until
    the next call of Ed_Used or Ed_Tell. } 
  external;

procedure Ed_Tell (last,error,start: ScriptIndex;
                   errfst: ScriptIndex;
                   errlen: integer);
  { Sends processor state to Editor.
    The index of the last input symbol read is 'last'.
    The index of the input symbol at which an error has been
    detected is 'error' (infinite if no error).
    If the processor is reporting an error, the erroneous
    input symbol is identified by 'errfst' and 'errlen' - the
    former is the index of the first symbol of the token, and
    the latter is the length of the token.
    The index of the input symbol the processor wants to
    read next is in 'start'. Usually, start = last+1. }
  external;

procedure Ed_Get (var msg: Ed_MsgType);
  { Get next input from Editor. }
  external;

procedure Ed_Getln (var line  : Ed_LineType;
                    var length: integer;
                    var msg   : Ed_MsgType);
  { Get inputs from Editor until end of line, end of file, or
    a cancel condition. A line of 'length' (possibly zero) characters
    is returned in 'line', and the terminating condition in 'msg'. }
  external;

function Ed_InfiniteIndex: ScriptIndex;
  { Return a position far, far away ... }
  external;

function Ed_BeginMode (newmode: Ed_ExecMode; newpos: ScriptIndex): boolean;
  { Start a new, nested execution mode.
    The first script position executed in the new mode is 'newpos'.
    This must be later than the start of the preceding active mode.
    The function returns 'true' if the mode could be started, and
    'false' for errors. }
  external;

function Ed_EndMode: boolean;
  { Terminate the current, nested execution mode.
    Returns 'true', unless the current mode is the original (bottom)
    mode, which should never terminate. }
  external;
