/* seed_scanline.h
   Constants, types and routines for scanline.p
*/

#if !defined(SEED_SCANLINE)
#define SEED_SCANLINE

/* line length limit (buffered mode) */
#define SCAN_MAXLINE ED_MAXLINE

typedef struct {
    scriptindex last;	/* last script position used by program */
    scriptindex mark;	/* a marked position (eg. start of token) */
    vstring     tag;	/* for tracing and debugging. */
    } scan_positiontype;

  /* Note that a Scan_PositionType.last is the last position used by the
    application program, not the last position read from the editor.
    The last position read is the eol on the line last read. */


extern char scan_eol;	/* newline symbol from Scan_Char */
extern char scan_can;	/* cancel (reverse) symbol from Scan_char */

extern void scan_store (scan_positiontype *safe, vstring tag);
  /* Store a checkpoint of line scanner state in the output parameter.
    The tag is for tracing and debugging purposes. */

extern scriptindex scan_next ();
  /* Position of first script symbol not looked at yet. */

extern char scan_char (scriptindex i);
  /* Return the input symbol at the indicated script position. */

extern void scan_used ();
  /* Report last-used position to editor. */

extern void scan_mark ();
  /* Mark the current position for possible later use in an error
    message. There is only one mark, so later marks override earlie
    ones. See Scan_Error for further info. */

extern void scan_error ();
  /* Signals an error at this point in the script.
    The error point is the highest ScriptIndex used in Scan_Char.
    The error substring is starts at the most recent Scan_Mark and
    ends at the error point. */

extern void scan_exit ();
  /* Signals that a grain has completed parsing with success. */

extern seed_bool scan_done ();
  /* Returns "true" if the current grain should be terminated 
    (with success or cancelled). Returns false if the current grain
    should be reprocessed. */

extern seed_bool scan_cancel();
  /* Whether the preceding Scan call "raised" the cancelled "exception".
    Any call that asks for more input can cause cancel.
  */

extern void scan_restore (scan_positiontype safe);
  /* Tries to restore the state at "safe". If it succeeds,
    resets Scan_cancel and Scan_done to false ("exception handled").
    If it fails, sets them true ("exception reraised"). */

extern void scan_init ();
  /* Initialize the ScanLine module. */

#endif
