#ifndef H_TELA_EXECUTION_MACHINERY
#define H_TELA_EXECUTION_MACHINERY

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:      Alg_Exec_Machinery 


		Contents:    TELA Algorithm Execution Machinery

		Description: handles the execution of external learning programs, 
                 builds argument vectors and prepares input, runs the 
                 learning programs in a separate process and collects 
                 statistics.

    Services:

    Input preparation
    =================

	  EXEC_write_atts_exs - write the attributes and examples 	
	  EXEC_build_RUN_instruction - build the argument vector for RUN-command
	  EXEC_build_TEST_instruction - build the argument vector for RUN-command

    Execution
    =========
		
	  EXEC_run_algorithm - execute the learning program in a separate process

    Output collection and cleanup
    =============================

	  EXEC_collect_stats - collect the recorded statistics
 	  EXEC_remove_temp_files - remove temporary attribute and example files


    Author: Juho Rousu (1995,1996)

\*                                                                            */

#include "tela.h"
#include "UI_parsers.h"

int EXEC_run_algorithm(char **param_table,int *p_status,float *time);

int EXEC_write_atts_exs(t_algorithms alg, t_ex_mask *p_mask);
int EXEC_build_RUN_instruction(t_run_1 param, char ***params,int mode);
int EXEC_build_TEST_instruction(t_test_1 param, char ***params,int mode);

int EXEC_run_algorithm(char **param_table,int *p_status,float *time);

int EXEC_collect_stats(t_algorithms alg,t_stats_list *p_stats);
int EXEC_remove_temp_files(t_algorithms alg);

#endif
