/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       DFD.y

		Contents:     Yacc parser sources for Algorithm Database
 
		Author:				Juho Rousu (1995,1996)

\*  																																					*/
 
%{
#include <alloca.h>
#include <stdlib.h>
#include "tela.h"
#include "Alg_Database.h"

#define GR_DEBUG

typedef char STRING_TYPE[256];
extern char buf[256];
extern char err_str[256];
t_alg_decl *p_alg_decl=0;

%}


%token ATTSEP NAMEVALSEP VALSEP ATTSEXSSEP NOMINAL NUMATTSEP
%token INT CONT UNKNOWN DONTCARE
%token FIRSTCHARSET CHARSET EXVALSEP EXAMPLESEP COMMENTSEP TYPEVALSEP
%token T_BEGIN END DESCRIPTION EOL FULL T_NULL DEFAULTS RULETYPE
%token EQUALS NEWLINE ALPHA OR EXENAME AE_FILE_SUPPORT
%token ATTSEXT EXSEXT BOTHEXT RULEEXT
%token EOL_TOKEN BLANK STRING CHAR QUOTED GRAMMAR OPTIONS RUNSTATS TESTSTATS
%token EXAMPLES ATTRIBUTES INTERFACE LAST FIRST UNNAMED CLASS
%token ATTSHEADER EXSHEADER BOTHHEADER RULEHEADER
%token ATTSFOOTER EXSFOOTER BOTHFOOTER RULEFOOTER


%%

alg_decl_file : alg_decls garbage ;
alg_decls : alg_decl alg_decls | ;
alg_decl : ad_header def_decl type_decl atts_decl exs_decl if_decls ad_trailer 
	{
		if (!ADB_add_algorithm(p_alg_decl))
		{
			yyerror("Invalid attribute description"); 
			return 0;
		}
	};

ad_header : 		T_BEGIN DESCRIPTION 	{
		p_alg_decl = (t_alg_decl*)malloc(sizeof(t_alg_decl));
		ADB_alg_decl_init(p_alg_decl);
	}
	alg_name NEWLINE; 
ad_trailer : 		END DESCRIPTION newlines {;}; 
garbage : STRING garbage | ;

def_decl : DEFAULTS EQUALS parent_name NEWLINE | ;
parent_name : STRING
	{
		if (!alloc_string(&p_alg_decl->parent_name,buf))
		{
			yyerror("Out of memory");
		}
	}

type_decl : RULETYPE EQUALS QUOTED NEWLINE
	{
		if (!alloc_string(&p_alg_decl->cf_type,buf))
		{
			yyerror("Out of memory");
		}
	} 

type_decl : ;

atts_decl	: ATTRIBUTES {;} NEWLINE attr_decl NEWLINE attr_decls | ; 
attr_decls : attr_decl NEWLINE attr_decls | ;
attr_decl	 : ATTSEP_DECL | NAMEVALSEP_DECL | VALSEP_DECL | nominal_decl
						| INT_DECL | CONT_DECL | ATTSEXSSEP_DECL | numattsep_decl 
						| UNKNOWN_DECL | DONTCARE_DECL | class_decl
						| FIRSTCHARSET_DECL | CHARSET_DECL | typevalsep_decl
						| COMMENTSEP_DECL | ATTSHEADER_DECL
						| EXSHEADER_DECL | BOTHHEADER_DECL | RULEHEADER_DECL
						| ATTSFOOTER_DECL | EXSFOOTER_DECL | BOTHFOOTER_DECL
						| RULEFOOTER_DECL
						;

exs_decl : EXAMPLES {;} NEWLINE ex_decls | ;
ex_decls : ex_decl NEWLINE ex_decls | ;
ex_decl : EXVALSEP_DECL | EXAMPLESEP_DECL | ex_class_decl	
					;

if_decls : INTERFACE {;} NEWLINE interface | ;
interface : if_decl interface | ;
if_decl : OPTIONS_DECLS | RUNSTATS_DECLS  | TESTSTATS_DECLS | EXENAME_DECL ;
if_decl : FILE_DECL NEWLINE ;
if_decl : FILE_SUPPORT_DECL ;

OPTIONS_DECLS : OPTIONS {;} NEWLINE OPTIONS_DECL;
OPTIONS_DECL: OPTION_DECL OPTIONS_DECL | ;
OPTION_DECL : STRING 
{
	char *new_buf=0;
	if (!realloc_string(&new_buf,buf))
		yyerror("Out of Memory");
	else if (!list_append(&p_alg_decl->interface.options,new_buf))
		yyerror("Out of Memory");
} NEWLINE ;

RUNSTATS_DECLS : RUNSTATS {;} NEWLINE RUNSTATS_DECL ;
RUNSTATS_DECL : RUNSTAT_DECL RUNSTATS_DECL | ;
RUNSTAT_DECL : STRING
{
	char *new_buf=0;

	if (!realloc_string(&new_buf,buf))
		yyerror("Out of Memory");
	else if (!list_append(&p_alg_decl->interface.runstats,new_buf))
		yyerror("Out of Memory");
} NEWLINE;

TESTSTATS_DECLS : TESTSTATS {;} NEWLINE TESTSTATS_DECL ;
TESTSTATS_DECL : TESTSTAT_DECL TESTSTATS_DECL | ;
TESTSTAT_DECL : STRING
{
	char *new_buf=0;

	if (!realloc_string(&new_buf,buf))
		yyerror("Out of Memory");
	else if (!list_append(&p_alg_decl->interface.teststats,new_buf))
		yyerror("Out of Memory");
} NEWLINE;

FILE_DECL : ATTS_EXT_DECL | EXS_EXT_DECL
	| BOTH_EXT_DECL | RULE_EXT_DECL ; 

ATTS_EXT_DECL : ATTSEXT EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.atts_ext,buf)) 
			yyerror("Out of Memory");
	};


EXS_EXT_DECL : EXSEXT EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.exs_ext,buf)) 
			yyerror("Out of Memory");
	};


BOTH_EXT_DECL : BOTHEXT EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.both_ext,buf)) 
			yyerror("Out of Memory");
	};


RULE_EXT_DECL : RULEEXT EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.rule_ext,buf)) 
			yyerror("Out of Memory");
	};

ATTSHEADER_DECL : ATTSHEADER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.atts_header,buf)) 
			yyerror("Out of Memory");
	}

EXSHEADER_DECL : EXSHEADER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.exs_header,buf)) 
			yyerror("Out of Memory");
	}

BOTHHEADER_DECL : BOTHHEADER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.both_header,buf)) 
			yyerror("Out of Memory");
	}

RULEHEADER_DECL : RULEHEADER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.rule_header,buf)) 
			yyerror("Out of Memory");
	}

ATTSFOOTER_DECL : ATTSFOOTER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.atts_footer,buf)) 
			yyerror("Out of Memory");
	}

EXSFOOTER_DECL : EXSFOOTER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.exs_footer,buf)) 
			yyerror("Out of Memory");
	}

BOTHFOOTER_DECL : BOTHFOOTER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.both_footer,buf)) 
			yyerror("Out of Memory");
	}

RULEFOOTER_DECL : RULEFOOTER EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.rule_footer,buf)) 
			yyerror("Out of Memory");
	}

COMMENTSEP_DECL: COMMENTSEP EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.comment_sep,buf)) 
			yyerror("Out of Memory");
	}

FILE_SUPPORT_DECL : AE_FILE_SUPPORT EQUALS STRING NEWLINE
	{
		if (!strcmp("SEPARATE",buf))
			p_alg_decl->interface.file_support = ONLY_SEP;
		else if (!strcmp("COMBINED",buf))
			p_alg_decl->interface.file_support = ONLY_COMB;
		else if (!strcmp("BOTH",buf))
			p_alg_decl->interface.file_support = COMB_N_SEP;
		else
		{
			sprintf(err_str,"Unknown file support specification: %s",buf);
			yyerror(err_str);
		}
	};



alg_name : STRING 
	{
		if (!realloc_string(&p_alg_decl->alg_name,buf)) yyerror("Out of Memory");
	};

class_decl: CLASS EQUALS class_specs;

class_specs : class_spec class_specs | ;

class_spec : FIRST
							{
								p_alg_decl->attr_decl.class.b_first = 1;
								if (p_alg_decl->attr_decl.class.b_unnamed == -1)
									p_alg_decl->attr_decl.class.b_unnamed = 0;
							}
class_spec : LAST
							{
								p_alg_decl->attr_decl.class.b_first = 0;
								if (p_alg_decl->attr_decl.class.b_unnamed == -1)
									p_alg_decl->attr_decl.class.b_unnamed = 0;
						}
class_spec : UNNAMED
							{
								p_alg_decl->attr_decl.class.b_unnamed = 1;							
								if (p_alg_decl->attr_decl.class.b_first == -1)
									p_alg_decl->attr_decl.class.b_first = 0;

							}

ex_class_decl: CLASS EQUALS ex_class_spec;

ex_class_spec: FIRST
							{
								p_alg_decl->ex_decl.b_class_first = 1;
							}
ex_class_spec: LAST
							{
								p_alg_decl->ex_decl.b_class_first = 0;
							}
ex_class_spec: ;

ATTSEP_DECL : ATTSEP EQUALS quoted_or_null { if
 (!realloc_string(&p_alg_decl->attr_decl.att_sep,buf)) 
			yyerror("Out of Memory");
	};

numattsep_decl : NUMATTSEP EQUALS quoted_or_null { if 
(!realloc_string(&p_alg_decl->attr_decl.numatt_sep,buf))
 	yyerror("Out of Memory");
	};


NAMEVALSEP_DECL : NAMEVALSEP EQUALS quoted_or_null { 
if (!realloc_string(&p_alg_decl->attr_decl.nameval_sep,buf)) 
			yyerror("Out of Memory");
	};

typevalsep_decl :TYPEVALSEP EQUALS quoted_or_null { 
if (!realloc_string(&p_alg_decl->attr_decl.typeval_sep,buf))
			yyerror("Out of memory");
	}

VALSEP_DECL : VALSEP EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->attr_decl.val_sep,buf)) 
			yyerror("Out of Memory");
	};

ATTSEXSSEP_DECL : ATTSEXSSEP EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->interface.atts_exs_sep,buf))
			yyerror("Out of Memory");
	}

nominal_decl : NOMINAL EQUALS quoted_or_null { 
		if (!realloc_string(&p_alg_decl->attr_decl.discrete_type,buf)) 
			yyerror("Out of Memory");
	};

INT_DECL : INT EQUALS quoted_or_null { 
		if (!realloc_string(&p_alg_decl->attr_decl.int_type,buf)) 
			yyerror("Out of Memory");
	};

CONT_DECL : CONT EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->attr_decl.cont_type,buf)) 
			yyerror("Out of Memory");
	};

UNKNOWN_DECL : UNKNOWN EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->attr_decl.val_unknown,buf)) 
			yyerror("Out of Memory");
	};


DONTCARE_DECL : DONTCARE EQUALS quoted_or_null { 
if (!realloc_string(&p_alg_decl->attr_decl.val_dont_care,buf)) 
			yyerror("Out of Memory");
	};

FIRSTCHARSET_DECL : FIRSTCHARSET EQUALS CHAR_SPECS ;

CHARSET_DECL : CHARSET EQUALS CHAR_SPECS ;
CHAR_SPECS : CHARSPEC CHAR_SPECS | ;

CHARSPEC : ALPHA
{
	yyerror("CHARSET = ALPHA\n");
};
CHARSPEC : FULL
{
	yyerror("CHARSET = FULL\n");
};
CHARSPEC : QUOTED 
{
	yyerror(buf);
};
 

EXVALSEP_DECL : EXVALSEP EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->ex_decl.ex_val_sep,buf)) 
		yyerror("Out of Memory");
};

EXAMPLESEP_DECL : EXAMPLESEP EQUALS quoted_or_null { if (!realloc_string(&p_alg_decl->ex_decl.example_sep,buf)) 
		yyerror("Out of Memory");
};

EXENAME_DECL : EXENAME EQUALS quoted_or_null { 
if (!realloc_string(&p_alg_decl->interface.exe_name,buf)) 
		yyerror("Out of Memory");
} NEWLINE;

quoted_or_null: QUOTED {$$ = QUOTED;} ;

quoted_or_null : {buf[0] = '\0'; $$ = 0;};

newlines : NEWLINE newlines | NEWLINE ; 

%%

#include "lex.yy.c"

yywrap()
{
	return 1;
} 
