/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       DL_main

		Contents:			the Data logging main program

		Services:

		main	- main program

		Internal functions:

		errmsg      - print error messages		
		print_usage - print a user interface syntax description

		Author:       Juho Rousu (1995,1996)

\*                                                                           */

#include "Alg_Database.h"
#include "IO_server.h"
#include "DL_routines.h"

#include <libgen.h>

char *TELA_call_path;

/* ========================================================================== *\
															INTERNAL FUNCTIONS
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
														Print error messages
\* -------------------------------------------------------------------------- */

void errmsg(char *msg)
{
	printf("%s.\n",msg);
}


/* -------------------------------------------------------------------------- *\
								Print a description of the user interface syntax
\* -------------------------------------------------------------------------- */

void print_usage()
{		
	printf("Usage: dl -a <algorithm name>");
		printf(" -i <example file> [-o <attr. file>] [-t <threshold>]\n");
		printf(
" -a <algorithm name>  : a learning algorithm name defined in TELA algorithm\n");
		printf(
"                     database.\n");
		printf(
" -i <example file>    : file containing the examples for which the attribute\n" );
		printf(
"                     declaration is to be inferred.\n");	
		printf(
" -o <attr. file>      : file where resulting attribute declaration is to be\n");
 		printf(
"                     written.\n");
		printf(
" -t <threshold>       : the number of unique values among the examples\n");
		printf(
"                     required to assign an attribute as 'integer'. Default\n");
		printf(
"                     value is 16.\n");
		return;
}



/* ========================================================================== *\
														EXTERNAL SERVICES
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
										Data logging main program
\* -------------------------------------------------------------------------- */

main(int argc, char *argv[])
{
	t_algorithms alg_ref;
	t_dl_node *p_dl_data;
	int threshold=16;
	int i,ex_cnt=0;
	char a_name[64],e_file[64],a_file[64],alg_db_file[256];

	a_name[0] = e_file[0] = a_file[0] = alg_db_file[0] = '\0';

	/* first, figure out, where we are called from */
	TELA_call_path = dirname(argv[0]);
	sprintf(alg_db_file,"%s/.TELA_dfd_db",TELA_call_path);


	if (argc < 5)
	{	
		print_usage();
		return;
	}

	i=1;
	while (i < argc-1)
	{
		if (strlen(argv[i]) < 2) 
		{
			print_usage(); return;
		}
		switch (argv[i][1])
		{
			case 'i': strcpy(e_file,argv[++i]); break;
			case 'o': strcpy(a_file,argv[++i]); break;
			case 'a': strcpy(a_name,argv[++i]); break;
			case 't': sscanf(argv[++i],"%d",&threshold);break;
			default: printf("Unknown option: %c\n",argv[i][1]);
		}
		i++;
	}

	if (!ADB_read_file(alg_db_file))	
	{
		printf("Algorithm database initialization did not succeed.\n");
		return;
	}

	ADB_get_alg_ref(a_name,&alg_ref);
	if (!DL_gen_attr_descr(alg_ref,e_file,threshold,&p_dl_data,&ex_cnt))
	{
		printf("Couldn't infer any description.\n");
	}
	else
	{
		int cl_set=0;

		cl_set = DL_set_class_attr(alg_ref);

		DL_display_inference_stats(p_dl_data,ex_cnt);

		if (strlen(a_file))
		{ 
			if (!IO_write_atts_file(a_file,alg_ref,cl_set))
			{
				printf("Couldn't write the attribute declaration\n");
			}
			else
			{
				printf("Attribute declaration written to file %s.\n",a_file);
			}
		}
	}
}
