/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       IO_scanner.h

		Contents:     Simple scanner routines for TELA I/O-server.							 

		Services:

		get_attr_token - get next token from attribute description	
		get_ex_token   - get next token from example description

		Author: Juho Rousu (1995,1996)
\*  									                                                        */

#include <string.h>

#include "IO_scanner.h"

#define DEBUG_

/* returns the next separator type */

char *eatwhite(char *src)
{
	while (*src == ' ') src++;

	return src;
}

t_separator_type get_attr_token(char *rest, int *len, char *nv_s, 
	char *tv_s,char *v_s,char *a_s, char *na_s, char *a_e_s, char *c_s)
{
	int i=0,comment_start = 0;
	char quote='\0';
	t_separator_type rv = ATT_SEP;

	if (*rest == '\"' || *rest == '\'') quote = *rest;

	while (rest[i] != '\0')
	{
		/* search for the closing quote */
		if (quote != '\0') 
		{
			if (rest[++i] != quote) 
				continue;
			else quote = '\0'; /* back to normal mode */
		}

		if (rest[i] == *v_s ? !strncmp(rest+i,v_s,strlen(v_s)) : 0)
		{
			rv = VAL_SEP;
			break;
		}
		else if (rest[i] == *nv_s ? !strncmp(rest+i,nv_s,strlen(nv_s)) : 0)
		{
			rv = NAME_VAL_SEP;
			break;
		}
		else if (rest[i] == *tv_s ? !strncmp(rest+i,tv_s,strlen(tv_s)) : 0)
		{
			rv = TYPE_VAL_SEP;
			break;
		}
		else if (rest[i] == *a_s ? !strncmp(rest+i,a_s,strlen(a_s)) : 0)
		{
			rv = ATT_SEP;
			break;
		}
		else if (rest[i] == *na_s ? !strncmp(rest+i,na_s,strlen(na_s)) : 0)
		{
			rv = NUMATT_SEP; /* note, that if a_s == na_s ATT_SEP is returned */
			break;
		}
		else if (rest[i] == *a_e_s ? !strncmp(rest+i,a_e_s,strlen(a_e_s)) : 0)
		{
			rv = ATTS_EXS_SEP;
			break;
		}
		else if (rest[i] == *c_s ? !strncmp(rest+i,c_s,strlen(c_s)): 0)
		{
			comment_start = i;
			i = strlen(rest)-1;
			rv = ATT_SEP;
		} 
		i++;
	}

	*len = comment_start ? comment_start : i; /* i characters from start are the potential 'identifier' */

	return rv;
}

int is_number(char c)
{
	if (c < '0' || c > '9') 
		return 0;
	else
		return 1;
}

int is_figure(char *c)
{
	float val;

	if (sscanf(c,"%f",&val)) return 1;	
}

#define NEXT_CH if ((tok_buf[i]=fgetc(exs_file))==EOF){rv = EXS_FOOTER;goto end;}
#define NEXT_i 	if (++i >= 511){ rv = EXS_FOOTER;goto end;}

char eat_until_sep_or_nl(FILE *fp,char evs,char es)
{
	char ch;

	do
	{
		if ((ch = fgetc(fp)) == EOF) break;
	}
	while ((ch != evs) && (ch != es) && (ch == ' '));

	return ch;
}

/* eat all spaces */ 
char eat_spaces_and_seps(FILE *fp,char evs, char es, char c_s)
{
	char ch;

	do
		if ((ch = fgetc(fp)) == EOF) 
			break;
		else if (ch == c_s)
		{
			while ((ch = fgetc(fp)) != '\n')
				if (ch == EOF) return ch;
		}
		
	while (ch == ' ' || ch == '\n' || ch == evs || 
		ch == es ||!isprint(ch));


	return ch;
}


t_separator_type get_ex_token(FILE *exs_file,char *tok_buf,char exval_sep,
	char ex_sep, char *footer, char comment_sep)
{
	char ch;
	int i= 0,j=0,rv;
	char quote = '\0';
	boolean in_number;

	if ((tok_buf[0] = eat_spaces_and_seps(exs_file,exval_sep,ex_sep,
		comment_sep)) == EOF)
	{
		rv = EXS_FOOTER;
		goto end;
	}
	else if (tok_buf[0] == footer[0]) 
	{
		while (i< (int)(strlen(footer)-1))
		{
			NEXT_i; NEXT_CH;
			if (footer[i] != tok_buf[i]) goto scan_on;
		}
		rv = EXS_FOOTER; goto end;
	}
	else if (tok_buf[i] == '\"' || tok_buf[i] == '\'') /* quoted string */
	{
		quote = tok_buf[i];
		do 
		{
			NEXT_i;	NEXT_CH;
		}
		while (tok_buf[i] != quote);
		NEXT_i; NEXT_CH;	/* ch after closing quote */
	}
	else if (is_number(tok_buf[i]) || tok_buf[i]=='-' || tok_buf[i]=='+') 
	/* the whole part of a number*/
	{
		if (!is_number(tok_buf[i])) /* a number must follow + or - */
		{
			NEXT_i; NEXT_CH;
			if (!is_number(tok_buf[i])) goto scan_on;
		}

		do
		{
			NEXT_i; NEXT_CH;
		}
		while (is_number(tok_buf[i]));

		if (tok_buf[i] == '.')                 /* decimal point */
		{
			NEXT_i; NEXT_CH; /* ch after decimal point */
		}	

		/* now, if this character is not a number, previously read point */
		/* could have been a separator */
		if (!is_number(tok_buf[i]))
		{
			if (tok_buf[i-1] == ex_sep)
			{
				i--;
				rv = EXAMPLE_SEP;
			}
			else if (tok_buf[i-1] == exval_sep)
			{
				i--;
				rv = EXVAL_SEP;
			}
		}
	}

	scan_on:

	do
	{		
		if (tok_buf[i]==ex_sep)
		{ 
			rv = EXAMPLE_SEP; break;
		}
		else if (tok_buf[i]==exval_sep || tok_buf[i]==' ' || tok_buf[i]=='\n')		
		{ 
			rv = EXVAL_SEP; 
			break;
		}
		else if (!isprint(tok_buf[i]))
		{
			rv = EXAMPLE_SEP;
			break;
		}
		else if (tok_buf[i] == comment_sep)
		{
			int comment_start = i;

			while (tok_buf[i] != '\n')
			{ 
				NEXT_i; NEXT_CH;
			}

			/* ok, end of line comment read */

			if (ex_sep != '\n') 
				rv = EXVAL_SEP;
			else
				rv = EXAMPLE_SEP;			

			i = comment_start;

			break;
		} 

		NEXT_i; NEXT_CH;
	}
	while (1);

	end:

	tok_buf[i] = '\0';

#ifdef DEBUG
	printf("*%s*\n",tok_buf);
#endif

	return rv;
}
/*
char eol_comment(FILE *file)
{
	char ch;

	do 
	{
		if ((ch=fgetc(exs_file)))==EOF) 
		{
			ch = '\0'; 
			break;
		}
	}
	while (ch != '\n')

	return ch; 
}
*/
