#ifndef H_TELA_IO_SERVER
#define H_TELA_IO_SERVER

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  IO_server 

		Contents:    TELA I/O-server module

		Description:  handles the TELA I/0 operations, namely, the reading 
									and writing of attributes, examples and classifiers (rules) 
									in algorithm specific formats to/from filesystem. 
		Services:

		General
		=======

		IO_initialize - initializes the I/0-server
		IO_set_round - tells the I/O-server which DO-round is in question (0 - none)
		IO_get_expanded_name - replaces '!' with current round index in the filename

	  Input operations
	  ================

	  IO_read_atts - read attribute description in given data format  
		IO_read_exs  - read examples in given data format
		IO_read_rule - read a rule

		IO_read_atts_file - read an attribute file
		IO_read_exs_file  - read an example file
  	IO_read_both_file - read a combined (atts&exs) file
  	IO_read_rule_file - read a rule file (with an attribute description) 

  	Output operations
  	=================

  	IO_write_atts - write attribute description in given data format  
  	IO_write_exs  - write examples in given data format
  	IO_write_rule - write a rule

  	IO_write_atts_file - write an attribute file
  	IO_write_exs_file   - write an example file
  	IO_write_both_file  - write a combined (atts&exs) file
  	IO_write_rule_file  - write a rule file (with an attribute description) 

		Author: Tommi Timonen (1992) Juho Rousu (1995,1996)

\*                                                                            */

#include <stdio.h>
#include "tela.h"

char *IO_get_expanded_name(char*);
void  IO_initialize(void);
void  IO_set_round(int round);

int IO_write_atts(FILE *a_file,t_algorithms alg,int classified);
int IO_write_exs (FILE *a_file,t_algorithms alg,t_ex_mask *p_mask,
		int is_cl, int *not_written);
int IO_write_rule(FILE *rule_file,t_algorithms alg,char *comment);

int IO_write_atts_file(char *fname,t_algorithms alg,int classified);
int IO_write_exs_file (char *fname,t_algorithms alg,t_ex_mask *p_msk,
	int is_cl, int *not_written);
int IO_write_both_file(char *fname,t_algorithms alg,t_ex_mask *p_msk,
	int is_cl, int *not_written);
int IO_write_rule_file(char *filename, t_algorithms alg,char *comment);

int IO_read_atts(FILE *atts_file,t_algorithms alg,t_filetype f_type);
int IO_read_exs(FILE *atts_file, t_algorithms alg,t_filetype f_type);
int IO_read_rule(FILE *rule_file,t_algorithms alg);

int IO_read_atts_file(char *filename,t_algorithms alg);
int IO_read_exs_file(char *filename, t_algorithms alg);
int IO_read_rule_file(char *filename, t_algorithms alg);
int IO_read_both_file(char *filename, t_algorithms alg);

#endif
