/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  TESLA_IO_cmds (part of TESLA command server)

		Contents: File system I/O related commands of TESLA language

		Services:

		READ_ATTS_command - reads an attribute description file
		READ_EXS_command  - reads an example vector file
		READ_BOTH_command - reads a combined atts/exs -file 
		READ_command      - a wrapper for above READ variants
		READ_RULE_command - reads a rule file

		WRITE_ATTS_command - writes an attribute description file
		WRITE_EXS_command  - an example vector file
		WRITE_BOTH_command - a combined atts/exs -file 
		WRITE_command      - a wrapper for above WRITE variants
		WRITE_RULE_command - writes a rule file

		Author: Margareta Bjorksten, Marko Nyman (1992), 
						Juho Rousu (1995,1996)
\*                                                                             */

#include <stdio.h>

#include "tela.h"
#include "WS_Database.h"
#include "IO_server.h"

#define PAR_W params.param.l_write_2

#define PAR_R params.param.l_read_2


/* -------------------------------------------------------------------------- *\
                            Read attributes
\* -------------------------------------------------------------------------- */

int READ_ATTS_command (t_command p) 
{ 
  
  FILE *fp;
  int palautus =0;

  char *uusi_nimi;
	char *tod_nimi; 

	char *FileName;

  WS_dispose_dbase();

  FileName = p.param.l_read_2.a_filename;
 
  uusi_nimi = IO_get_expanded_name(p.param.l_read_2.a_filename);

	if (!(fp = fopen(uusi_nimi,"r")))
	{
		char msg[256];
		sprintf(msg,"No such file: %s",uusi_nimi);
		errmsg(msg);
		return EXECUTION_FAILED;
	}
	else fclose(fp);
     
	if (IO_read_atts_file(uusi_nimi,p.param.l_read_2.alg))
	{
		printf("Attribute description ");
		printf("read from file %s.\n",uusi_nimi);
		return EXECUTION_OK;
	}
	else
		return EXECUTION_FAILED;

} 


/* -------------------------------------------------------------------------- *\
							              Read Examples
\* -------------------------------------------------------------------------- */

int READ_EXS_command(t_command p) 
{

  FILE * fp;
  char *uusi_nimi, *tod_nimi; 
  char *FileName;
  int palautus =0;
  /* PoistaEsimerkit(); */


  if (!WS_get_attr_cnt ()) 
	{
		errmsg("No attribute description in memory");
		return EXECUTION_FAILED;
  }

    
  FileName = p.param.l_read_2.e_filename;

  uusi_nimi = IO_get_expanded_name(p.param.l_read_2.e_filename);

	if (!(fp = fopen(uusi_nimi,"r")))
	{
		char msg[256];
		sprintf(msg,"No such file: %s",uusi_nimi);
		errmsg(msg);
		return EXECUTION_FAILED;
	}
	else fclose(fp);
	
	if (IO_read_exs_file(uusi_nimi,p.param.l_read_2.alg))
	{
		printf("%d examples read from file %s.\n",WS_get_ex_cnt(),uusi_nimi);
		return EXECUTION_OK;
	}
	else
		return EXECUTION_FAILED;

} 


/* -------------------------------------------------------------------------- *\
					Read attributes and examples for ma combined file
\* -------------------------------------------------------------------------- */

int READ_BOTH_command (t_command p) {

  static int palautus = 0;

  FILE *fp;
  char *uusi_nimi, *tod_nimi; 
	t_algorithms alg = p.param.l_read_2.alg;

  WS_dispose_dbase ();

  uusi_nimi = IO_get_expanded_name(p.param.l_read_2.a_filename);

	if (!(fp = fopen(uusi_nimi,"r")))
	{
		char msg[256];
		sprintf(msg,"No such file: %s",uusi_nimi);
		errmsg(msg);
		return EXECUTION_FAILED;
	}
	else fclose(fp);

	if (IO_read_both_file(uusi_nimi,alg))
	{
		printf("Attribute description and %d examples ",WS_get_ex_cnt());
		printf("read from file %s.\n",uusi_nimi);
		return EXECUTION_OK;
	}
	else
		return EXECUTION_FAILED;
         
}


/* -------------------------------------------------------------------------- *\
					           Read ATTS, EXS, BOTH or RULE
\* -------------------------------------------------------------------------- */

int READ_command (t_command p) {
	
  switch (p.cmd_id) {
    case K_READ_ATTS:
      return READ_ATTS_command (p);
    case K_READ_EXS:
      return READ_EXS_command (p);
    case K_READ_BOTH:
      return READ_BOTH_command (p);
    case K_READ_ATTS_AND_EXS:
      if (READ_ATTS_command (p) != NO_ERROR)
        return NON_FATAL_ERROR;
      else
        return READ_EXS_command (p);
		case K_READ_RULE: 
			return READ_RULE_command (p);
    default:
      printf ("Invalid command id.\n");
      return NON_FATAL_ERROR;
  }
}


/* -------------------------------------------------------------------------- *\
				                      Write Attributes
\* -------------------------------------------------------------------------- */

int WRITE_ATTS_command (t_command p) 
{
	t_algorithms alg = p.param.l_write_2.alg;
	char *name; 
	int palautus =0 ;
	
  /* Annetaan virheilmoitus jos muistissa ei ole attribuutteja */
  if (!WS_get_attr_cnt ()) 
	{
		errmsg("No attribute description in memory");
		return EXECUTION_FAILED;
  }

	name = IO_get_expanded_name(p.param.l_write_2.a_filename);

	if (IO_write_atts_file(name,alg,WS_is_class_attr()))
	{
		printf ("Attributes written to file %s.\n", name);
		return EXECUTION_OK;
	}
	else
  	return EXECUTION_FAILED;                    
} 


/* -------------------------------------------------------------------------- *\
											       Write Examples
\* -------------------------------------------------------------------------- */

int WRITE_EXS_command (t_command p) {

  int not_written = 0;
	char *filename;
	t_algorithms alg = p.param.l_write_2.alg;

	filename = IO_get_expanded_name(p.param.l_write_2.e_filename); 

  if (!WS_get_ex_cnt ()) 	
	{
		errmsg("No examples in memory");
		return EXECUTION_FAILED;
  }


	if (IO_write_exs_file(filename,alg,0,WS_is_class_attr(),&not_written))
	{
		printf ("%d examples written to file %s",
 			WS_get_ex_cnt()-not_written,filename);
		if (not_written)
			printf (",%d examples discarded.\n",not_written);
		else
			printf(".\n");

		return EXECUTION_OK;
	}
	else
  	return EXECUTION_FAILED;                    
} 


/* -------------------------------------------------------------------------- *\
							Write attributes and examples in a combined file
\* -------------------------------------------------------------------------- */

int WRITE_BOTH_command(t_command p) {

  int not_written =0;
	char *file_name;
	t_algorithms alg = p.param.l_write_2.alg;

	file_name = IO_get_expanded_name(p.param.l_write_2.a_filename);

  if (!WS_get_attr_cnt ()) 
	{
		errmsg("No attribute description in memory");
		return EXECUTION_FAILED;
  }

 	if (!WS_get_ex_cnt ()) 	
	{
		errmsg("No examples in memory");
		return EXECUTION_FAILED;
  }

	if (IO_write_both_file(file_name,alg,0,WS_is_class_attr(),&not_written))
	{
		printf ("Attributes and %d examples written to file %s", 
  	  WS_get_ex_cnt()-not_written,p.param.l_write_2.a_filename);
		if (not_written)
			printf(", %d examples discarded.\n",not_written);
		else
			printf(".\n");

		return EXECUTION_OK;
 	}
	else
 		return EXECUTION_FAILED;                    

} 


/* -------------------------------------------------------------------------- *\
					             Write ATTS, EXS, BOTH or RULE
\* -------------------------------------------------------------------------- */

int WRITE_command (t_command p) {
	
  switch (p.cmd_id) {
    case K_WRITE_ATTS:
      return WRITE_ATTS_command (p);
    case K_WRITE_EXS:
      return WRITE_EXS_command (p);
    case K_WRITE_BOTH:
      return WRITE_BOTH_command (p);
    case K_WRITE_ATTS_AND_EXS:
			if (!strcmp(p.param.l_write_1.a_filename,
				p.param.l_write_1.e_filename))
			{
				errmsg(
				"Different filenames should be provided for attributes and examples");
				return EXECUTION_FAILED;
			}
      if (WRITE_ATTS_command (p) != NO_ERROR)
        return NON_FATAL_ERROR;
      else
        return WRITE_EXS_command (p);
		case K_WRITE_RULE:
			return WRITE_RULE_command (p);
    default:
      printf ("Invalid command id.\n");
      return NON_FATAL_ERROR;
  }
}


/* -------------------------------------------------------------------------- *\
							                Read a Rule
\* -------------------------------------------------------------------------- */

int READ_RULE_command(t_command params)
{
  
  if (IO_read_rule_file(IO_get_expanded_name(PAR_R.r_filename),
		PAR_R.alg))
	{
		printf("Classifier read from file %s.\n",PAR_R.r_filename);
		WS_set_active_classifier(PAR_R.alg);
		return EXECUTION_OK;
	}
	else
  {
		errmsg("Couldn't read the rule");
    return EXECUTION_FAILED;
  }
}


/* -------------------------------------------------------------------------- *\
						                 Write a Rule 
\* -------------------------------------------------------------------------- */

int WRITE_RULE_command (t_command params)
{
	
  FILE *tiedosto;
	t_algorithms alg;
  int virhekoodi = NO_ERROR;

	if (!WS_get_active_classifier(&alg))
	{
		errmsg("No active classifier");
		return EXECUTION_FAILED;
	}
	else if (alg != PAR_W.alg)
	{
		errmsg("Classifier was generated by another algorithm");
		return EXECUTION_FAILED;
	}
  
  if (IO_write_rule_file(IO_get_expanded_name(PAR_W.r_filename),
		PAR_W.alg,PAR_W.comment)) 
	{
		printf("Classifier written to file %s.\n",PAR_W.r_filename);
		return EXECUTION_OK;
  }

	return EXECUTION_FAILED;

}
