#ifndef H_TELA_TESLA_DB_CMDS
#define H_TELA_TESLA_DB_CMDS

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:  TESLA_db_cmds (part of TESLA command server)

		Contents: Working Set Manipulation and Supporting commands of 
			TESLA language

		Services:

		Working Set Manipulation commands
		=================================

		CLASS_command		    - sets the class attribute
		DIVIDE_command      - partitions the active example set
		EXCLUDE_command		  - excludes specified attributes
		INCLUDE_command		  - includes (irrelevant) attributes
		MAP_command			    - maps and attributes domain to a new one
		NOISE_command		    - adds random noise
		RANDOMIZE_command	  - adds random noise
		SELECT_command      - chooses a subset of examples as the new example set 
		STAT_command		    - displays basic statistics of the dataset

		Supporting commands
		===================

		SET_command     - sets global variables, initializes the RND generator
		DELETE_command  - deletes a file
		INIT_command    - initializes the working set database
		BACKUP_command  - backups specified components of the working set
		RESTORE_command - restores a previously backed up working set portions
		QUIT_command   - exits TELA after cleanup

		Author: Ilkka Hyvarinen (1992,1993,1994), Juho Rousu (1995,1996)

\*                                                                            */

#include "tela.h"

int CLASS_command		(t_command params);
int DIVIDE_command		(t_command params);
int EXCLUDE_command		(t_command params);
int INCLUDE_command		(t_command params);
int MAP_command			(t_command params);
int NOISE_command		(t_command params);
int RANDOMIZE_command		(t_command params);
int SELECT_command (t_command params);
int STAT_command		(t_command params);

int SET_command (t_command params);
int DELETE_command (t_command params);
int INIT_command (t_command params);
int BACKUP_command(t_command cmd);
int RESTORE_command(t_command cmd);
int QUIT_command (t_command cmd);

#endif
