#ifndef H_TESLA_BASIC
#define H_TESLA_BASIC

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       TESLA_basic

		Contents:     Parsers for some common language elements in TESLA 

		Services:

		TESLA_word             - parse an unquoted string
		TESLA_string           - parse a quoted string, strip quotes
		TESLA_qstring		       - parse a quotes string, keep the quotes
		TESLA_nat_val          - parse s natural number
		TESLA_cont_val         - parse a floating point number
		TESLA_int_val          - parse an integer value
		TESLA_nominal_val      - parse a nominal value: character, word or string
		TESLA_attribute_value  - parse a value that is legal for an attribute
		TESLA_percent_val      - parse an percentage
		TESLA_relation         - parse an arithmetic relation
		TESLA_attribute_name   - parse an attribute name
		TESLA_algorithm        - parse an algorithm name
		TESLA_filename         - parse a filename
		TESLA_param_value      - parse a parameter value
		TESLA_symbol           - parse a reserved word
		TESLA_datatype         - parse a data (file) type
		TESLA_wordtoken_cmp	   - compare a string to current (word) token
		TESLA_is_separator     - is current token given separator character
		TESLA_is_reserved_mark - is given character a reserved separator
		TESLA_is_reserved_word - is current token given reserved word 

		Author: Juho Rousu (1995,1996)

\*                                                                            */


#include "tela.h"

#include "cseed.h"
#include "seed_lexline.h"

#include "UI_parsers.h"

#define PARSE_OK 0
#define PARSE_FAILED 1
#define PARSE_CANCELLED 2
#define EXECUTION_FAILED 3
#define USER_QUIT 4

/* Parsers of basic building blocks of TESLA language 				*/
/* these return a non-zero if parse didn't succeed. In that case, 	*/
/* the arguments can point to rubbish.												*/

int TESLA_word(char **wd);
int TESLA_string(char **str);
int TESLA_qstring(char **str);
int TESLA_nat_val(unsigned *p_n);
int TESLA_cont_val(float *p_f);
int TESLA_int_val(int *p_n);
int TESLA_nominal_val(char **p_val);
int TESLA_attribute_value(char **pp_val);
int TESLA_percent_val(float *p_n);
int TESLA_relation(t_relation *p_rel);
int TESLA_attribute_name(char **name);
int TESLA_algorithm(t_algorithms *p_alg_ref);
int parse_algorithm(t_algorithms *p_alg_ref);
int TESLA_filename(t_algorithms alg, t_filetype f_type, char **fn);
int TESLA_param_value(char **pp_val);
int TESLA_symbol(int *id);
int TESLA_datatype(t_filetype *p_id);
int TESLA_wordtoken_cmp(char* str);
int TESLA_is_separator(char c);
int TESLA_is_reserved_mark(char c);
int TESLA_is_reserved_word(int id);

#endif
