/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       UI_chk_pnt.h

		Contents:     Checkpoint maintenance routines for SEED interface
									 
		Services:

		new_chk_pnt         - add a new checkpoint
		free_chk_pnt        - free a checkpoint
		free_cmd_list		    - free a list of commands
		cmd_type					  - return the type of the command
		add_cmd_list			  - add a command at the end of the  command list
		*cmd_name						- return the name of the command
		insert_to_cmd_list	- insert a command at the beginning of the command list

		Author: Juho Rousu (1995,1996)
\*  																																					*/
   
#include "UI_chk_pnt.h"
#include "UI_parsers.h"


/* -------------------------------------------------------------------------- *\
   											add a new checkpoint node
\* -------------------------------------------------------------------------- */

int new_chk_pnt(t_chk_pnt **pp_chk_pnt,t_cmd_type c_type,lex_positiontype lp)
{
	t_chk_pnt *p_new_chk_pnt;
	if (!(p_new_chk_pnt = (t_chk_pnt *)malloc(sizeof(t_chk_pnt))))
	{
		errmsg("Out of memory");
		return 1;	
	}
	memset(p_new_chk_pnt,0,sizeof(t_chk_pnt));
	p_new_chk_pnt->cp_type = c_type;
	p_new_chk_pnt->lex_pos = lp;
	p_new_chk_pnt->p_cmds = 0;
	p_new_chk_pnt->p_rev_cmds = 0;
	p_new_chk_pnt->unsafe = 0;
	p_new_chk_pnt->ae_backup = 0;
	p_new_chk_pnt->cf_backup = 0;

	p_new_chk_pnt->p_prev = *pp_chk_pnt;
	*pp_chk_pnt =  p_new_chk_pnt;
	return 0;
}


/* -------------------------------------------------------------------------- *\
			free all the memory allocated to the checkpoint structures
\* -------------------------------------------------------------------------- */

void free_chk_pnt(t_chk_pnt **pp_chk)
{
	free_cmd_list(&(**pp_chk).p_rev_cmds);
	free(*pp_chk);
	*pp_chk = 0;
}


/* -------------------------------------------------------------------------- *\
												free a list of commands
\* -------------------------------------------------------------------------- */

void free_cmd_list(t_command **pp_cmd)
{
	t_command *p_next,*p_delete;

	if (!(*pp_cmd)) return;

	/* ok, at least one to free, leave it unremoved for a moment */
	p_next = (*pp_cmd)->next;	
	p_delete = p_next;

	/* delete the commands from the ring */
	while (p_next != *pp_cmd && p_next) 
	{	
		p_next = p_next->next;
		free_cmd(p_delete,1); /* this should work,since the prev-fields are empty */
		p_delete = p_next;
	}

	/* and finally, the last node */
	free(*pp_cmd);
	return;
}


/* -------------------------------------------------------------------------- *\
										return the type of the command
\* -------------------------------------------------------------------------- */

t_cmd_type cmd_type(t_command *p_cmd)
{
	switch (p_cmd->cmd_id)
	{
  	case K_READ_ATTS: 
		case K_READ_EXS: 
		case K_READ_ATTS_AND_EXS: 
		case K_READ_BOTH: 
		case K_READ_RULE:
   	case K_DIVIDE:
  	case K_NOISE:
  	case K_RANDOMIZE:
  	case K_INCLUDE:
  	case K_MAP:	

		case K_SELECT:  	
		case K_EXCLUDE:
		case K_EXCLUDE_ALL_EXCEPT:
			return C_WSMANIP;
		break;
			
		case K_DO: /* K_DO deserves a checkpoint for it's own, otherwise it's
			too tedious */ 

  	case K_RUN:
		case K_CV:
  	case K_TEST:
 		case K_WRITE_ATTS: 
		case K_WRITE_EXS: 
		case K_WRITE_ATTS_AND_EXS: 
		case K_WRITE_BOTH: 
		case K_WRITE_RULE:
  	case K_CLASS:
		case K_STAT: 
		case K_STAT_ALL:
 		case K_END:
			return C_PROCESSING;
		break;
		default:
			return C_WSMANIP;
	}
}


/* -------------------------------------------------------------------------- *\
											return the name of the command
\* -------------------------------------------------------------------------- */

char *cmd_name(t_command *p_cmd)
{
	char name[256];
	switch (p_cmd->cmd_id)
	{
		case K_NULL:  sprintf(name,"K_NULL"); break;
		case K_INIT:  sprintf(name,"K_INIT"); break;
		case K_BACKUP_ATTS_EXS:  sprintf(name,"K_BACKUP_ATTS_EXS"); break;
		case K_BACKUP_RULE:  sprintf(name,"K_BACKUP_RULE"); break;
		case K_RESTORE_ATTS_EXS:  sprintf(name,"K_RESTORE_ATTS_EXS"); break;
		case K_RESTORE_RULE:  sprintf(name,"K_RESTORE_RULE"); break;
		case K_BEGIN:  sprintf(name,"K_BEGIN"); break;
		case K_DELETE:  sprintf(name,"K_DELETE"); break;

  	case K_READ_ATTS: 
		case K_READ_EXS: 
		case K_READ_ATTS_AND_EXS:  
		case K_READ_BOTH: sprintf(name,"K_READ_AEB"); break;
		case K_READ_RULE: sprintf(name,"K_READ_RULE"); break;
   	case K_DIVIDE: sprintf(name, "K_DIVIDE"); break;
  	case K_NOISE: sprintf(name, "K_NOISE"); break;
  	case K_RANDOMIZE: sprintf(name, "K_RANDOMIZE"); break;
  	case K_INCLUDE: sprintf(name, "K_INCLUDE"); break;
  	case K_MAP:	 sprintf(name, "K_MAP"); break;
		case K_DO:  sprintf(name, "K_DO"); break;

		case K_SELECT: sprintf(name, "K_SELECT"); break; 	
		case K_EXCLUDE: 		
		case K_EXCLUDE_ALL_EXCEPT:sprintf(name, "K_EXCLUDE"); break;

  	case K_RUN: sprintf(name, "K_RUN"); break;
		case K_CV: sprintf(name, "K_CV"); break;
  	case K_TEST: sprintf(name, "K_TEST"); break;
 		case K_WRITE_ATTS:  
		case K_WRITE_EXS: 
		case K_WRITE_ATTS_AND_EXS: 
		case K_WRITE_BOTH: sprintf(name, "K_WRITE_AEB"); break;
		case K_WRITE_RULE: sprintf(name, "K_WRITE_RULE"); break;
  	case K_CLASS: sprintf(name, "K_CLASS"); break;
		case K_STAT:
		case K_STAT_ALL: sprintf(name, "K_STAT"); break;
 		case K_END: sprintf(name, "K_END"); break;
		default: sprintf(name, "UNKNOWN"); break;
	}
	return name;
}


/* -------------------------------------------------------------------------- *\
							add a command to the end of a command list
\* -------------------------------------------------------------------------- */

void add_cmd_list(t_command **pp_list1,t_command *p_list2)
{
	if (!p_list2) return;
	if (*pp_list1)
	{
		t_command *p_tmp = (**pp_list1).next;
		(**pp_list1).next = p_list2->next;
		p_list2->next = p_tmp;

	}
	else p_list2->next = p_list2; /* make it a ring */
	*pp_list1 = p_list2;
}


/* -------------------------------------------------------------------------- *\
						insert a command at the beginning of the command list
\* -------------------------------------------------------------------------- */

void insert_to_cmd_list(t_command **pp_list1,t_command *p_list2)
{
	if (!p_list2) return;
	if (!(*pp_list1))
		*pp_list1 = p_list2;
	else
	{
		t_command *p_tmp = (**pp_list1).next;
		(**pp_list1).next = p_list2->next;
		p_list2->next = p_tmp;
	}
}
