/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       UI_parsers

		Contents:     Parsers of TESLA commands for SEED interface  

		Services:

		

		Author: Ilkka Hyvrinen (1992,1993,1994) Juho Rousu (1995,1996)

\*                                                                            */

#include <stdio.h>
#include <string.h>

#include "cseed.h"
#include "seed_lexline.h"

#include "TESLA_statistics.h"
#include "Alg_Database.h"
#include "IO_server.h"
#include "UI_parsers.h"

#define private static

 
/* -------------------------------------------------------------------------- *\
												Parse an algorithm name 
\* -------------------------------------------------------------------------- */

int parse_algorithm(t_algorithms *p_alg_ref)
{
  lex_next ();           /* Algorithm name */
  if (lex_cancel ())
    return PARSE_CANCELLED;

	if (TESLA_algorithm(p_alg_ref) != PARSE_OK)
	{
		errmsg("No such algorithm");
		return PARSE_FAILED;
	}

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
   			 		Output an error message on the screen
\* -------------------------------------------------------------------------- */

void errmsg (char *msg) 
{
	
  printf ("%s.\n", msg);
  lex_error ();
}


/* -------------------------------------------------------------------------- *\
							Parse an into - clause
\* -------------------------------------------------------------------------- */

int parse_into (char **into_fname) 
{
	FILE *fp;	

	lex_positiontype here;
	
  *into_fname = NULL;


  /* INTO or end of cmd */
  if (TESLA_is_reserved_word(into_sym)) 
	{
		/* File name */
    lex_next (); if (lex_cancel()) return 0;

		if (TESLA_filename(0,FT_STATS,into_fname) == PARSE_FAILED)
		{
			errmsg("Filename expected");
			return PARSE_CANCELLED;
		}
  
    lex_next ();    if (lex_cancel ()) return 0; /* semicolon */
  } 
	else if (!TESLA_is_separator(';')) 
	{
    errmsg ("INTO or semicolon expected");
    return 0;
  }
  return 1;
}


/* -------------------------------------------------------------------------- *\
												Parse a semicolon
\* -------------------------------------------------------------------------- */    
void parse_semicolon (void) 
{

  lex_get (); if (lex_cancel ()) return;
  if (!lex_ismark (';')) 
	{
    errmsg ("Semicolon expected");
    return;
  } 
  lex_null();
}

 
/* -------------------------------------------------------------------------- *\
 									Allocate memory for one command node 
\* -------------------------------------------------------------------------- */

t_command *alloc_cmd (void) 
{
  t_command      *cmd=0;

  cmd = (t_command *) malloc (sizeof (t_command));
  if (cmd != NULL)
	{
    memset (cmd, 0, sizeof (t_command));
		cmd->next = 0;
		cmd->prev = 0;
	}
	else
	{
		errmsg("Memory allocation failed: alloc_cmd");
	}
  return cmd;
}


/* -------------------------------------------------------------------------- *\
								Free the value list of INCLUDE-command
\* -------------------------------------------------------------------------- */

void free_incl_val_list (struct t_incl_val_node *val_list) 
{

  if (val_list != NULL) 
	{
    free_incl_val_list (val_list->next);
    free (val_list);
  }
}

/* -------------------------------------------------------------------------- *\
							Free the attribute list of INCLUDE-command
\* -------------------------------------------------------------------------- */

free_incl_list (t_include_1 *incl_list) 
{
  if (incl_list != NULL) 
	{
    free_incl_list (incl_list->next);
    free_incl_val_list (incl_list->val_list_head.next);
    free (incl_list);
  }
}

/*--------------------------------------------------------------------------*\
 		Remove the most recently allocated command node and all its 
 	  related dynamically allocated structures from memory
\*--------------------------------------------------------------------------*/

t_command *free_cmd (t_command *tree, int optimized) 
{	
  struct t_exclude_node *ep;
  t_snode		*sp;
  t_command	     	*p;
  
  p = tree->prev;
  switch (tree->cmd_id) {
    case K_CLASS:
      if (tree->param.l_class_1.attr_name != NULL)
        free (tree->param.l_class_1.attr_name);
      break;
    case K_EXCLUDE: 
    case K_EXCLUDE_ALL_EXCEPT:
      ep = tree->param.l_exclude_1.first;
      while (ep != NULL) {
        tree->param.l_exclude_1.first = ep->next;
        if (ep->attr_name != NULL)
          free (ep->attr_name); 
        free (ep);
        ep = tree->param.l_exclude_1.first;
      }
      break;
    case K_INCLUDE:
      if (optimized) {
/*        if (tree->param.l_include_2.attr_name != NULL)
          free (tree->param.l_include_2.attr_name);
        if (tree->param.l_include_2.attribute_table != NULL)
          WS_free_attr (tree->param.l_include_2.attribute_table); */
      } else {
        if (tree->param.l_include_1.attr_name != NULL)
          free (tree->param.l_include_1.attr_name);
        free_incl_val_list (tree->param.l_include_1.val_list_head.next);
        free_incl_list (tree->param.l_include_1.next);
      }
      break;
    case K_MAP:
      if (optimized) {

      	t_map_node *mp;

      	if (tree->param.l_map_2.attr_name != NULL)
      	  free (tree->param.l_map_2.attr_name);
      	if (tree->param.l_map_2.else_val != NULL)
      	  free (tree->param.l_map_2.else_val);
      	if (tree->param.l_map_2.dont_know_val != NULL)
      	  free (tree->param.l_map_2.dont_know_val);
      	if (tree->param.l_map_2.dont_care_val != NULL)
      	  free (tree->param.l_map_2.dont_care_val);
      	if (tree->param.l_map_2.attribute_table != NULL)
      	  WS_free_attr (tree->param.l_map_2.attribute_table);
        while (tree->param.l_map_2.first != NULL) {
	  mp = tree->param.l_map_2.first;
	  tree->param.l_map_2.first = mp->next;
	  if (mp->new_val != NULL)
	    free (mp->new_val);
	  free (mp);
        }
      } else {

      	t_map_node     *mp;
      	t_map_node_oth *omp;

      	if (tree->param.l_map_1.attr_name != NULL)
      	  free (tree->param.l_map_1.attr_name);
      	if (tree->param.l_map_1.else_val != NULL)
      	  free (tree->param.l_map_1.else_val);
        while (tree->param.l_map_1.first != NULL) {
	  mp = tree->param.l_map_1.first;
	  tree->param.l_map_1.first = mp->next;
	  if (mp->new_val != NULL)
	    free (mp->new_val);
	  free (mp);
        }
        while (tree->param.l_map_1.others != NULL) {
	  omp = tree->param.l_map_1.others;
	  tree->param.l_map_1.others = omp->next;
	  if (omp->attr_val != NULL)
	    free (omp->attr_val);
	  free (omp);
        }  /* while */
      }  /* else of if (optimized) */
      break;
    case K_READ_ATTS:
    case K_READ_BOTH:
      if (tree->param.l_read_1.a_filename != NULL)
        free (tree->param.l_read_1.a_filename);
      break;
    case K_READ_ATTS_AND_EXS:
      if (tree->param.l_read_1.a_filename != NULL)
        free (tree->param.l_read_1.a_filename); /* Note lack of break! */
    case K_READ_EXS:
      if (tree->param.l_read_1.e_filename != NULL)
        free (tree->param.l_read_1.e_filename);
      break;
    case K_READ_RULE:
      break;
    case K_STAT: 
    case K_STAT_ALL:
      sp = tree->param.l_stat_1.first;
      while (sp != NULL) {
        tree->param.l_stat_1.first = sp->next;
        if (sp->attr_name != NULL)
          free (sp->attr_name); 
        free (sp);
        sp = tree->param.l_stat_1.first;
      }
      if (tree->param.l_stat_1.stat_file_name != NULL)
        free (tree->param.l_stat_1.stat_file_name);
      if (tree->param.l_stat_1.comment != NULL)
        free (tree->param.l_stat_1.comment);
      break;
    case K_WRITE_ATTS:
    case K_WRITE_BOTH:
      if (tree->param.l_write_1.a_filename != NULL)
        free (tree->param.l_write_1.a_filename);
        break;
    case K_WRITE_ATTS_AND_EXS:
      if (tree->param.l_write_1.a_filename != NULL)
        free (tree->param.l_write_1.a_filename); /* Note lack of break! */
    case K_WRITE_EXS:
      if (tree->param.l_write_1.e_filename != NULL)
        free (tree->param.l_write_1.e_filename);
      break;
    case K_WRITE_RULE:
      break;
    default:
      break;
  }
  free (tree);
  
  return p;
  /* Laajennettava ! */
}


/* -------------------------------------------------------------------------- *\
		Confirm the overwriting of given file (if it exists) from the user
\* -------------------------------------------------------------------------- */

int confirm_overwrite(char *file_name,int *ask_again)
{
	char *tmp_name;
	FILE *fp;
	lex_positiontype here;

	*ask_again = 1;

	IO_set_round(1);
	tmp_name = IO_get_expanded_name(file_name);

	if (fp = fopen(tmp_name,"r"))
	{
		fclose(fp);

		printf("File %s exists already. Overwrite?",tmp_name);

		lex_store(&here);

		do
		{
			lex_restore(here);

			lex_next();

			if (!TESLA_wordtoken_cmp("yes") || !TESLA_wordtoken_cmp("YES"))
				lex_exit();
			else if (TESLA_is_reserved_word(all_sym))
			{
				*ask_again = 0;
				lex_exit();
			}
			else if	(!TESLA_wordtoken_cmp("NO") || !TESLA_wordtoken_cmp("no"))
			{
				errmsg("Please give another filename");
				return PARSE_CANCELLED;
			}
			else 
			{
				errmsg("Please answer 'yes','no' or 'all'");
				lex_restore(here);
			}
		}
		while (!lex_done());

		return PARSE_OK;
	}
}


/* -------------------------------------------------------------------------- *\
													Parse a class command
\* -------------------------------------------------------------------------- */

int parse_class (t_command **next_cmd)
{ 
  char           *token;
      
  lex_next (); if (lex_cancel ()) return PARSE_FAILED;

  *next_cmd = alloc_cmd ();
  if (*next_cmd == NULL) {
    errmsg ("Out of memory");
    return PARSE_FAILED; 
  }

  (**next_cmd).cmd_id = K_CLASS;
	(**next_cmd).param.l_class_1.attr_spec = A_NORMAL;

  if (TESLA_attribute_name(&(**next_cmd).param.l_class_1.attr_name) != PARSE_OK)
		if (TESLA_is_reserved_word(none_sym))
		{
			(**next_cmd).param.l_class_1.attr_spec = A_NONE;
		}
		else goto cancel;
        
	lex_null();	
	parse_semicolon();
	if (lex_cancel()) goto cancel;

	return PARSE_OK;

	cancel:
	
	free_cmd(*next_cmd,0);  
  return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
      					Add a node to the list of DIVIDE filenames
\* -------------------------------------------------------------------------- */

int expand_divide_list (t_div_node **curr_node) 
{

  t_div_node *p;
  
  p = (t_div_node *) malloc (sizeof (t_div_node));
  if (p != NULL) {
    p->next = NULL; p->f_name = NULL;
    p->handle = NULL; p->part = 0.0;

    if (*curr_node == NULL)
      *curr_node = p;
    else {
      (*curr_node)->next = p; *curr_node = p;
    }
  }
  return (p != NULL);
}


/* -------------------------------------------------------------------------- *\
														Parse a DIVIDE command
\* -------------------------------------------------------------------------- */

int parse_divide (t_command **next_cmd) 
{

  t_command	*wcmd;
	int ask;
  char		*token;
  t_div_node	*curr_node=0, **nextptr, **wptr,*p_div_node;
  float		sum = 0;      /* sum of the parts */

  wcmd = alloc_cmd ();
  if (wcmd == NULL) {
    errmsg ("Out of memory");
    return 1;
  }

  wcmd->cmd_id = K_DIVIDE;

  wptr = nextptr = &wcmd->param.l_divide_1.first;

  do 
	{
    float f;
    lex_next ();  if (lex_cancel ()) goto cancel;

    do 
		{   
     	if (!expand_divide_list (&curr_node)) 
			{	
        errmsg ("Out of memory");
				goto cancel;      
			} 

			wcmd->param.l_divide_1.count++;	/* update list length */	
 	
      if (TESLA_filename(0,FT_BOTH,&curr_node->f_name) != PARSE_OK) 
			{
				errmsg ("Filename expected"); 
				goto cancel;
      }

 			if (wcmd->param.l_divide_1.first == NULL)
        wcmd->param.l_divide_1.first = curr_node;

      nextptr = &(curr_node->next);
  
      /* Another file name? */
      lex_next (); if (lex_cancel ()) goto cancel;    
		} 
		while (!lex_ismark (':'));

    lex_next (); if (lex_cancel()) goto cancel;

		if (TESLA_percent_val(&f) != PARSE_OK)
		{
      errmsg ("A precentage expected");
			goto cancel;    
		}

		f /= 100;

    /* fill in the part -fields */
    for (; wptr != nextptr; wptr = &((*wptr)->next)) {
      (*wptr)->part = f;
      sum += f;
    }

    lex_next ();
    if (lex_cancel ()) goto cancel;
	}
  while (lex_ismark (','));

  if (sum > 1.00) 
	{
		errmsg("Precentages sum over 100");
		goto cancel;  
  }

  if (TESLA_is_reserved_word(others_sym)) 
	{
    int	  count = 0;
    float f;

    lex_next ();  /* tiedostonimi */
    if (lex_cancel ()) goto cancel;

    do 
		{   /* file list for OTHERS */          
			if (!expand_divide_list (&curr_node)) {
        errmsg ("Out of memory");
        goto cancel;
      }
 
			if (TESLA_filename(0,FT_BOTH,&curr_node->f_name) != PARSE_OK) 
			{
				errmsg ("Filename expected"); 
				goto cancel;
      }

   		if (wcmd->param.l_divide_1.first == NULL)
        wcmd->param.l_divide_1.first = curr_node;
			
			count ++;

      nextptr = &(curr_node->next);
 
      /* Another file name? */
      lex_next (); if (lex_cancel ()) goto cancel;    
		} while (!lex_ismark (';'));

    /* fill the part -fields */
    f = (1.00 - sum) / count;
    for (; wptr != nextptr; wptr = &((*wptr)->next)) 
		{
      (*wptr)->part = f;
    }
  }
	else if (sum < 1.00)
	{
		errmsg("Precentages don't add up to 100");
		goto cancel;
	}
  
  parse_semicolon ();
  if (lex_cancel ()) goto cancel; 

	/* here, check if some of the filenames exist already */
	ask = 1;
	p_div_node = wcmd->param.l_divide_1.first;
	while (ask && p_div_node)
	{
		if (confirm_overwrite(p_div_node->f_name,&ask) != PARSE_OK) goto cancel;
		p_div_node = p_div_node->next;	
	}

	*next_cmd = wcmd;

  return 0;

	cancel:

	free_cmd (wcmd, 0);
	return 1;

 }


/* -------------------------------------------------------------------------- *\
 			Add new node to the end of list of attributes to EXCLUDE
\* -------------------------------------------------------------------------- */

int expand_exclude_list (struct t_exclude_node **curr_node) 
{

  struct t_exclude_node *p;
  
  p = (struct t_exclude_node *) malloc (sizeof (struct t_exclude_node));
  if (p != NULL) {
    p->next = NULL; p->attr_name = NULL;
    if (*curr_node == NULL)
      *curr_node = p;
    else {
      (*curr_node)->next = p; *curr_node = p;
    }
  }
  return (p != NULL);
}

/* -------------------------------------------------------------------------- *\
												Parse an exclude command
\* -------------------------------------------------------------------------- */

int parse_exclude (t_command **next_cmd) 
{ 
  int			cmd_id = K_EXCLUDE;
  struct t_exclude_node *curr_node;
  char			*token;
  t_command     *wcmd;
      
  /* Attribute name or ALL */
  lex_next ();	if (lex_cancel ()) return 1;

  /* Do we have ALL EXCEPT -syntax? */
  if (TESLA_is_reserved_word(all_sym)) 
	{
		lex_next ();  	if (lex_cancel ()) return 1;
    if (!TESLA_is_reserved_word(except_sym)) {
      errmsg ("EXCEPT expected");
      return 1;
    }    
    cmd_id = K_EXCLUDE_ALL_EXCEPT;
    lex_next ();	if (lex_cancel ()) return 1;
  } 
  
  /* Allocate command structure */
  wcmd = alloc_cmd ();
  if (wcmd == NULL) {
    errmsg ("Out of memory");
    return 1; 
  }

  wcmd->cmd_id = cmd_id;

	
  /* Get names of attributes to exclude (or not to exclude) */
  while (1) 
	{
		if (lex_ismark(';')) 
			break;
		else if (TESLA_attribute_name(&token) == PARSE_OK)  
		{ 

      /* Add attribute to list */      
      if (wcmd->param.l_exclude_1.first == NULL) {
        expand_exclude_list (&(wcmd->param.l_exclude_1.first));
        curr_node = wcmd->param.l_exclude_1.first;
      } else
        expand_exclude_list (&curr_node);
      curr_node->attr_name = token;
      
      /* Next token */
      lex_next ();	if (lex_cancel ()) return 1;
    } 
		else 
		{
      errmsg ("Attribute name expected");
      return 1;
    } 
  }
  
  parse_semicolon ();	if (lex_cancel ()) return 1;
  
	*next_cmd = wcmd;

  return 0;
}


/* -------------------------------------------------------------------------- *\
 					Add new node to the end of list of attributes to INCLUDE
\*--------------------------------------------------------------------------*/

int expand_incl_list (t_include_1 **curr_node) 
{

  t_include_1 *p;
  
  p = (t_include_1 *) malloc (sizeof (t_include_1));
  if (p != NULL) 
	{
		p->val_cnt = 0;
    p->next = NULL;
    (*curr_node)->next = p;
    (*curr_node) = p;
  }
  return (p != NULL);
}

/* -------------------------------------------------------------------------- *\
 				Add a new node to the end of list of available values for an 
 	  attribute to be included by INCLUDE-command.
\* -------------------------------------------------------------------------- */

int expand_val_list (struct t_incl_val_node **curr_node) 
{

  struct t_incl_val_node *p;
  
  p = (struct t_incl_val_node *) malloc (sizeof (struct t_incl_val_node));
  if (p != NULL) 
{
    p->next = NULL;
    (*curr_node)->next = p;
    (*curr_node) = p;
  }
  return (p != NULL);
}


/* -------------------------------------------------------------------------- *\
											Parse an INCLUDE - command
\* -------------------------------------------------------------------------- */

int parse_include (t_command **next_cmd) 
{ 
  char				*token;
  int  				is_more = 1,
  				is_values = 1;
  struct t_incl_val_node	*curr_val_node;
  t_command     		*wcmd;
  t_include_1			*curr_incl_node;
      

  if ((wcmd = alloc_cmd ()) == NULL) 
	{
    errmsg ("Out of memory!");
    return 1;
  }
  wcmd->cmd_id = K_INCLUDE;
  curr_incl_node = NULL;

  while (is_more) 
	{

    /* Attribute name */
    lex_next ();	if (lex_cancel ()) return 1;
    if (TESLA_attribute_name(&token) != PARSE_OK) {
      errmsg ("Attribute name expected");
      free_cmd (wcmd, 0); return 1;
    }
    if (curr_incl_node == NULL)
      curr_incl_node = &(wcmd->param.l_include_1);
    else if (!expand_incl_list (&curr_incl_node)) {
      errmsg ("Out of memory"); 
      free_cmd (wcmd, 0); return 1;
    }
    curr_incl_node->attr_name = token;

    /* Colon */
    lex_next ();	if (lex_cancel ()) return 1;
    if (!lex_ismark (':')) {
      errmsg (": expected");
      free_cmd (wcmd, 0); return 1;
    }
    
    /* 1st value to go to value list */
    lex_next ();		if (lex_cancel ()) return 1;
    is_values = 1;
    curr_val_node = NULL;
    while (is_values) 
		{
      if (TESLA_attribute_value(&token) == PARSE_OK)  { 

        /* Add value to list */
        if (curr_val_node == NULL)
          curr_val_node = &(curr_incl_node->val_list_head);
        else if (!expand_val_list (&curr_val_node)) {
          errmsg ("Out of memory");
          free_cmd (wcmd, 0); return 1;
        }
        alloc_string (&curr_val_node->attr_val, token);
        curr_incl_node->val_cnt++;

        /* next token */
        lex_null ();
        lex_get ();	if (lex_cancel ()) return 1;
      } else if (lex_ismark (',')) {
        is_values = 0; lex_null ();
      } else if (lex_ismark (';'))
        is_more = is_values = 0;
      else {
        errmsg ("Attribute value expected");
        return 1;
      }
    }  /* while (is_values) */
  }
  parse_semicolon();
  if (lex_cancel ()) {
    free_cmd (wcmd, 0);
    return 1;
  }    

	*next_cmd = wcmd;

	return PARSE_OK;
}




/* -------------------------------------------------------------------------- *\
   Adds a new node to the end of list of old value-relation-new value
 	  triplets
\* -------------------------------------------------------------------------- */

expand_map_list (t_map_node **curr_node, t_val old_val,t_val_type ov_type, 
		 t_relation curr_rel, char *new_val) 
{
  t_map_node *p;
  
  p = (t_map_node *) malloc (sizeof (t_map_node));
  if (p != NULL) 
	{
    p->next = NULL;
		p->old_type = ov_type;
    p->old_val = old_val;
    p->relation = curr_rel;
    p->new_val = new_val; /* ??? */
    if ((*curr_node) != NULL)
      (*curr_node)->next = p;
    (*curr_node) = p;
  }

  return (p != NULL);
}


/*--------------------------------------------------------------------------*\
   				Add a new node to other values -list of a map-command node
\*--------------------------------------------------------------------------*/

expand_oth_list (t_map_node_oth **head, char *new_val) 
{
	
  t_map_node_oth *p;
  
  p = (t_map_node_oth *) malloc (sizeof (t_map_node_oth));
  if (p != NULL) {
    p->attr_val = new_val;
    p->next = (*head);
    (*head) = p;
  }
  return (p != NULL);
}


/*--------------------------------------------------------------------------*\
												Parse a MAP-command
\*--------------------------------------------------------------------------*/

int parse_map (t_command **next_cmd) 
{
	
	int		a_ref;
	t_attr_types	a_type;
	t_map_node	*curr_node=0;
	t_relation	curr_rel;
	int		oth_values = 1;
	t_command	*new_cmd;
	char		*new_val;
	t_val_type ov_type;
	t_val		old_val;	
	float		s = 1.0;
	char		*token;
	char		wstr[256];
	
  /* Attribute name */
  lex_next ();				if (lex_cancel ())  return 1;
  if (TESLA_attribute_name(&token) != PARSE_OK) 
	{
		errmsg("Attribute name expected");
		return PARSE_CANCELLED;
	}

  /* Allocate space for the cmd node and init it. */
  new_cmd = alloc_cmd ();
  if (new_cmd == NULL) {
    errmsg ("Out of memory");
    return 1; 
  }
  new_cmd->cmd_id = K_MAP; 
  new_cmd->param.l_map_1.attr_name = token;
  
  /* Get reference to and type of attribute */
  if (WS_get_attr_ref (new_cmd->param.l_map_1.attr_name, &a_ref) != NO_ERROR)
		return PARSE_FAILED;
  WS_get_attr_type (a_ref, &a_type);
  
  /* as */
  lex_next (); if (lex_cancel ()) goto cancel;

  if (!TESLA_is_reserved_word(as_sym)) {
    errmsg ("AS expected");
		goto cancel;  
	}
 
  /* relation list */
  while (1) {
    lex_next (); if (lex_cancel ()) goto cancel;    
		if (lex_ismark ('<')) {
      /* < or <= */
      if (a_type == T_USER) 
			{
        errmsg ("Relation invalid for unordered attributes"); 
  			goto cancel;
      }
      lex_next (); if (lex_cancel ()) goto cancel;      
			if (lex_ismark ('=')) 
			{
      	curr_rel = LE;
      	lex_next (); if (lex_cancel ()) goto cancel;      
			} 
			else
        curr_rel = LT;
    } 
		else if (lex_ismark ('=')) 
		{
      curr_rel = EQ;
      lex_next (); if (lex_cancel ()) goto cancel;    
		} 
		else if (new_cmd->param.l_map_1.first == NULL) 
		{
      errmsg ("Relation expected."); 
			goto cancel;
    } 
		else break;
    
    /* Old value */
    s = 1.0;  /* Sign for numeric values */
    switch (a_type) {
      case T_USER:
        if (TESLA_attribute_value(&token) != PARSE_OK) 
				{
          errmsg ("Attribute value expected");
					goto cancel;        
				}
				else if (!WS_is_attr_val_name(a_ref,token))
				{
					errmsg("Attribute value expected");
					goto cancel;
				}
        alloc_string(&old_val.discrete, token);
				ov_type = l_discrete;
        break;
      case T_FLOAT:
				if (!TESLA_cont_val(&old_val.f))
					ov_type = l_float;
				else
				{
					errmsg("Numeric value expected");
					goto cancel;
				}	
				break;
      case T_INT:
				if (!TESLA_int_val(&old_val.i))
					ov_type = l_int;
				else
				{
					errmsg("Integer value expected");
					goto cancel;
				}	
        break;
    }
    
    /* Colon */
    lex_next (); if (lex_cancel ()) goto cancel;

    if (!lex_ismark (':')) 
		{
      errmsg (": expected");
			goto cancel;
    }
    
    /* New value */
    lex_next (); if (lex_cancel ()) goto cancel;    

		if (TESLA_attribute_value(&token) != PARSE_OK) 
		{
      errmsg ("Attribute value expected");
			goto cancel; 
    }
    new_val = token;
    
    if (!expand_map_list (&curr_node, old_val, ov_type,curr_rel, new_val)) 
		{
      errmsg ("Out of memory");
 			goto cancel;
    } 
		else if (new_cmd->param.l_map_1.first == NULL)
      new_cmd->param.l_map_1.first = curr_node;
  }  /* while loop collecting the relation list */
  
  /* Else/others-branch (if any) */
  if (TESLA_is_reserved_word(else_sym)) 
	{
    lex_next (); if (lex_cancel ()) goto cancel;

    if (TESLA_attribute_value(&token) != PARSE_OK ) 
		{
      errmsg ("Invalid value");
      goto cancel;
    }
    alloc_string(&new_cmd->param.l_map_1.else_val, token);
    lex_next (); if (lex_cancel ()) goto cancel;
  } 

	if (TESLA_is_reserved_word(others_sym)) 
	{
    /* Others-branch */
    lex_next (); if (lex_cancel ()) goto cancel;

    while (oth_values) 
		{
      if (!(oth_values = !TESLA_attribute_value(&token))) 
			{
        errmsg ("Invalid value");
        goto cancel;
      }
      alloc_string(&new_val, token);
      if (!expand_oth_list (&(new_cmd->param.l_map_1.others), new_val)) 
			{
        errmsg ("Out of memory");
        goto cancel;
      }

      lex_next ();if (lex_cancel ()) goto cancel;

			if (lex_ismark(';')) break;
    }  /* while (oth_values) */
  } 
	else if (!lex_ismark (';')) 
	{
    errmsg ("OTHERS or ; expected");
		goto cancel;
  }  

	if (lex_cancel ()) goto cancel;

	if (!lex_ismark(';')) 
	{
		errmsg("\';\' expected");
		goto cancel;
	}

	lex_null();	

	*next_cmd = new_cmd;

	/* optimize it right away */
	if (optimize_map(new_cmd))
		return PARSE_FAILED;
	else
		return PARSE_OK;

	cancel:
	
   free_cmd (new_cmd, 0);
	 return 1;

}


/*--------------------------------------------------------------------------*\
 		Transform the MAP-command structure to a more suitable
 	  form for executing the command
\*--------------------------------------------------------------------------*/

int optimize_map (t_command *cmd) 
{
  t_map_node 	 *ptr, *ptr1, **pptr;
  t_map_node_oth *ptr2;
  int             retcode = 0;
  t_attr_node    *a_node;
  int             go;
  
  if (WS_create_attr (cmd->param.l_map_1.attr_name, T_USER, &a_node) 
		!= NO_ERROR ) 
	{
    printf ("Out of memory.\n");
    return 1;
  }

  /* We'll add the values in value list to the attribute node and 
     at the same time handle * and ? -mappings by copying them and 
     removing them from the mapping list. */
  for (pptr = &(cmd->param.l_map_1.first); *pptr; pptr = &((*pptr)->next)) {
    ptr = *pptr;
    retcode = retcode | WS_join_val_low (a_node, ptr->new_val);
    if (ptr->old_type == l_dont_care) {
      cmd->param.l_map_2.dont_care_val = ptr->new_val;
      *pptr = ptr->next;
      free (ptr);
    } 
    if (ptr->old_type == l_dont_know) {
      cmd->param.l_map_2.dont_know_val = ptr->new_val;
      *pptr = ptr->next;
      free (ptr);
    } 
  } 
  
  /* We'll add others-values to the new attribute node */
  for (ptr2 = cmd->param.l_map_1.others; ptr2; ptr2 = ptr2->next)
    retcode = retcode | WS_join_val_low (a_node, ptr2->attr_val);
 
  if (cmd->param.l_map_1.else_val != NULL)
    retcode = retcode | WS_join_val_low (a_node, cmd->param.l_map_1.else_val);
  cmd->param.l_map_2.attribute_table = a_node;
  cmd->param.l_map_2.attr_name = strdup (cmd->param.l_map_1.attr_name);
  
  /* We'l sort the mapping list */ 
  do {
    go = 0;
    for (pptr = &(cmd->param.l_map_1.first); 
         (*pptr) && ((*pptr)->next); pptr = &((*pptr)->next)) {
      ptr = *pptr;
      ptr1 = ptr->next;
    
      if (ptr->old_type == l_float &&
          ptr->old_val.f > ptr1->old_val.f ||
          ptr->old_type == l_int &&
          ptr->old_val.i > ptr1->old_val.i ||
          ptr->old_type == l_discrete &&
          strcmp (ptr->old_val.discrete, ptr1->old_val.discrete) > 0) {
        go = 1;
        *pptr = ptr1;   
        ptr->next = ptr1->next;
        ptr1->next = ptr;
      }
    }
  } while (go);  
      
  return retcode;
}


/*--------------------------------------------------------------------------*\
 	Add new node to the end of list of attributes to apply NOISE to.
\*--------------------------------------------------------------------------*/

int expand_noise_list (struct t_noise_node **curr_node) 
{

  struct t_noise_node *p;
  
  p = (struct t_noise_node *) malloc (sizeof (struct t_noise_node));
  if (p != NULL) 
	{
    p->next = NULL; p->noise_p = 0.0;
    if (*curr_node == NULL)
      *curr_node = p;
    else {
      (*curr_node)->next = p; *curr_node = p;
    }
  }
  return (p != NULL);
}

/* -------------------------------------------------------------------------- *\
 												Parser for a NOISE-command
\* -------------------------------------------------------------------------- */

int parse_noise (t_command **next_cmd) 
{ 

  struct t_noise_node	*curr_node;
  char			*token;
  t_command     	*wcmd;
      
  /* Name of 1st attribute */
  lex_next ();	if (lex_cancel ()) return 1;

  /* Allocate command structure */
  wcmd = alloc_cmd ();
  if (wcmd == NULL) {
    errmsg ("Out of memory");
    return 1; 
  }
  wcmd->cmd_id = K_NOISE;

  /* Get names of attributes to apply noise to */
  while (1) 
	{
		if (lex_ismark(';'))
			break;	
		else if (TESLA_is_reserved_word(others_sym)) 
		{ 		
			lex_next (); if (lex_cancel ()) goto cancel;				
      
			/* others-percent */
			if (TESLA_percent_val(&wcmd->param.l_noise_1.others) != PARSE_OK)
			{
				errmsg("Invalid noise percent");
				goto cancel;
			}
			wcmd->param.l_noise_1.others /= 100.0;
			lex_null(); break;
		}	
		else if (TESLA_attribute_name(&token) == PARSE_OK)  
		{ 
      /* Add attribute to list */      
      if (wcmd->param.l_noise_1.first == NULL) 
			{
        expand_noise_list (&(wcmd->param.l_noise_1.first));
        curr_node = wcmd->param.l_noise_1.first;
      } 
			else
        expand_noise_list (&curr_node);
      
			curr_node->attr_name = token;
      
      /* Colon */
      lex_next ();
      if (lex_cancel ()) goto cancel;

      if (!lex_ismark (':')) 
			{
        errmsg (": expected");
				goto cancel;  
      }
      
      /* Noise percent */
      lex_next (); if (lex_cancel ()) goto cancel;

			if (TESLA_percent_val(&curr_node->noise_p) != PARSE_OK)
			{
				errmsg("Invalid noise percent");
				goto cancel;
			}

      curr_node->noise_p /= 100.0;
      
			/* Next token */
			lex_next();			
			
    } 
		else 
		{
			errmsg("An attribute name or OTHERS expected");
			goto cancel;
		}
	}  /* while (1) */
  
  parse_semicolon ();
  if (lex_cancel ()) goto cancel; 
	*next_cmd = wcmd;
  return PARSE_OK;

	cancel:
	
	free_cmd (wcmd, 0); 
	return PARSE_FAILED;

}


/* -------------------------------------------------------------------------- *\
			Update the type of the READ-command according to the given datatype.
		Note: the new datatype is a combination of the old command type and
		the given datatype
\* -------------------------------------------------------------------------- */

int read_update_cmd_TESLA_datatype(t_command *p_cmd,
	t_filetype TESLA_datatype,char *filename)
{
	switch (TESLA_datatype)
	{
		case FT_ATTS:					
			p_cmd->param.l_read_1.a_filename = filename;
			switch (p_cmd->cmd_id)
			{
				case K_READ_EXS:p_cmd->cmd_id = K_READ_ATTS_AND_EXS;break;
				case K_NULL:p_cmd->cmd_id = K_READ_ATTS;break;
				case K_READ_RULE: goto rulesep;
				default: goto spurious;
			} 		
		break;
		case FT_EXS:				
			p_cmd->param.l_read_1.e_filename = filename;
			switch (p_cmd->cmd_id)
			{
				case K_READ_ATTS:p_cmd->cmd_id = K_READ_ATTS_AND_EXS;break;	
				case K_NULL: p_cmd->cmd_id = K_READ_EXS;break;					
				case K_READ_RULE: goto rulesep;
				default: goto spurious;
			}
		break;
		case FT_BOTH:
			p_cmd->param.l_read_1.a_filename = filename;					
			switch (p_cmd->cmd_id)
			{
				case K_NULL: p_cmd->cmd_id = K_READ_BOTH; break;
				case K_READ_RULE: goto rulesep;
				default: goto spurious;
			}
		break;
		case FT_ATTSRULE:					
			p_cmd->param.l_read_1.r_filename = filename;
			switch (p_cmd->cmd_id)
			{					
				case K_NULL: p_cmd->cmd_id = K_READ_RULE; break;
				case K_READ_RULE: goto spurious;
				default: goto rulesep;
			}
		break;
	}

	return PARSE_OK;

	rulesep:
		errmsg("Rule must be read in separate command");
		return PARSE_CANCELLED;

	spurious:
		errmsg("Spurious data type specifications");
		return PARSE_CANCELLED;
}


/* -------------------------------------------------------------------------- *\
													Parse a READ-command
\* -------------------------------------------------------------------------- */

int parse_read(t_command **cmd)
{
	t_algorithms alg;
	t_command *p_cmd;
	char *f_name=0;
	t_filetype d_type;	


	if (!(*cmd = alloc_cmd()))
	{
		errmsg("Out of memory");
		return PARSE_CANCELLED;
	}

	p_cmd = *cmd;
	p_cmd->cmd_id = K_NULL;	

	/* algorithm name : whose grammar to use */
	if (parse_algorithm(&p_cmd->param.l_read_1.alg) != PARSE_OK) goto cancel;

	lex_next(); if (lex_cancel()) goto cancel;

	/* parse <TESLA_datatype> <filename> pairs */
	if (TESLA_datatype(&d_type) == PARSE_OK)
	{
		do
		{
			lex_next(); if (lex_cancel()) goto cancel;

			if (TESLA_filename(p_cmd->param.l_read_1.alg,d_type,&f_name) != PARSE_OK)
			{
				errmsg("Filename expected");
				goto cancel;
			}
			else if (read_update_cmd_TESLA_datatype(*cmd,d_type,f_name) != PARSE_OK)
				goto cancel; /* error message has been displayed already */

			lex_next(); if (lex_cancel()) goto cancel;
		}
		while (TESLA_datatype(&d_type) == PARSE_OK);
	}
	else
	{
		errmsg("Datatype expected");
		goto cancel;
	}

	if (!TESLA_is_separator(';'))
	{
	 errmsg("Datatype or ';' expected");
	 goto cancel; 
	}
	lex_null();

	return PARSE_OK;

	cancel:
		free_cmd(*cmd,0);
		return PARSE_CANCELLED;
}

/* -------------------------------------------------------------------------- *\
				Add a node to the end of the list of attributes to STAT	
\* -------------------------------------------------------------------------- */

int expand_stat_list (t_snode **curr_node) 
{

  t_snode *p;
  
  p = (t_snode *) malloc (sizeof (t_snode));
  if (p != NULL) {
    p->next = NULL; p->attr_name = NULL;
    if (*curr_node == NULL)
      *curr_node = p;
    else {
      (*curr_node)->next = p; *curr_node = p;
    }
  }
  return (p != NULL);
}


/* -------------------------------------------------------------------------- *\
 										Parse a STAT-command
\* -------------------------------------------------------------------------- */

int parse_stat (t_command **next_cmd) 
{
  int	    cmd_id = K_STAT;
	FILE *fp;
  t_snode   *curr_node = NULL;
  char	    *into_fname = NULL;
  char	    *token;
  t_command *wcmd;
      
  /* Attribute name or ALL */
  lex_next ();	if (lex_cancel ()) return 1;

  /* Do we have ALL -syntax? */
  if (TESLA_is_reserved_word(all_sym)) {
    cmd_id = K_STAT_ALL;

    /* INTO !! */
    lex_next ();  	if (lex_cancel ()) return 1;
    if (!parse_into (&into_fname))
      return 1;
  }
  
	if (into_fname ? !(fp = STATS_open_file(into_fname)) : 0)
	{
		char msg[256];
		sprintf(msg,"File %s is not a valid TELA statistics file");
		errmsg(msg);
		return 1;
	}

  /* Allocate command structure */
  wcmd = alloc_cmd ();
  if (wcmd == NULL) {
    errmsg ("Out of memory");
    return 1; 
  }

  wcmd->cmd_id = cmd_id;
  if (into_fname != NULL) {
    alloc_string(&wcmd->param.l_stat_1.stat_file_name, into_fname);
    into_fname = NULL;
  }

  /* Get names of attributes to stat if not all */
  if (cmd_id == K_STAT) {
    while (1) 
		{ 
			if (lex_ismark (';')) 
				break;       
			else if (TESLA_attribute_name(&token) == PARSE_OK)  
			{ 
        /* Add attribute to list */      
        expand_stat_list (&curr_node);
        if (wcmd->param.l_stat_1.first == NULL)
          wcmd->param.l_stat_1.first = curr_node;
        alloc_string(&curr_node->attr_name, token);
      
        /* Next token */
        lex_next ();	
        if (lex_cancel ()) {
          free_cmd (wcmd, 0); return 1;
        }
        if (TESLA_is_reserved_word(into_sym))
          break;
      } 
			else
			{
				errmsg("Attribute name expected");
				goto cancel;
			}
    }
    
    /* INTO part */
    if (!parse_into (&into_fname)) {
      free_cmd (wcmd, 0); return 1;
    }
    if (into_fname != NULL) {
      alloc_string(&wcmd->param.l_stat_1.stat_file_name, into_fname);
      into_fname = NULL;
    }
  }  
  parse_semicolon ();
  if (lex_cancel ()) {
    free_cmd (wcmd, 0); return 1;
  }
  
	*next_cmd = wcmd;

  return 0;

	cancel:

	free_cmd(wcmd,0); return PARSE_FAILED;		
}


/* -------------------------------------------------------------------------- *\
			Update the type of the write command on the basis of given datatype
\* -------------------------------------------------------------------------- */

int write_update_cmd_TESLA_datatype(t_command *p_cmd,t_filetype TESLA_datatype,
	char *filename)
{
	switch (TESLA_datatype)
	{
		case FT_ATTS:					
			p_cmd->param.l_write_1.a_filename = filename;
			switch (p_cmd->cmd_id)
			{
				case K_WRITE_EXS:p_cmd->cmd_id = K_WRITE_ATTS_AND_EXS;break;
				case K_NULL:p_cmd->cmd_id = K_WRITE_ATTS;break;
				case K_WRITE_RULE: goto rulesep;
				default: goto spurious;
			} 		
		break;
		case FT_EXS:				
			p_cmd->param.l_write_1.e_filename = filename;
			switch (p_cmd->cmd_id)
			{
				case K_WRITE_ATTS:p_cmd->cmd_id = K_WRITE_ATTS_AND_EXS;break;	
				case K_NULL: p_cmd->cmd_id = K_WRITE_EXS;break;					
				case K_WRITE_RULE: goto rulesep;
				default: goto spurious;
			}
		break;
		case FT_BOTH:
			p_cmd->param.l_write_1.a_filename = filename;					
			switch (p_cmd->cmd_id)
			{
				case K_NULL: p_cmd->cmd_id = K_WRITE_BOTH; break;
				case K_WRITE_RULE: goto rulesep;
				default: goto spurious;
			}
		break;
		case FT_ATTSRULE:					
			p_cmd->param.l_write_1.r_filename = filename;
			switch (p_cmd->cmd_id)
			{					
				case K_NULL: p_cmd->cmd_id = K_WRITE_RULE; break;
				case K_WRITE_RULE: goto spurious;
				default: goto rulesep;
			}
		break;
	}

	return PARSE_OK;

	rulesep:
		errmsg("Rule must be written in separate command");
		return PARSE_CANCELLED;

	spurious:
		errmsg("Spurious data type specifications");
		return PARSE_CANCELLED;
}


/* -------------------------------------------------------------------------- *\
													Parse a write command
\* -------------------------------------------------------------------------- */

int parse_write(t_command **cmd)
{
	t_algorithms alg;
	int ask;          /* ask further confirmations ? */
	t_command *p_cmd;
	char *f_name=0;
	t_filetype d_type;	


	if (!(*cmd = alloc_cmd()))
	{
		errmsg("Out of memory");
		return PARSE_CANCELLED;
	}

	p_cmd = *cmd;
	p_cmd->cmd_id = K_NULL;	

	/* algorithm name : whose grammar to use */
	if (parse_algorithm(&p_cmd->param.l_write_1.alg) != PARSE_OK) goto cancel;

	lex_next(); if (lex_cancel()) goto cancel;

	/* parse <TESLA_datatype> <filename> pairs */
	if (TESLA_datatype(&d_type) == PARSE_OK)
	{
		do	
		{
			lex_next(); if (lex_cancel()) goto cancel;

			if (TESLA_filename(p_cmd->param.l_write_1.alg,d_type,&f_name) != PARSE_OK)
			{
				errmsg("Filename expected");
				goto cancel;
			}
			else if (write_update_cmd_TESLA_datatype(*cmd,d_type,f_name) != PARSE_OK)
				goto cancel;		

			lex_next(); if (lex_cancel()) goto cancel;
		}	
		while (TESLA_datatype(&d_type) == PARSE_OK);
	}
	else
	{
		errmsg("Datatype expected");
		goto cancel;
	}

	if (!TESLA_is_separator(';')) 
	{
		errmsg("Datatype or ';' expected");
		goto cancel; 
	}	

	ask = 1;

	if (p_cmd->param.l_write_1.a_filename)
		if (confirm_overwrite(p_cmd->param.l_write_1.a_filename,&ask) != PARSE_OK) 
			goto cancel;
	if (p_cmd->param.l_write_1.e_filename && ask)
		if (confirm_overwrite(p_cmd->param.l_write_1.e_filename,&ask) != PARSE_OK) 
			goto cancel;
	if (p_cmd->param.l_write_1.r_filename && ask)
		if (confirm_overwrite(p_cmd->param.l_write_1.r_filename,&ask) != PARSE_OK) 
			goto cancel;

	lex_null();

	return PARSE_OK;

	cancel:
		free_cmd(*cmd,0);
		return PARSE_CANCELLED;
}


/* -------------------------------------------------------------------------- *\
								Parse the RECORD-clause for a DO-command
\* -------------------------------------------------------------------------- */

int do_record(t_do_1 *p_do)
{
	char *stat;
	t_algorithms alg;
	t_stats_node *p_stats_node;
	list_init(&p_do->stats);

	if (!TESLA_is_reserved_word(record_sym)) return PARSE_FAILED;

	do
	{
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (TESLA_is_reserved_word(time_sym))
		{
			p_stats_node = STATS_create_node("TIME");
			p_stats_node->printing = 2; /* averages are to be printed */
			list_insert(&p_do->stats,p_stats_node);
		}
		else if (TESLA_is_reserved_word(testtime_sym))
		{		
			p_stats_node = STATS_create_node("TESTTIME");
			p_stats_node->printing = 2; 
			list_insert(&p_do->stats,p_stats_node);
		}
		else if (!TESLA_word(&stat))
		{
			p_stats_node = STATS_create_node(stat);
			p_stats_node->printing = 2;
			list_insert(&p_do->stats,p_stats_node);
			free(stat);
		}
		else if (TESLA_is_reserved_word(into_sym)) break;
		else if (TESLA_is_separator(';')) return PARSE_OK;
		else if (TESLA_is_reserved_word(comment_sym)) return PARSE_OK;
		else
		{
			errmsg("No such statistic");
			return PARSE_CANCELLED;
		}
	} 
	while (1);
	/* we have parsed INTO, if get here, now get the filename */
	lex_next(); if (lex_cancel()) return PARSE_CANCELLED;

	/* use TELA's default extensions if filename is unquoted */
	ADB_get_alg_ref("TELA",&alg);
	if (TESLA_filename(alg,FT_STATS,&p_do->stat_file_name) == PARSE_FAILED)
	{
		errmsg("Filename expected");
		return PARSE_CANCELLED;	
	}

	/* get token that hopefully is ';' or comment_sym */
	lex_next(); if (lex_cancel()) return PARSE_CANCELLED; 

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
												Parse a COMMENT-clause
\* -------------------------------------------------------------------------- */

int comment(char **comment_str)
{
	if (!TESLA_is_reserved_word(comment_sym)) return PARSE_FAILED;

	lex_next(); if (lex_cancel()) return PARSE_CANCELLED;

	if (TESLA_string(comment_str))
	{
		errmsg("Double quoted (\") string expected");
		return PARSE_CANCELLED;
	}	

	/* get token that hopefully is ';' */
	lex_next(); if (lex_cancel()) return PARSE_CANCELLED; 

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
									Parse the header of the DO-loop
\* -------------------------------------------------------------------------- */

int parse_do_header(t_command **pp_cmd) 
{
	t_do_1 *p_do;
	FILE *fp;

	*pp_cmd = alloc_cmd();

	(**pp_cmd).cmd_id = K_DO;
	
	p_do = &(**pp_cmd).param.l_do_1;

	p_do->stat_file_name = p_do->comment = 0;

	/* iteration count */
	lex_next(); if (lex_cancel()) goto cancel;
	if (TESLA_nat_val(&p_do->do_cnt))
	{
		errmsg("Iteration count expected");
		goto cancel;
	}

	lex_next(); if (lex_cancel()) goto cancel;

	/* optional record-clause */
	if (do_record(p_do) == PARSE_CANCELLED) goto cancel;
	/* optional comment-clause */
	if (comment(&p_do->comment) == PARSE_CANCELLED) goto cancel;	

	if (TESLA_is_separator(';'))
	{
		lex_null();

		if (p_do->stat_file_name)
		{
			if (fp = STATS_open_file(p_do->stat_file_name))
				fclose(fp);
			else goto cancel;
		}

		return PARSE_OK;
	}	

	cancel:

	free_cmd(*pp_cmd,0);
	return PARSE_CANCELLED;
	
}


/* -------------------------------------------------------------------------- *\
			Does the given statistics exist in the stats-list of given command
\* -------------------------------------------------------------------------- */

int stat_is_covered(char *name,t_command *p_cmd)
{
	t_stats_node *dummy;

	while (p_cmd)
	{
		switch (p_cmd->cmd_id)
		{
			case K_RUN: 
				if (STATS_find_node(&p_cmd->param.l_run_1.stats,name,&dummy) 
					== EXECUTION_OK)
					return 1;
			break;
			case K_TEST: 
				if  (STATS_find_node(&p_cmd->param.l_test_1.stats,name,&dummy)
					== EXECUTION_OK)
					return 1;
			break;
			case K_CV:
				if  (STATS_find_node(&p_cmd->param.l_cv.stats,name,&dummy)
					== EXECUTION_OK)
					return 1;
      break;
		}
		p_cmd = p_cmd->next;
	}
	return 0;
}


/* -------------------------------------------------------------------------- *\
													Parse a DO-loop
\* -------------------------------------------------------------------------- */

int parse_do(t_command **pp_do_cmd)
{
	int rv=PARSE_OK,end = 0;
	t_do_1 *p_do; 

	/* first parse the header of the do-clause */
	if (!parse_do_header(pp_do_cmd))
	{
		t_command **pp_cmd;			
		p_do = &(**pp_do_cmd).param.l_do_1;
		pp_cmd = &p_do->commands;
		/* parse commands inside the do-clause */
		do
		{
	
			if (rv = parse_cmd(pp_cmd,1,0))
				goto end;
			switch ((**pp_cmd).cmd_id)
			{
				case K_END: end = 1; break;
				case K_RUN: 
				{	
					t_run_1 *p_run = &(**pp_cmd).param.l_run_1;
					p_run->pr_mode = 
						STATS_merge_lists(p_run->alg,0,&p_run->stats,&p_do->stats);
					break;
				}
				case K_TEST:					
				{
					t_test_1 *p_test = &(**pp_cmd).param.l_test_1;
					p_test->pr_mode = 
						STATS_merge_lists(p_test->alg,1,&p_test->stats,&p_do->stats);
					break;
				}
				case K_CV:
				{
					t_cv *p_cv = &(**pp_cmd).param.l_cv;
					p_cv->pr_mode =
						STATS_merge_lists(p_cv->alg,-1,&p_cv->stats,&p_do->stats);
					break;
				}
			}
			pp_cmd = &(**pp_cmd).next;	
		}
		while (!end);
 
		/* so, we got out. Now, either we found K_END or parse failed */	
		*pp_cmd = 0; /* end of list */
	}

	if (p_do->commands->cmd_id == K_END)
	{
		errmsg("Empty DO-loop");
		return PARSE_CANCELLED;
	}

end:

	if (lex_cancel() || rv)
	{
		free_cmd(*pp_do_cmd,0);
		return PARSE_CANCELLED;
	}
	else
	{
		/* check if all the statistics in DO-header ended up in some */
		/* commands statistics list */
		t_stats_node *p_stats_node;		

		if (list_get_first(&p_do->stats,(void**)&p_stats_node))
		{
			do
			{
				if (!stat_is_covered(p_stats_node->name,p_do->commands))
				{
					char msg[256];
					sprintf(msg,"Statistic %s not applicable",p_stats_node->name);
					errmsg(msg);
					return PARSE_CANCELLED;
				}	
			}
			while (list_get_next(&p_do->stats,(void**)&p_stats_node));
		}

		return PARSE_OK;
	}

}


/* -------------------------------------------------------------------------- *\
												Parse an END-command
\* -------------------------------------------------------------------------- */

int parse_end(t_command **pp_cmd)
{
	lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
	if (!TESLA_is_separator(';')) return PARSE_FAILED;

	*pp_cmd = alloc_cmd();
	(**pp_cmd).cmd_id = K_END;	
	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
												Allocate a parameter node
\* -------------------------------------------------------------------------- */

t_run_node *alloc_param_node(char* p_tmp_sel,char* p_tmp_val)
{
	t_run_node *p_tmp;

	p_tmp = (t_run_node *)malloc(sizeof(t_run_node));
	p_tmp->selector = p_tmp_sel;
	p_tmp->val = p_tmp_val;
	p_tmp->next = 0;

	return p_tmp;
}


/* -------------------------------------------------------------------------- *\
								Parse the algorithms input parameters
\* -------------------------------------------------------------------------- */

int alg_params(t_algorithms alg,t_run_node **head)
{
	char *p_tmp_sel=0,*p_tmp_val=0;
	
	t_run_node *p_tmp = 0;  

	if (TESLA_is_separator('('))
	{
		do																				
		{
			lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
			if (TESLA_word(&p_tmp_sel) == PARSE_FAILED)						/* selector */
			{
				errmsg("Parameter name expected");
				goto cancel;
			}
			lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
			if (TESLA_is_separator('='))													/* = 				*/
			{
				lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
				if (TESLA_param_value(&p_tmp_val) ) 								/* value    */
				{
					errmsg("Parameter name expected");
					goto cancel;	
				}			
				lex_next(); 																						/* , or )   */
				if (lex_cancel()) goto cancel;  
			}
			/* ok, now we have a selector and possibly a value*/
			*head = alloc_param_node(p_tmp_sel,p_tmp_val);			

			/* check if they make any sense */	
			if (!ADB_alg_params_ok(*head,alg))
			{
				errmsg("Invalid parameter");
				goto cancel;
			}

			/* add them to the parameter list */ 

			(*head)->next = p_tmp;
			p_tmp = *head;
			p_tmp_val = 0; /* for safety */


		}		
		while (TESLA_is_separator(','));						

		if (TESLA_is_separator(')'))	
			return PARSE_OK;
		else
		{
			errmsg("\')\' expected");
			goto cancel;
		}  
	}	

	cancel:

	/* free parameter list here ? */

	return PARSE_CANCELLED;
}


/* -------------------------------------------------------------------------- *\
								Parse a RECORD-cluse of a RUN-command
\* -------------------------------------------------------------------------- */

int run_record(t_command *p_tmp)
{
	char token[256],*stat;
	t_stats_node *p_stats_node;
	t_run_1 *p_run_1 = &p_tmp->param.l_run_1;

	if (!TESLA_is_reserved_word(record_sym)) return PARSE_FAILED;

	do
	{
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (TESLA_is_reserved_word(time_sym))
		{ 
			p_stats_node = STATS_create_node("TIME");
			list_insert(&p_run_1->stats,p_stats_node);
		}
		else if (!TESLA_word(&stat))
		{
			if (ADB_alg_statistic_ok(stat,p_run_1->alg,0))
			{
				p_stats_node = STATS_create_node(stat);
				list_insert(&p_run_1->stats,p_stats_node);
				free(stat);
			}
			else
			{
				errmsg("No such statistic");
				free(stat);
				return PARSE_CANCELLED;
			}
		}
		else if (TESLA_is_reserved_word(into_sym)) break;
		else if (TESLA_is_separator(';')) return PARSE_OK;
		else
		{
			errmsg("No such statistic");
			return PARSE_CANCELLED;
		}
		p_run_1->pr_mode = 1; /* print statistics individually */
	} 
	while (1);

	/* we have parsed INTO, if get here, now get the filename */
	lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
	if (TESLA_filename(p_run_1->alg,FT_STATS,&p_run_1->stat_file_name) == PARSE_FAILED)
	{
		errmsg("Filename Expected");
		return PARSE_CANCELLED;	
	}

	lex_next(); if (lex_cancel()) return PARSE_CANCELLED; 

	if (!TESLA_is_separator(';'))
	{
		errmsg("\';\' expected");
		return PARSE_CANCELLED;	
	}

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
													Parse a RUN-command			
\* -------------------------------------------------------------------------- */

int parse_run(t_command **next_cmd)
{
	char *token;
	t_command *tmp;
	t_algorithms alg;
	FILE *fp;

	lex_next(); if (lex_cancel()) return PARSE_CANCELLED; 

	if (!TESLA_word(&token))
	{
		if (!ADB_get_alg_ref(token,&alg))
		{
			errmsg("No such algorithm"); 															
			return PARSE_CANCELLED;
		}
	}
	else
	{
		errmsg("Unquoted string expected");
		return PARSE_CANCELLED;
	}

	tmp = alloc_cmd();
	if (tmp == NULL)
	{
    errmsg ("Out of memory");   /* also this error handling could be 		*/
    return PARSE_FAILED; 				/* performed in alloc_cmd(), would save */
																/* some code */
  }

	tmp->cmd_id = K_RUN;
	tmp->param.l_run_1.alg = alg;
	tmp->param.l_run_1.a_mask = 0; /* all examples active */
	tmp->param.l_run_1.pr_mode = 0;	
	tmp->param.l_run_1.b_time = 
		!ADB_alg_statistic_ok("TIME",alg,0);

	lex_next(); if (lex_cancel()) return 1;

	/* optional: algorithm specific parametrs */
	if (!alg_params(alg,&(tmp->param.l_run_1.params)))
	{
		lex_next(); if (lex_cancel()) goto cancel;
	}
				
	if (run_record(tmp)== PARSE_CANCELLED) goto cancel;

	if (!TESLA_is_separator(';'))
	{
		errmsg("; expected");
		goto cancel;	 	
	}

	if (tmp->param.l_run_1.stat_file_name)
	{
		if (fp = STATS_open_file(tmp->param.l_run_1.stat_file_name))
			fclose(fp);
		else goto cancel;
	}

	*next_cmd = tmp;

	lex_null();
	return PARSE_OK; 

	cancel:	
	
	free_cmd(tmp,0);
	return PARSE_CANCELLED;

}


/* -------------------------------------------------------------------------- *\
												Parse a TEST-command
\* -------------------------------------------------------------------------- */

int parse_test(t_command **next_cmd)
{
	char *token,*stat;
	t_stats_node *p_stats_node;
	FILE *fp;

	t_command *p_tmp;
	t_algorithms alg;
	t_test_1 *p_test;

	lex_next(); if (lex_cancel()) return PARSE_CANCELLED; 

	if (!TESLA_word(&token) ? !ADB_get_alg_ref(token,&alg) : 1)
	{
		errmsg("No such algorithm"); 															
		printf("%s\n", token);
		return PARSE_CANCELLED;
	}

	p_tmp = alloc_cmd();
	if (p_tmp == NULL)
	{
    errmsg ("Out of memory");   /* also this error handling could be 		*/
    return PARSE_FAILED; 				/* performed in alloc_cmd(), would save */
																/* some code */
  }

	p_tmp->cmd_id = K_TEST;
	p_test = &p_tmp->param.l_test_1;
	p_test->alg = alg;
	p_test->a_mask = 0; /* all examples active */
	p_test->pr_mode = 0;
	p_test->b_testtime=
		!ADB_alg_statistic_ok("TESTTIME",p_test->alg,1);

	lex_next(); if (lex_cancel()) goto cancel;

	if (TESLA_is_reserved_word(record_sym))
	{
		do
		{
			lex_next(); if (lex_cancel()) goto cancel;
			if (TESLA_is_reserved_word(testtime_sym)) 
			{
				p_stats_node = STATS_create_node("TESTTIME");
				list_insert(&p_test->stats,p_stats_node);
			}
			else if (!TESLA_word(&stat))
			{
				if (ADB_alg_statistic_ok(stat,p_test->alg,1))
				{
					p_stats_node = STATS_create_node(stat);
					list_insert(&p_test->stats,p_stats_node);
					free(stat);
				}
				else
				{
					errmsg("No such statistic");
					free(stat);
					return PARSE_CANCELLED;
				}
			}
			else if (TESLA_is_reserved_word(into_sym)) break;
			else if (TESLA_is_separator(';')) 
			{
				*next_cmd = p_tmp;
				lex_null(); 
				return PARSE_OK;
			}
			else
			{
				errmsg("No such statistic");
				goto cancel;
			}	
			p_test->pr_mode = 1; /* print statistics individually */
		}	 
		while (1);

		/* we have parsed INTO, if get here, now get the filename */
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (TESLA_filename(p_test->alg,FT_STATS,&p_test->stat_file_name) == PARSE_FAILED)
		{
			errmsg("Filename expected");
			goto cancel;	
		}
		
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
	}

	if (!TESLA_is_separator(';'))
	{
		errmsg("SYNTAX_ERROR");
		goto cancel;
	}

	
	if (p_test->stat_file_name)
	{
		if (fp = STATS_open_file(p_test->stat_file_name))
			fclose(fp);
		else goto cancel;
	}

	*next_cmd = p_tmp;

		lex_null();
		return PARSE_OK;

	cancel:
	if (lex_cancel())
	{
		free_cmd(p_tmp,0);
		return PARSE_CANCELLED;
	}

}


/* -------------------------------------------------------------------------- *\
										An Attribute value
\* -------------------------------------------------------------------------- */

int attribute_val(t_attr_ref a_ref, t_val_node **pp_val_node)
{
	t_attr_types a_type;
	WS_get_attr_type(a_ref,&a_type);

	if (!(*pp_val_node = (t_val_node *)malloc(sizeof(t_val_node))))
	{
		errmsg("Out of Memory.");
		return PARSE_FAILED;
	}
	(**pp_val_node).p_next = 0;

	switch (a_type)
	{
		case T_USER:
		{
			char *a_val;
			if (TESLA_nominal_val(&a_val))
			{
				errmsg("Discrete attribute value expected");
				goto cancel;
			}
			else
			{
				int index;
				if (WS_get_val_index(a_ref,a_val,&index) == NO_ERROR) {
					alloc_string(&(**pp_val_node).val.discrete,a_val); }
				else 
				{
					char msg[256];
					sprintf(msg,"Invalid attribute value: %s",a_val);
					errmsg(msg);
					goto cancel;
				}	
			}
		} break;
		case T_FLOAT:
		{
			if (TESLA_cont_val(&(**pp_val_node).val.f))
			{
				errmsg("Real number expected");
				goto cancel;
			}		
		} break;
		case T_BINARY:
		case T_INT:	
		{
			if (TESLA_int_val(&(**pp_val_node).val.i))
			{
				errmsg("Number expected");
				goto cancel;
			}	 
			else if (a_type == T_BINARY)
			{
				if ((**pp_val_node).val.i != 0 && (**pp_val_node).val.i != 1)
					goto cancel;
			}
		} break;
		default: errmsg("Internal error: No such case."); goto cancel;
	}
	return PARSE_OK;

	cancel:
	free(*pp_val_node);
	return PARSE_FAILED;
}


/* -------------------------------------------------------------------------- *\
										Parse a single SELECT-test
\* -------------------------------------------------------------------------- */

int parse_sel_test(t_select_node **pp_select)
{
	char *attr_name;
	int limit;
	t_val_node **pp_val_node;
	t_attr_ref a_ref;
	t_attr_types a_type;

	if (TESLA_attribute_name(&attr_name) != PARSE_OK)
		goto cancel;
	else
	{
		if (!(*pp_select = (t_select_node *)malloc(sizeof(t_select_node))))
		{
			errmsg("Out of memory.");
			goto cancel;
		}
		(**pp_select).p_next = 0;
		(**pp_select).val_list = 0;
		(**pp_select).neg = 0;
		alloc_string(&(**pp_select).attr_name, attr_name);				
		if (WS_get_attr_ref((**pp_select).attr_name,&a_ref) != NO_ERROR)
			return PARSE_FAILED;
		WS_get_attr_type(a_ref,&a_type);	
		lex_next(); if (lex_cancel()) goto cancel;
	}

	if (TESLA_is_reserved_word(not_sym))
	{ 
		(**pp_select).neg = 1;
		lex_next(); if (lex_cancel()) goto cancel;
	}

	/* IN is the only allowed relation now */
	if (TESLA_is_reserved_word(in_sym))
	{
		(**pp_select).rel = IN;
		lex_next(); if (lex_cancel()) goto cancel;
	}

	if (TESLA_is_separator('['))
		(**pp_select).start_encl = 1;
	else if (TESLA_is_separator(']'))
		(**pp_select).start_encl = 0;
	else	
	{

		errmsg("[ or ] expected.");
		return PARSE_CANCELLED;
	}
	
	pp_val_node = &(**pp_select).val_list;

	if (a_type == T_USER || a_type == T_BINARY)
		/* parse as many attribute values as there is */
		do
		{
			lex_next(); if (lex_cancel()) goto cancel;
			if (attribute_val(a_ref,pp_val_node) != PARSE_FAILED)
			{
				pp_val_node = &(**pp_val_node).p_next;
				lex_next(); if (lex_cancel()) goto cancel;	
			}
		} while (TESLA_is_separator(','));
	else /* int or float */
	{
		/* lower boundary */
		lex_next(); if (lex_cancel()) goto cancel;
		if (TESLA_is_separator(','))
			(**pp_select).start_encl = -1; /* -oo */
		else if (attribute_val(a_ref,pp_val_node) != PARSE_FAILED)
		{
			pp_val_node = &(**pp_val_node).p_next;	
			lex_next(); if (lex_cancel()) goto cancel;
		}
		else
		{	
			errmsg("Attribute value or , excepted");
			goto cancel;
		}

			if (!TESLA_is_separator(','))
			{
				errmsg(", expected");
				goto cancel;
			}

		/* upper boundary */

		lex_next(); if (lex_cancel()) goto cancel;
		if (TESLA_is_separator('[') || TESLA_is_separator(']'))
		{
			(**pp_select).end_encl = -1; /* -oo */
			lex_null();
			return PARSE_OK;
		}
		else if (attribute_val(a_ref,pp_val_node) == PARSE_FAILED)
		{	
			errmsg("Attribute value, [ or ] excepted");
			goto cancel;
		}

		lex_next(); if (lex_cancel()) goto cancel;

	}
	

	
	if (TESLA_is_separator(']'))
		(**pp_select).end_encl = 1;
	else if (TESLA_is_separator('['))
		(**pp_select).end_encl = 0;
	else
	{ 
		errmsg("] or [ expected");
		goto cancel;
	}
		
	return PARSE_OK;

	cancel:
	/* free the command structure here */	
	
	return PARSE_FAILED;	
}


/* -------------------------------------------------------------------------- *\
											Parse a SELECT-command
\* -------------------------------------------------------------------------- */

int parse_select(t_command **pp_cmd)
{
	t_select_node **pp_select;

	
	if (!(*pp_cmd = alloc_cmd() ) ) return PARSE_FAILED;
	(**pp_cmd).cmd_id = K_SELECT;

	pp_select = &(**pp_cmd).param.l_select.p_first;

	do
	{
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (parse_sel_test(pp_select)==PARSE_FAILED) 
		{
			errmsg("Test Expected.");
			goto cancel;
		}
		pp_select = &(*pp_select)->p_next;
		lex_next(); if (lex_cancel()) goto cancel;
	} while (TESLA_is_separator(','));

	if (!TESLA_is_separator(';'))
	{
		errmsg("; expected.");
	}
	else
		lex_null();

	cancel:
	if (lex_cancel())
	{
		free_cmd(*pp_cmd,0);
		return PARSE_CANCELLED;
	}

	return PARSE_OK;
}


/* -------------------------------------------------------------------------- *\
													Parse a CV-command
\* -------------------------------------------------------------------------- */

int parse_cv(t_command **pp_cmd)
{
	char *token,*stat;
	FILE *fp;
	t_stats_node *p_stats_node;
	t_algorithms alg;
	t_cv *p_cv;
	int count;

	lex_next(); if (lex_cancel()) return PARSE_CANCELLED;

	if (TESLA_nat_val(&count) == PARSE_FAILED)
	  count = 10;
	else
	{
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED; 
	}

	if (!TESLA_word(&token) ? !ADB_get_alg_ref(token,&alg) : 1)
	{
		errmsg("No such algorithm");
		return PARSE_CANCELLED;															
	}

	if (!(*pp_cmd = alloc_cmd()))
	{
    errmsg ("Out of memory");   
    return PARSE_FAILED; 				
	}

	(**pp_cmd).cmd_id = K_CV;
	p_cv = &(**pp_cmd).param.l_cv;
	p_cv->alg = alg;
	p_cv->count = count;

	lex_next(); if (lex_cancel()) goto cancel;

	/* optional algorithm input parameters */
	if (!alg_params(alg,&(p_cv->params)))
	{
		lex_next(); if (lex_cancel()) goto cancel;
	}

	p_cv->pr_mode = 0;

	if (TESLA_is_reserved_word(record_sym))
	{
		p_cv->pr_mode = 1; /* independent output of statistics */
		do
		{
			lex_next(); if (lex_cancel()) goto cancel;		
			if (TESLA_is_reserved_word(time_sym))
			{
				p_stats_node = STATS_create_node("TIME");
				list_insert(&p_cv->stats,p_stats_node);
			}
			else if (TESLA_is_reserved_word(testtime_sym))
			{		
				p_stats_node = STATS_create_node("TESTTIME");
				list_insert(&p_cv->stats,p_stats_node);
			}
			else if (!TESLA_word(&stat))
			{
				if (ADB_alg_statistic_ok(stat,alg,0) || 
						ADB_alg_statistic_ok(stat,alg,1))
				{
					p_stats_node = STATS_create_node(stat);
					list_insert(&p_cv->stats,p_stats_node);
					free(stat);
				}
				else
				{
					errmsg("No such statistic");
					return PARSE_FAILED;
				}
			} 
			else if (TESLA_is_reserved_word(into_sym)) break;
			else if (TESLA_is_separator(';')) 
			{
				lex_null(); 
				return PARSE_OK;
			}
			else
			{
				errmsg("No such statistic");
				goto cancel;
			}
		}	 
		while (1);

	/* we have parsed INTO, if get here, now get the filename */
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
		if (TESLA_filename(alg,FT_STATS,&p_cv->stat_file_name) == PARSE_FAILED)
		{
			errmsg("Filename expected");
			goto cancel;	
		}
		
		lex_next(); if (lex_cancel()) return PARSE_CANCELLED;
	} /*
	else
	{
		errmsg("No statistics to record");
		goto cancel;
	} */

	if (!TESLA_is_separator(';'))
	{
		errmsg("Semicolon expected");
		goto cancel;
	}

	if (p_cv->stat_file_name)
	{
		if (fp = STATS_open_file(p_cv->stat_file_name))
			fclose(fp);
		else goto cancel;
	}




		lex_null();
		return PARSE_OK;

	cancel:
	if (lex_cancel())
	{
		free_cmd(*pp_cmd,0);
		return PARSE_CANCELLED;
	}

}


/* -------------------------------------------------------------------------- *\
												Parse a SET-command
\* -------------------------------------------------------------------------- */

int parse_set(t_command **pp_cmd)
{
	int type;

  if (!(*pp_cmd = alloc_cmd() ) ) return PARSE_FAILED;

	lex_next(); if (lex_cancel()) goto cancel;

	if (TESLA_symbol(&type) != PARSE_OK)
		type = -1;

	lex_next(); if (lex_cancel()) goto cancel;

	switch (type)
	{
		case randseed_sym:
			(**pp_cmd).cmd_id = K_SET_RANDSEED;
			if (TESLA_int_val(&(**pp_cmd).param.l_set.seed) != PARSE_OK)
			{
				if (TESLA_is_reserved_word(clock_sym))
				{
				  (**pp_cmd).cmd_id = K_SET_RANDCLOCK;
				}
				else
				{
					errmsg("Integer value or CLOCK expected");
					goto cancel;
				}
			}			
		break;
		case algorithm_sym:
			(**pp_cmd).cmd_id = K_SET_ALG;
			if (parse_algorithm(&(**pp_cmd).param.l_set.alg) != PARSE_OK)
			{
				errmsg("Algorithm name expected");
				goto cancel;
			}
		break;
	  default:
		{
			errmsg("RANDSEED or ALGORITHM expected");
			goto cancel;
		}
	}

	lex_next(); if (lex_cancel()) goto cancel;
	if (!TESLA_is_separator(';'))
	{
		errmsg("Semicolon expected");
		goto cancel;
	}

	return PARSE_OK;

	cancel:

	free_cmd(*pp_cmd,0);

	return PARSE_CANCELLED;

}


/* -------------------------------------------------------------------------- *\
												Parse a QUIT-command
\* -------------------------------------------------------------------------- */

int parse_quit(t_command **pp_cmd)
{
	*pp_cmd = alloc_cmd();

	(**pp_cmd).cmd_id = K_QUIT;

	lex_next(); if (lex_cancel()) return PARSE_CANCELLED;

	if (TESLA_is_separator(';'))
	{
		lex_null();
		return PARSE_OK;
	} 
	else
	{
		errmsg("Semicolon expected");
		return PARSE_FAILED;
	}
}

