/*                                                                            *\

    TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

    Module:       WS_Database

    Contents:      TELA Working Set Database

    Services:

  
    Author: Marko Nyman (1992), Ilkka Hyvarinen (1992,1993,1994),
            Juho Rousu (1995,1996)

\*                                                                            */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>

#include "tela.h"
#include "Alg_Database.h"
#include "WS_Database.h"

/* initialize: no classifier present */
t_rule_base rule_base = {0,0};

#define ALLOCSTRING(dest,source) \
  (dest) = (char *) malloc (strlen (source) + 1); \
  if ((dest) == NULL) { \
    errmsg ("Out of memory"); \
  } \
  strcpy ((dest), (source));

/* #defines */
#define private static

typedef struct 
{
  int    WS_attr_iter_by_read_order, 
        WS_attr_iter_by_name, 
        WS_attr_count, 
        WS_first_read_attr, 
        WS_exs_count,
        WS_class_attr;
  t_attr_node  *WS_attr_table;
} tVStruct;

/* Global variables. */
t_attr_node   *WS_attr_table = NULL;   /* osoitin attr rakenteen alkuun */

t_exs_db *WS_p_exs_db_list = 0;
int WS_exs_db_count = 0;

/* Moduulin sisiset globaalit muuttujat */
private int   WS_attr_count = 0;      /* Attribuuttien lkm */
private int   WS_exs_count = 0;      /* Ja esimerkkien */
private int   WS_first_read_attr = -1;      /* 1. luettu attr */
private int   WS_class_attr = -1;    /* Nyk. luokka attr */

/* iteraattorimuuttujat attribuuttien selausta varten. */
private int WS_attr_iter_by_name = -1;
private int WS_attr_iter_by_read_order = -1;




/* ====================================================================== *\
                          INTERNAL FUNCTIONS
\* ====================================================================== */

private int handle_err (char *caller, int errCode);
private int attr_cmp (const void *key, const void *a_item);
private int attr_binary_search (char *a_name);
private int attr_val_cmp (const void *avain, const void *value_node);
private int binary_search_val_index (char *mJono, t_attr_node *attr);
private int map_ok_exval (t_ex_node source, t_ex_node *dest,
             t_attr_node oldAttr, t_attr_node *newAttr,
             t_map_node *mList, char *elseVal);
private void table_cmp_init (tVStruct *v);
private void table_cmp_restore (tVStruct v);


/* -------------------------------------------------------------------------- *\
        Internal error handler for WS Database
\* -------------------------------------------------------------------------- */

private int handle_err (char *caller, int errCode) {
  
#define NBR_ERR_MSGS 26

struct vTauluAlkio {
  char *errMsg;
  int errClass;
};

static struct vTauluAlkio vTaulu[NBR_ERR_MSGS] = {
  "",                 NO_ERROR, /*  0 */
  "Internal error: NULL pointer not allowed!",      FATAL_ERROR, /*  1 */
  "Internal error: Memory allocation failed!",      FATAL_ERROR, /*  2 */
  "Attribute with same name already exists",      NON_FATAL_ERROR, /*  3 */
  "Name of an attribute can't be an empty string",    NON_FATAL_ERROR, /*  4 */
  "Internal error: Attribute handle out of range",    FATAL_ERROR, /*  5 */
  "No such attribute",          NON_FATAL_ERROR, /*  6 */
  "No attributes currently loaded",        NON_FATAL_ERROR, /*  7 */
  "Last attribute already returned",        NON_FATAL_ERROR, /*  8 */
  "The class-attribute is not specified",      NON_FATAL_ERROR, /*  9 */
  "Can't use a numeric attribute as the class attribute"
    ,    NON_FATAL_ERROR, /* 10 */
  "Internal error: Operation valid only for discrete attributes!", FATAL_ERROR, /* 11 */
  "Internal error: Value index out of range!",      FATAL_ERROR, /* 12 */
  "Internal error: Example row out of range!",      FATAL_ERROR, /* 13 */
  "Internal error: Attempt to get an unknown or not-cared-about value!", FATAL_ERROR, /* 14 */
  "Internal error: Operation valid only for type INT!",    FATAL_ERROR, /* 15 */
  "Internal error: Operation valid only for type FLOAT!",  FATAL_ERROR, /* 16 */
  "No examples currently loaded",        NON_FATAL_ERROR, /* 17 */
  "Last example has been returned already",      NON_FATAL_ERROR, /* 18 */
  "Internal error: Invalid seach mode!",      FATAL_ERROR, /* 19 */
  "BINARY-type attribute already has 2 values",    NON_FATAL_ERROR, /* 20 */
  "Invalid attribute value",  NON_FATAL_ERROR,  /* 21 */
  "Internal error: Can't free single attribute node within array!", FATAL_ERROR,  /* 22 */
  "Internal error: Link structure of attribute table scrambled!", FATAL_ERROR,  /* 23 */
  "Discrete value already exists",        NON_FATAL_ERROR, /* 24 */
  "All old values not specified in MAP!",      FATAL_ERROR
};

  /* Begin code */
  if (errCode) {
    if (vTaulu[errCode].errClass == FATAL_ERROR)
      errmsg(vTaulu[errCode].errMsg);
    else
      errmsg(vTaulu[errCode].errMsg);
  }
  return (vTaulu[errCode].errClass);
}


/* -------------------------------------------------------------------------- *\
      Compare an attribute node against a key     
\* -------------------------------------------------------------------------- */

private int attr_cmp (const void *key, const void *a_item) 
{
  return eq_q_str((char*)key,*(char**)a_item);
}


/* -------------------------------------------------------------------------- *\
      Find the attribute of given name from the attribute table
\* -------------------------------------------------------------------------- */

private int attr_binary_search (char *a_name) 
{
  t_attr_node *aOs;

  if ((aOs = (t_attr_node *) 
    bsearch ((const void *) a_name, (const void *) WS_attr_table, 
      WS_attr_count, sizeof (t_attr_node), attr_cmp)) == NULL)
    return (-1);
  else 
    return (aOs - WS_attr_table);
}



/* -------------------------------------------------------------------------- *\
              Compare an attribute node against a key       
\* -------------------------------------------------------------------------- */

private int attr_val_cmp (const void *avain, const void *value_node) 
{
  int  retVal;

  retVal = eq_q_str ((char*)avain, ((t_discrete_val *)value_node)->val);
  return retVal;
}


/* -------------------------------------------------------------------------- *\
              binary_search_val_index
\* -------------------------------------------------------------------------- */

private int binary_search_val_index (char *mJono, t_attr_node *attr) {
  
t_discrete_val *aOs;
  
  if ((aOs = (t_discrete_val *) 
             bsearch ((const void *) mJono, 
              (const void *) attr->disc_val_table,
           attr->val_cnt, 
           sizeof (t_discrete_val), attr_val_cmp)) == NULL)
    return (-1);
  else
    return (aOs - (attr->disc_val_table));
}


/* -------------------------------------------------------------------------- *\
            Perform a map for a single example column cell
\* -------------------------------------------------------------------------- */

private int map_ok_exval (t_ex_node source, t_ex_node *dest,
             t_attr_node oldAttr, t_attr_node *newAttr,
             t_map_node *mList, char *elseVal) 
{

  t_attr_types  aType;  
  float    floatVal;
  int    found, intVal;
  t_map_node   *wPtr;

  dest->ok = T_OK; aType = oldAttr.a_type;

  switch (aType) 
  {
    case T_USER :
    case T_BINARY :
      /* only allowed relation is equality (==) */
      found = 0; wPtr = mList;
      while (!found && (wPtr != NULL)) 
      {
        found = !eq_q_str(oldAttr.disc_val_table[source.val.int_val].val,
             wPtr->old_val.discrete);
        if (!found)
          wPtr = wPtr->next;
      }
      break; 
    case T_INT :
    case T_FLOAT :
      found = 0; wPtr = mList;
      while (!found && (wPtr != NULL)) 
      {
        if (aType == T_INT && wPtr->old_type == l_float) 
        {
          wPtr->old_type = l_int;
          wPtr->old_val.i = (int) wPtr->old_val.f;
        }
   
        switch (wPtr->relation) 
        {
          case LT :
            found = 
              ((aType == T_INT) ? 
                 (source.val.int_val < wPtr->old_val.i) :
               (source.val.float_val < wPtr->old_val.f));
            break;
          case EQ:
            found = 
              ((aType == T_INT) ? 
                 (source.val.int_val == wPtr->old_val.i) :
               (source.val.float_val == wPtr->old_val.f));
            break;
          default :
            found = 
              ((aType == T_INT) ? 
                 (source.val.int_val <= wPtr->old_val.i) :
               (source.val.float_val <= wPtr->old_val.f));
            break;
        }
        
        if (!found)
          wPtr = wPtr->next;
      } 
      break;  
  }
          
  /* Set value in new column */
  if (found) 
  {
    if ((dest->val.int_val = 
        binary_search_val_index (wPtr->new_val, newAttr)) == -1)
      return (handle_err ("map_domain", 12));
  } 
  else 
  {
    if (elseVal == NULL || !strcmp (elseVal, ""))
    {
      char msg[256];
      strcpy(msg,"Unspecified value: ");
      switch (aType)
      {
        case T_USER: case T_BINARY:
          sprintf(msg+strlen(msg),"%s",
            oldAttr.disc_val_table[source.val.int_val].val);
        break;
        case T_INT: sprintf(msg+strlen(msg),"%d",source.val.int_val); break;
        case T_FLOAT: sprintf(msg+strlen(msg),"%f",source.val.float_val); break;
      }
      errmsg(msg);
      return EXECUTION_OK;
    }
 
   if ((dest->val.int_val = binary_search_val_index (elseVal, newAttr)) == -1)
      return (handle_err ("map_domain", 12));
  }

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
            Initialize a attribute table for comparison
\* -------------------------------------------------------------------------- */

private void table_cmp_init (tVStruct *v) 
{
  v->WS_attr_table = WS_attr_table; v->WS_attr_count = WS_attr_count;
  v->WS_exs_count = WS_exs_count; v->WS_first_read_attr = WS_first_read_attr;
  v->WS_class_attr = WS_class_attr; v->WS_attr_iter_by_name = WS_attr_iter_by_name;
  v->WS_attr_iter_by_read_order = WS_attr_iter_by_read_order;
}


/* -------------------------------------------------------------------------- *\
              Restore a stored attribute table
\* -------------------------------------------------------------------------- */

private void table_cmp_restore (tVStruct v) 
{
  
int  i, j;

  WS_dispose_dbase ();
  
  WS_attr_table = v.WS_attr_table; WS_attr_count = v.WS_attr_count;
  WS_exs_count = v.WS_exs_count; WS_class_attr = v.WS_class_attr;
  WS_attr_iter_by_name = v.WS_attr_iter_by_name; WS_attr_iter_by_read_order = v.WS_attr_iter_by_read_order;
  WS_first_read_attr = v.WS_first_read_attr;
}
  

/* ========================================================================== *\
                            EXTERNAL SERVICES
\* ========================================================================== */


/* -------------------------------------------------------------------------- *\
                    Set the type of the active classfier
\* -------------------------------------------------------------------------- */

void WS_set_active_classifier(t_algorithms alg)
{
  rule_base.cf_type = alg;
  if (alg) rule_base.cf_exists = 1;
  return;
}


/* -------------------------------------------------------------------------- *\
                       Get the type of the active classfier
\* -------------------------------------------------------------------------- */

int WS_get_active_classifier(t_algorithms *alg)
{
  *alg = rule_base.cf_type;

  if (!rule_base.cf_exists) return 0;

  return 1;
}


/* -------------------------------------------------------------------------- *\
                          Remove an existing classifier      
\* -------------------------------------------------------------------------- */

void WS_remove_active_classifier()
{
  rule_base.cf_exists = 0;
  rule_base.cf_type = 0;
  return;
}
 
 
/* -------------------------------------------------------------------------- *\
        Back up the current learning result for later restoring
\* -------------------------------------------------------------------------- */

int WS_rule_backup(t_algorithms alg,char *fname)
{
  FILE *bkp = fopen(fname,"w");

  if (IO_write_rule(bkp,alg,0))
  {
    fclose(bkp);
    return EXECUTION_OK;
  }
  else 
  {
    fclose(bkp);
    return EXECUTION_FAILED;
  }
}


/* -------------------------------------------------------------------------- *\
                  Restore the previously stored classifier
\* -------------------------------------------------------------------------- */

int WS_rule_restore(t_algorithms alg,char *fname)
{
  FILE *rule_file = fopen(fname,"r");

  if (IO_read_rule(rule_file,alg))
  {
    fclose(rule_file);
    WS_set_active_classifier(alg);
    return EXECUTION_OK;
  }
  
  fclose(rule_file);
  return EXECUTION_FAILED;
}


/* -------------------------------------------------------------------------- *\
        Output a filename for the internal storage for a classifier
\* -------------------------------------------------------------------------- */

void WS_get_internal_rulename(char *buf)
{
  /* keep internal rulefiles uniquely named by augmenting by pid */
  sprintf(buf,".TELA_%ld_cf",getpid());
  return;
}


/* -------------------------------------------------------------------------- *\
            Get the number of attribute in active working set    
\* -------------------------------------------------------------------------- */

int WS_get_attr_cnt (void) 
{
  return WS_attr_count;
}


/* -------------------------------------------------------------------------- *\
            Get the number of examples in active working set      
\* -------------------------------------------------------------------------- */

int WS_get_ex_cnt (void) 
{
  return WS_exs_count;
}


/* -------------------------------------------------------------------------- *\
            Set the number of examples in active working set

    Note: this function does not free or allocate memory and does not 
    quarantee that the example count makes sense. Use with caution!

\* -------------------------------------------------------------------------- */

int WS_set_ex_cnt (int e_count) 
{

  WS_exs_count = e_count;
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
            Delete current example set and attribute description      
\* -------------------------------------------------------------------------- */

void WS_dispose_dbase (void) 
{
  
  int  i, j;
  
  if (!WS_attr_table) return;

  for (i = 0; i < WS_attr_count; i++) {
    if (WS_attr_table[i].examples != NULL)
      free (WS_attr_table[i].examples);
    if (WS_attr_table[i].a_type == T_USER || 
        WS_attr_table[i].a_type == T_BINARY) 
    {
      for (j = 0; j < WS_attr_table[i].val_cnt; j++)
        free (WS_attr_table[i].disc_val_table[j].val);
      free (WS_attr_table[i].disc_val_table);
    }
  }
  free (WS_attr_table);
  
  WS_attr_table = 0; 

  WS_first_read_attr = WS_attr_iter_by_name = 
    WS_attr_iter_by_read_order = WS_class_attr = -1;

  WS_attr_count = WS_exs_count = 0;
}


/* -------------------------------------------------------------------------- *\
                    Create a new attribute node
\* -------------------------------------------------------------------------- */

int WS_create_attr (char *a_name, t_attr_types a_type, t_attr_node **a_node) 
{ 

  static char rutname[] = "create_attr";

  /* Koodi */
   if ((*a_node = (t_attr_node *) malloc (sizeof (t_attr_node))) == NULL)
    return (handle_err (rutname,  2));
    
  /* Init. kentt */
  alloc_string(&(*a_node)->name,a_name);
  (*a_node)->a_type = a_type;
  (*a_node)->val_cnt = 0;
  (*a_node)->first_read = -1;
  (*a_node)->next_attr = -1;
  (*a_node)->disc_val_table = NULL;
  (*a_node)->examples = NULL;

  return NO_ERROR;
}


/* -------------------------------------------------------------------------- *\
                Free an attribute node

    Note: use only for nodes that are not attached into the attribute table
\* -------------------------------------------------------------------------- */

int WS_free_attr (t_attr_node *a_node) 
{
  
  static char   rutname[] = "free_attr";
  int    i;

  if (WS_attr_table <= a_node && a_node < (WS_attr_table + WS_attr_count))
    return (handle_err (rutname, 22));
       
  if (a_node->examples != NULL)
    free (a_node->examples);
  if (a_node->a_type == T_USER || a_node->a_type == T_BINARY) {
    for (i = 0; i < a_node->val_cnt; i++)
      free (a_node->disc_val_table[i].val);
    free (a_node->disc_val_table);
  }
  
  free (a_node);
  return (NO_ERROR);
}
  

/* -------------------------------------------------------------------------- *\
            Add an attribute node into the attribute table
\* -------------------------------------------------------------------------- */

int WS_add_attr (t_attr_node *a) 
{
  static char rutname[] = "add_attr";

  int   i, place;
  char  *tmp1, *tmp2;

  /* Here we go ... */
  if (a == NULL)
    return (handle_err (rutname, 1));
  if (!strcmp (a -> name, ""))
    return (handle_err (rutname, 4));
  if (WS_attr_table != NULL && attr_binary_search (a->name) >= 0)
    return (handle_err (rutname, 3));
    
  if ((WS_attr_table = 
    (t_attr_node *) realloc (WS_attr_table, (WS_attr_count + 1) * 
                            sizeof (t_attr_node))) == NULL)
    return (handle_err (rutname, 2));
  

   /* Find a place for the attribute */

   place = 0;
 
   if (WS_get_attr_cnt ()) 
  {
    while ((place < WS_attr_count) && 
      (eq_q_str(WS_attr_table[place].name, a->name)) <= 0) 
      place++;
  }
    
  /* Move the attributes following the insertion place */
  for (i = WS_attr_count - 1; i >= place; i--) 
  {
    if (WS_attr_table[i].next_attr >= place)
      WS_attr_table[i].next_attr++;
    else if (WS_attr_table[i].next_attr == -1)
      WS_attr_table[i].next_attr = place;
    WS_attr_table[i + 1] = WS_attr_table[i];
  }
  
  /* Ensure that also the preceding links point correctly */
  for (i = 0; i < place; i++) 
  {
    if (WS_attr_table[i].next_attr >= place)
      WS_attr_table[i].next_attr++;
    else if (WS_attr_table[i].next_attr == -1)
      WS_attr_table[i].next_attr = place;
  }
  
  /* put the new attribute node in its place */
  WS_attr_table[place] = *a;
  WS_attr_table[place].next_attr = -1;
  if ((WS_attr_table[place].a_type == T_USER ||
       WS_attr_table[place].a_type == T_BINARY) &&
      WS_attr_table[place].disc_val_table != NULL) {
    for (i = 0; i < WS_attr_table[place].val_cnt; i++) {
      alloc_string(&WS_attr_table[place].disc_val_table[i].val,
				a->disc_val_table[i].val);
      WS_attr_table[place].disc_val_table[i].next_val =
 				a->disc_val_table[i].next_val;
    }
  }

  WS_attr_count++;
  if (WS_first_read_attr == -1)
    WS_first_read_attr = 0;
  else if (WS_first_read_attr >= place)
    WS_first_read_attr++;
  if (WS_class_attr != -1 && (WS_class_attr >= place))
    WS_class_attr++;
  
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                      Create and Add a new attribute      
\* -------------------------------------------------------------------------- */

int WS_create_and_add_attr (char *a_name, t_attr_types a_type, 
	t_attr_ref *a_ref) 
{ 

  static char rutname[] = "LuoJaadd_attr";
  t_attr_node a_node;
  int retVal;

  /* Init. kentt */
  alloc_string(&a_node.name,a_name);
  a_node.a_type = a_type;
  a_node.disc_val_table = NULL; a_node.examples = NULL;
  a_node.val_cnt = 0;
  a_node.first_read = a_node.next_attr = -1;
  
  /* Listn attribuutti ja haetaan sen viite */
  if (retVal = WS_add_attr (&a_node))
    return (retVal);

  if ((*a_ref = attr_binary_search (a_name)) == -1)
    return (handle_err (rutname, 6));
  else 
    return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Remove an attribute from the attribute table      
\* -------------------------------------------------------------------------- */

int WS_remove_attr (t_attr_ref a_ref) 
{
   static char rutname[] = "remove_attr";
  int i, linkSave;

  if (a_ref < 0 || a_ref > WS_attr_count - 1)
    return (handle_err (rutname, 5));
     
  if (WS_attr_table[a_ref].examples != NULL)
    free (WS_attr_table[a_ref].examples);
    
  /* Free discrete value table, is applicable */
  if (WS_attr_table[a_ref].a_type == T_USER || 
      WS_attr_table[a_ref].a_type == T_BINARY)
    for (i = 0; i < WS_attr_table[a_ref].val_cnt; i++)
      free (WS_attr_table[a_ref].disc_val_table[i].val);
      
  free (WS_attr_table[a_ref].name);
     
  /* Reorganize the attribute table */
  linkSave = (WS_attr_table[a_ref].next_attr < a_ref) ? 
     WS_attr_table[a_ref].next_attr : WS_attr_table[a_ref].next_attr - 1;
  for (i = a_ref + 1; i < WS_attr_count; i++) 
  {
    if (WS_attr_table[i].next_attr == a_ref)
      WS_attr_table[i].next_attr = linkSave;
    else if (WS_attr_table[i].next_attr > a_ref)
      WS_attr_table[i].next_attr--;
    WS_attr_table[i - 1] = WS_attr_table[i];
  }
  
  for (i = 0; i < a_ref; i++) 
  {
    if (WS_attr_table[i].next_attr == a_ref)
      WS_attr_table[i].next_attr = linkSave;
    else if (WS_attr_table[i].next_attr > a_ref)
      WS_attr_table[i].next_attr--;
  }
  
  /* Ja glob muuttujat */
  WS_attr_count--;
  if (WS_first_read_attr == a_ref)
    WS_first_read_attr = linkSave;
  else if (WS_first_read_attr > a_ref)
    WS_first_read_attr--;
  if (WS_class_attr != -1 && (WS_class_attr > a_ref))
    WS_class_attr--;
  else if (WS_class_attr == a_ref)
    WS_class_attr = -1;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                Get a reference to an attribute by name
\* -------------------------------------------------------------------------- */

int WS_get_attr_ref (char *a_name, t_attr_ref *a_ref) 
{
  static char rutname[] = "get_attr_ref";
  t_attr_node *aOs;

  /* Virhetarkistukset */
  if (WS_attr_table == NULL)
    return (handle_err (rutname, 7));
  if (!strcmp (a_name, ""))
    return (handle_err (rutname, 6));
  
  /* Vars. ty, haetaan attribuutti binaarihakua kytten */
  if ((*a_ref = attr_binary_search (a_name)) == -1)
  {
    char buf[256];
    sprintf(buf,"No such attribute: %s",a_name);
    errmsg(buf);
    return EXECUTION_FAILED;
  }
  else 
    return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Is given name already in use or otherwise reserved      
\* -------------------------------------------------------------------------- */

int WS_is_attr_name_used (char *name)
{
  if (WS_attr_table == NULL)
    return 0;
  else if (!strcmp (name, "")) /* empty name is always reserved */
    return 1;
  else if (!strcmp (name, "class"))
    return 1;

  if ((attr_binary_search (name)) == -1)
    return 0;
  else
    return 1;
}


/* -------------------------------------------------------------------------- *\
              Get an attribute's name by reference
\* -------------------------------------------------------------------------- */

int WS_get_attr_name (t_attr_ref a_ref, char **a_name) 
{
  
  static char rutname[] = "WS_get_attr_name";

  if (a_ref < 0 || a_ref > (WS_attr_count - 1))
    return (handle_err (rutname, 5));
  
  *a_name = WS_attr_table[a_ref].name;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Replace an attribute node with a new one       
\* -------------------------------------------------------------------------- */

int WS_replace_attr (t_attr_ref old_attr, t_attr_node *new_attr)
{
  
  static char   rutname[] = "replace_attr";
  int    i;

  if (new_attr == NULL)
    return (handle_err (rutname, 1));
  if (old_attr < 0 || old_attr >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (strcmp (new_attr->name, WS_attr_table[old_attr].name)) 
    return (handle_err (rutname, 4));
    
  /* Demolish old attribute node */
  if (WS_attr_table[old_attr].examples != NULL)
    free (WS_attr_table[old_attr].examples);
  if (WS_attr_table[old_attr].a_type == T_USER || 
      WS_attr_table[old_attr].a_type == T_BINARY) 
  {
    for (i = 0; i < WS_attr_table[old_attr].val_cnt; i++)
      free (WS_attr_table[old_attr].disc_val_table[i].val);
  }
     
  new_attr->next_attr = WS_attr_table[old_attr].next_attr;
  WS_attr_table[old_attr] = *new_attr;
  
  alloc_string(&WS_attr_table[old_attr].name,new_attr->name);
  if (WS_attr_table[old_attr].a_type == T_USER || 
      WS_attr_table[old_attr].a_type == T_BINARY) 
  {
    for (i = 0; i < WS_attr_table[old_attr].val_cnt; i++)
      alloc_string(&WS_attr_table[old_attr].disc_val_table[i].val,
        new_attr->disc_val_table[i].val);
  }
  
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
           Get first attribute using given search mode
\* -------------------------------------------------------------------------- */

int WS_get_first_attr (t_attr_ref *a_ref, t_search_mode search_mode) 
{

  static char rutname[] = "get_first_attr";
  
  if (WS_attr_count == 0) 
    return (handle_err (rutname, 7));
  if (search_mode == A_JARJ)
    *a_ref = WS_attr_iter_by_name = 0;
  else if (search_mode == M_JARJ)
    *a_ref = WS_attr_iter_by_read_order = WS_first_read_attr;
  else
    return (handle_err (rutname, 19));

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Get next attribute using given search mode
\* -------------------------------------------------------------------------- */

int WS_get_next_attr (t_attr_ref *a_ref, t_search_mode search_mode) 
{
  static char rutname[] = "get_next_attr";

  switch (search_mode) {
    case A_JARJ : 
      if (WS_attr_iter_by_name == -1 || WS_attr_iter_by_name == WS_attr_count - 1)
        return (NON_FATAL_ERROR);
      else
        *a_ref = (t_attr_ref) ++WS_attr_iter_by_name;
      break;
    case M_JARJ : 
      if (WS_attr_iter_by_read_order == -1)
       return (NON_FATAL_ERROR);
      else
        if ((WS_attr_iter_by_read_order = WS_attr_table[WS_attr_iter_by_read_order].next_attr) == -1)
          return (1);
        else
        *a_ref = (t_attr_ref) WS_attr_iter_by_read_order;
      break;
    default : 
      return (handle_err (rutname, 19));
      break;
  }
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Get last attribute using given search mode      
\* -------------------------------------------------------------------------- */

int WS_get_last_attr (t_attr_ref *a_ref, t_search_mode search_mode) 
{
  static char   rutname[] = "get_last_attr";
  int i;
  
  if (WS_attr_count == 0) 
    return (handle_err (rutname, 7));
  if (search_mode == A_JARJ)
    *a_ref = WS_attr_count - 1;
  else if (search_mode == M_JARJ) {
    for (i = 0; i < WS_attr_count && WS_attr_table[i].next_attr != -1; i++);
    if (i < WS_attr_count)
      *a_ref = i;
    else
      return (handle_err (rutname, 23));
  } else
    return (handle_err (rutname, 19));
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Get the name of the current class attribute      
\* -------------------------------------------------------------------------- */

int WS_get_class_attr_name(char *p_buf)
{
  if (WS_is_class_attr())
  {
    char *p_tmp_buf;
    t_attr_ref a_ref;
    WS_get_class_attr(&a_ref);
    WS_get_attr_name(a_ref,&p_tmp_buf);
    strcpy(p_buf,p_tmp_buf);
    return 0;
  }
  return 1;  /* no previous class attribute */
}


/* -------------------------------------------------------------------------- *\
                    Is a class attribute defined
\* -------------------------------------------------------------------------- */

int WS_is_class_attr (void) {
  
  return (WS_class_attr != -1);
}


/* -------------------------------------------------------------------------- *\
              Get a reference to the class attribute
\* -------------------------------------------------------------------------- */

int WS_get_class_attr (t_attr_ref *a_ref) 
{
  
  static char rutname[] = "WS_get_class_attr";
  
  if (WS_attr_table == NULL)
    return handle_err (rutname, 7);
  else if (WS_class_attr == -1)
    return handle_err (rutname, 9);
    
  *a_ref = WS_class_attr;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
        Set the referenced attribute as the current class attribute 
\* -------------------------------------------------------------------------- */

int WS_set_class_attr (t_attr_ref a_ref) 
{
  
  static char rutname[] = "set_class_attr";
  
  if (a_ref != -1)
  {
    if (a_ref >= WS_attr_count)
      return handle_err (rutname, 5);
    else if (WS_attr_table[a_ref].a_type == T_FLOAT || 
             WS_attr_table[a_ref].a_type == T_INT)
      return handle_err (rutname, 10);
  }  
  
  WS_class_attr = a_ref;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                    Get the type of an attribute      
\* -------------------------------------------------------------------------- */

int WS_get_attr_type (t_attr_ref a_ref, t_attr_types *a_type) 
{  
  static char rutname[] = "get_attr_type";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
    
  *a_type = WS_attr_table[a_ref].a_type;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                    Set the type of an attribute
\* -------------------------------------------------------------------------- */

int WS_set_attr_type (t_attr_ref a_ref, t_attr_types a_type) 
{
  
  static char rutname[] = "set_attr_type";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
    
  WS_attr_table[a_ref].a_type = a_type;
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                            Copy an attribute node
\* -------------------------------------------------------------------------- */

int WS_copy_attr (t_attr_node *source, t_attr_node **dest) 
{
  
  int    i;
  static char   rutname[] = "WS_copy_attr";
  t_attr_node  *wPtr; /* Work Pointer, you see */

  if ((wPtr = (t_attr_node *) malloc (sizeof (t_attr_node))) == NULL)
    return (handle_err (rutname, 2));
    
  (*wPtr) = (*source);
  
  /* If this is a discrete attribute we must copy the value strings. */
  if (wPtr->a_type == T_USER || wPtr->a_type == T_BINARY) 
  {
    if (!(wPtr->disc_val_table = 
      (t_discrete_val *) malloc (wPtr->val_cnt * sizeof (t_discrete_val))) )
      return (handle_err (rutname, 2));
    for (i = 0; i < wPtr->val_cnt; i++) 
    {
      alloc_string(&wPtr->disc_val_table[i].val,source->disc_val_table[i].val);
      wPtr->disc_val_table[i].next_val = source->disc_val_table[i].next_val;
    }
  }

  *dest = wPtr;  

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Get the size of an attribute's domain
\* -------------------------------------------------------------------------- */

int WS_get_attr_domain_size (t_attr_ref a_ref, int *val_cnt) {
  
static char rutname[] = "get_attr_domain_size";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (WS_attr_table[a_ref].a_type != T_USER &&
      WS_attr_table[a_ref].a_type != T_BINARY)
    return (handle_err (rutname, 11));
    
  *val_cnt = WS_attr_table[a_ref].val_cnt;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Get attribute value corresponding to an index
\* -------------------------------------------------------------------------- */

int WS_get_val_index (t_attr_ref a_ref, char *mJono, int *index) 
{
  
  static char   rutname[] = "get_val_index";

  int    wInd, retVal;

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (WS_attr_table[a_ref].a_type != T_USER &&
      WS_attr_table[a_ref].a_type != T_BINARY)
    return (handle_err (rutname, 11));

  if ((wInd = binary_search_val_index (mJono, &WS_attr_table[a_ref])) == -1) {
    retVal = EXECUTION_FAILED;
    fprintf (stdout, "Received value was: %s\n", mJono);
    fprintf (stdout, "Allowed values are: \n");
    for (wInd = 0; wInd < WS_attr_table[a_ref].val_cnt; wInd++)
      fprintf (stdout, "   %s\n", 
        WS_attr_table[a_ref].disc_val_table[wInd].val);
    return retVal;
  }
    
  *index = wInd;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Is given value part of the attribute's domain              
\* -------------------------------------------------------------------------- */

int WS_is_attr_val_name(t_attr_ref a_ref,char *val_name)
{
  static char   rutname[] = "WS_is_attr_val_name";

  if (a_ref < 0 || a_ref > WS_attr_count)
  {
    handle_err (rutname, 5);
    return 0;
  }

  if (binary_search_val_index(val_name,&WS_attr_table[a_ref]) == -1)
    return 0;
  else
    return 1;
}


/* -------------------------------------------------------------------------- *\
        Get the attribute value corresponding to a given index                
\* -------------------------------------------------------------------------- */

int WS_get_index_attr_val (t_attr_ref a_ref, int index, char **mJono)
{
  
  static char rutname[] = "get_index_attr_val";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (index > WS_attr_table[a_ref].val_cnt - 1)
    return (handle_err (rutname, 12));
  if (WS_attr_table[a_ref].a_type != T_USER &&
      WS_attr_table[a_ref].a_type != T_BINARY)
    return (handle_err (rutname, 11));
    
  *mJono = WS_attr_table[a_ref].disc_val_table[index].val;
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                Add a new value to an attributes domain              
\* -------------------------------------------------------------------------- */

int WS_join_val_low (t_attr_node *a, char *new_value) 
{
  
  const int   step = 2;

  static char   rutname[] = "add_attr_val";
  int     place, i;
  char    *tmp1, *tmp2;

  if (a == NULL)
    return (handle_err (rutname, 1));
  if (!strcmp (new_value, ""))
    return (handle_err (rutname, 4));
  if (a->a_type != T_USER && a->a_type != T_BINARY)
    return (handle_err (rutname, 11));
  if (a->val_cnt > 0 && binary_search_val_index (new_value, a) >= 0)
    return (NO_ERROR);
  if (a->a_type == T_BINARY && a->val_cnt == 2)
    return (handle_err (rutname, 20));
   
  /* allocate memory if needed */ 
  if ((a->val_cnt % 2) == 0)
    if (!(a->disc_val_table =  (t_discrete_val *) realloc (a->disc_val_table,
          (a->val_cnt + step) * sizeof (t_discrete_val))))
      return (handle_err (rutname, 2));
          
  /* find a place for the value */
  place = 0;
  if (a->val_cnt) 
  {
    while ((place < a->val_cnt) && 
       (eq_q_str (a->disc_val_table[place].val, new_value)) <= 0) 
    {
      place++;
    }
  }
    
  /* Organize the table from the end to the insertion place */
  for (i = a->val_cnt - 1; i >= place; i--) {
    if (a->disc_val_table[i].next_val >= place)
      a->disc_val_table[i].next_val++;
    else if (a->disc_val_table[i].next_val == -1)
      a->disc_val_table[i].next_val = place;
    a->disc_val_table[i + 1] = a->disc_val_table[i];
  }
  
  /* Organize the table form start to the insertion place*/
  for (i = 0; i < place; i++) 
	{
    if (a->disc_val_table[i].next_val >= place)
      a->disc_val_table[i].next_val++;
    else if (a->disc_val_table[i].next_val == -1)
      a->disc_val_table[i].next_val = place;
  }
  
  /* Insert the item */
  alloc_string(&a->disc_val_table[place].val,new_value);
  a->disc_val_table[place].next_val = -1;
  a->val_cnt++;
  if (a->first_read == -1)
    a->first_read = 0;
  else if (a->first_read >= place)
    a->first_read++;
  
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Add a value to an attributes domain
\* -------------------------------------------------------------------------- */

int WS_add_attr_val (t_attr_ref a_ref, char *new_value) 
{
  return (WS_join_val_low (&WS_attr_table[a_ref], new_value));
}


/* -------------------------------------------------------------------------- *\
          Get the kind of an example value, Ok, Unknown of Don't Care  
\* -------------------------------------------------------------------------- */

int WS_get_exval_kind (t_attr_ref a_ref, int e_line, t_ex_ok *kind) 
{
  
  static char rutname[] = "get_exval_kind";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  
  *kind = WS_attr_table[a_ref].examples[e_line].ok;

  return (NO_ERROR);
}
  

/* -------------------------------------------------------------------------- *\
      Set the kind for an example value, Ok, Unknonw or Don't Care              
\* -------------------------------------------------------------------------- */

int WS_set_exval_kind (t_attr_ref a_ref, int e_line, t_ex_ok kind) {
  
static char rutname[] = "set_exval_kind";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  
  WS_attr_table[a_ref].examples[e_line].ok = kind;
  return (NO_ERROR);
}
  

/* -------------------------------------------------------------------------- *\
    Get the value of given nominal attribute on given example as an index          \* -------------------------------------------------------------------------- */

int WS_get_nominal_exval (t_attr_ref a_ref, int e_line, int *index) 
{
  
  static char rutname[] = "get_ex_val";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  if (WS_attr_table[a_ref].a_type != T_USER &&
      WS_attr_table[a_ref].a_type != T_BINARY)
    return (handle_err (rutname, 11));
  if (WS_attr_table[a_ref].examples[e_line].ok != T_OK)
    return (handle_err (rutname, 14));
    
  *index = WS_attr_table[a_ref].examples[e_line].val.int_val;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
          Set the value of given nominal attribute on given example              \* -------------------------------------------------------------------------- */

int WS_set_nominal_exval (t_attr_ref a_ref, int e_line, int index) 
{
  
  static char rutname[] = "set_ex_val";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (WS_attr_table[a_ref].a_type != T_USER &&
      WS_attr_table[a_ref].a_type != T_BINARY)
    return (handle_err (rutname, 11));
  if (index >= WS_attr_table[a_ref].val_cnt)
    return (handle_err (rutname, 12));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 13));
    
  WS_attr_table[a_ref].examples[e_line].val.int_val = index;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
      Get the value of given integer-valued attribute on given example      
\* -------------------------------------------------------------------------- */

int WS_get_integer_exval (t_attr_ref a_ref, int e_line, int *value)
{
  
  static char rutname[] = "get_integer_exval";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  if (WS_attr_table[a_ref].a_type != T_INT)
    return (handle_err (rutname, 15));
  if (WS_attr_table[a_ref].examples[e_line].ok != T_OK)
    return (handle_err (rutname, 14));
    
  *value = WS_attr_table[a_ref].examples[e_line].val.int_val;
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
      Set the value of given integer-valued attribute on given example    
\* -------------------------------------------------------------------------- */

int WS_set_integer_exval (t_attr_ref a_ref, int e_line, int value) {
  
  static char rutname[] = "set_integer_exval";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  if (WS_attr_table[a_ref].a_type != T_INT)
    return (handle_err (rutname, 15));
    
  WS_attr_table[a_ref].examples[e_line].val.int_val = value;
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
    Get the value of given continuous-valued attribute on given example      
\* -------------------------------------------------------------------------- */

int WS_get_float_exval (t_attr_ref a_ref, int e_line, float *value) {
  
  static char rutname[] = "get_float_exval";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  if (WS_attr_table[a_ref].a_type != T_FLOAT)
    return (handle_err (rutname, 16));
  if (WS_attr_table[a_ref].examples[e_line].ok != T_OK)
    return (handle_err (rutname, 14));
    
  *value = WS_attr_table[a_ref].examples[e_line].val.float_val;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
    Set the value of given continuous-valued attribute on given example      
\* -------------------------------------------------------------------------- */

int WS_set_float_exval (t_attr_ref a_ref, int e_line, float value) {
  
static char rutname[] = "set_float_exval";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (e_line >= WS_exs_count)
    return (handle_err (rutname, 12));
  if (WS_attr_table[a_ref].a_type != T_FLOAT)
    return (handle_err (rutname, 16));
    
  WS_attr_table[a_ref].examples[e_line].val.float_val = value;
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
     Create a column for example vector values of an attribute
\* -------------------------------------------------------------------------- */

int WS_create_ex_column (t_attr_types a_type, int e_count, t_ex_node **column) 
{
  
  static char rutname[] = "create_ex_column";

  if (e_count == 0)
    *column = NULL;
  else
    if (!(*column = (t_ex_node *) malloc (e_count * sizeof (t_ex_node))))
      return (handle_err (rutname, 2));

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
             Get a handle to the example value column
\* -------------------------------------------------------------------------- */

int WS_get_ex_column (t_attr_ref a_ref, t_ex_node **column) 
{
  
  static char rutname[] = "get_ex_column";

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (WS_exs_count == 0)
    return (handle_err (rutname, 17));
  
  *column = WS_attr_table[a_ref].examples;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                   Clone the example value column
\* -------------------------------------------------------------------------- */

int WS_clone_ex_column (t_attr_ref a_ref, t_ex_node *kohde, int lkm) 
{
  
  static char rutname[] = "clone_ex_column";
  register int i;

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (kohde == NULL)
    return (handle_err (rutname, 1));
  if (WS_exs_count == 0)
    return (handle_err (rutname, 17));
    
  for (i = 1; i <= lkm; i++) 
  {
    memcpy (kohde, WS_attr_table[a_ref].examples, 
          WS_exs_count * sizeof (t_ex_node));
    kohde += WS_exs_count;
  }

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
     Replace an existing example value column with a new one
\* -------------------------------------------------------------------------- */

int WS_replace_ex_column (t_attr_ref a_ref, t_ex_node *new_column) 
{
  
  static char rutname[] = "replace_ex_column";
  register int i;

  if (a_ref < 0 || a_ref >= WS_attr_count)
    return (handle_err (rutname, 5));
  if (new_column == NULL)
    return (handle_err (rutname, 1));

  free (WS_attr_table[a_ref].examples);
  WS_attr_table[a_ref].examples = new_column;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                      Move an example vector
\* -------------------------------------------------------------------------- */

int WS_move_ex_row (int source, int dest) 
{

  static char   rutname[] = "move_ex_row";
  t_attr_ref  aRef;  

  if (source == dest) return (NO_ERROR);
  if (source >= WS_exs_count)
    return (handle_err (rutname, 13));

  for (aRef = 0; aRef < WS_attr_count; aRef++)
    WS_attr_table[aRef].examples[dest] = WS_attr_table[aRef].examples[source];

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                      Add an example vector
\* -------------------------------------------------------------------------- */

int WS_add_ex_row (void) 
{
  
  static char rutname[] = "add_ex_row";
  register int i;

  if (WS_attr_count == 0) 
    return (handle_err (rutname, 7));
  
  for (i = 0; i < WS_attr_count; i++) 
  {
    if (WS_exs_count == 0) 
    {
      if ((WS_attr_table[i].examples = 
           (t_ex_node *) malloc (sizeof (t_ex_node))) == NULL)
        return (handle_err (rutname, 2));
    } 
    else 
    {
      if (!(WS_attr_table[i].examples = 
          (t_ex_node *) realloc (WS_attr_table[i].examples, 
          (WS_exs_count + 1) * sizeof (t_ex_node))) )
        return (handle_err (rutname, 2));
    }
  }

  WS_exs_count++;

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
              Remove examples form working set
\* -------------------------------------------------------------------------- */

int WS_remove_examples (void) 
{
  
  static char rutname[] = "remove_examples";
  register int i;

  if (WS_exs_count != 0) 
  {
    for (i = 0; i < WS_attr_count; i++) 
    {
      free (WS_attr_table[i].examples);
      WS_attr_table[i].examples = NULL;
    }
    WS_exs_count = 0;
  }

  return (NO_ERROR);
}  


/* -------------------------------------------------------------------------- *\
            Perform a mapping of an attribute's domain to a new one                  
\* -------------------------------------------------------------------------- */

int WS_map_domain (t_attr_ref aRef, int i, t_ex_node *newColPtr, 
           t_attr_node *a, t_map_node *mList, char *elseVal, 
      char *dKnowVal, char *dCareVal) 
{         
  int  newInd;

  switch (WS_attr_table[aRef].examples[i].ok) 
  {
    case T_DONT_KNOW :
      if (!strcmp (dKnowVal, "*"))
        newColPtr[i].ok = T_DONT_CARE;
      else if (dKnowVal != NULL && strcmp (dKnowVal, "?") && 
               strcmp (dKnowVal, "")) 
      {
        if ((newInd = binary_search_val_index (dKnowVal, a)) != -1)
           newColPtr[i].val.int_val = newInd;
      } 
      else /* Straight copy */
        newColPtr[i] = WS_attr_table[aRef].examples[i];
      break;
    case T_DONT_CARE :
      if (!strcmp (dCareVal, "?"))
        newColPtr[i].ok = T_DONT_KNOW;
      else if (dCareVal != NULL && strcmp (dCareVal, "?") && 
               strcmp (dCareVal, "")) {
        if ((newInd = binary_search_val_index (dCareVal, a)) != -1)
         newColPtr[i].val.int_val = newInd;
      } else /* Straight copy */
        newColPtr[i] = WS_attr_table[aRef].examples[i];
      break;

    case T_OK :
      return (map_ok_exval (WS_attr_table[aRef].examples[i], &newColPtr[i], 
              WS_attr_table[aRef], a, mList, elseVal));
      break;
  }

  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
                     Compare attribute tables
\* -------------------------------------------------------------------------- */

int WS_attr_table_cmp (char *backup_name) 
{
  int      i, newClass, newInx, oldClass, oldInx, result;
  char      *tmp1, *tmp2;
  t_exs_db *p_ex_db;
  tVStruct v;
  
  table_cmp_init(&v);  

  if (WS_example_db_restore(backup_name))
  {
    errmsg("Couldn't restore the example database");
    return 0;
  }

  /* Tarkista relevantit globaalit muuttujat */
  if (v.WS_attr_count != WS_attr_count)
    return (0);

  oldClass = v.WS_class_attr; newClass = WS_class_attr;

  if ((oldClass == -1) != (newClass == -1))
    return (0);

  oldInx = v.WS_first_read_attr; newInx = WS_first_read_attr;
  if (oldInx == oldClass)
    oldInx = v.WS_attr_table[oldInx].next_attr;
  if (newInx == newClass)
    newInx = WS_attr_table[newInx].next_attr;
  while (oldInx != -1 && newInx != -1) {
    
    /* Vertaa perussolmu */
    result = (!eq_q_str (WS_attr_table[newInx].name, v.WS_attr_table[oldInx].name))&&
       (WS_attr_table[newInx].a_type == v.WS_attr_table[oldInx].a_type) &&
       (WS_attr_table[newInx].val_cnt == v.WS_attr_table[oldInx].val_cnt) &&
       (WS_attr_table[newInx].first_read == v.WS_attr_table[oldInx].first_read);
    if (!result)
      return (0);
      
    /* Vertaa diskreetin attribuutin mahdollisten valuejen yhdenmukaisuus */
    if (WS_attr_table[newInx].a_type == T_USER || 
      WS_attr_table[newInx].a_type == T_BINARY) {
      for (i = 0; i < WS_attr_table[newInx].val_cnt; i++) {
          result = (!eq_q_str (WS_attr_table[newInx].disc_val_table[i].val,
          v.WS_attr_table[oldInx].disc_val_table[i].val )) &&
          (WS_attr_table[newInx].disc_val_table[i].next_val ==
          v.WS_attr_table[oldInx].disc_val_table[i].next_val);
          if (!result)
          return (0);
      }
    }
    
    /* Haetaan seuraava solmu */
    oldInx = v.WS_attr_table[oldInx].next_attr;
    newInx = WS_attr_table[newInx].next_attr;
    if (oldInx == oldClass)
      oldInx = v.WS_attr_table[oldInx].next_attr;
    if (newInx == newClass)
      newInx = WS_attr_table[newInx].next_attr;
  }  /* while */
  

  /* compare class attribute nodes */
	if ((oldClass != -1) && (newClass != -1))
	{
  	tmp1 = WS_attr_table[newClass].name;
  	tmp2 = v.WS_attr_table[oldClass].name;
  	result = (!strncmp(tmp1, "unnamed",7)) ||
        (!strncmp(tmp2,"unnamed",7)) ||
       (!strcmp (tmp1, tmp2));

  	result = result &&
   		(WS_attr_table[newClass].a_type == v.WS_attr_table[oldClass].a_type) &&
   		(WS_attr_table[newClass].val_cnt == v.WS_attr_table[oldClass].val_cnt) &&
   		(WS_attr_table[newClass].first_read == 	
				v.WS_attr_table[oldClass].first_read);
 
 		if (!result)
    	return (0);
      
  	if (WS_attr_table[newClass].a_type == T_USER || 
      WS_attr_table[newClass].a_type == T_BINARY) 
		{
    	for (i = 0; i < WS_attr_table[newClass].val_cnt; i++) 
			{
      	result = (!eq_q_str (WS_attr_table[newClass].disc_val_table[i].val,
        	v.WS_attr_table[oldClass].disc_val_table[i].val )) &&
         (WS_attr_table[newClass].disc_val_table[i].next_val ==
                v.WS_attr_table[oldClass].disc_val_table[i].next_val);
      	if (!result)
        return (0);
    	}
  	}
	} 
 
  return (1);
}


/* -------------------------------------------------------------------------- *\
          Get an attributes's value on given example in a string form
\* -------------------------------------------------------------------------- */

int WS_get_exval (char *a_name, int row, char *value) {
  
int  apuI, aRef;
float  apuF;
  
  if ((aRef = attr_binary_search (a_name)) == -1)
    return (NON_FATAL_ERROR);
    
  if (row >= WS_exs_count)
    return (handle_err ("get_exval", 13));
    
  if (WS_attr_table[aRef].examples[row].ok != T_OK)
    return (NON_FATAL_ERROR);
    
  switch (WS_attr_table[aRef].a_type) {
    case T_USER : 
    case T_BINARY :
      apuI = WS_attr_table[aRef].examples[row].val.int_val;
      strcpy (value, WS_attr_table[aRef].disc_val_table[apuI].val);
      break;
    case T_INT :
      apuI = WS_attr_table[aRef].examples[row].val.int_val;
      sprintf (value, "%d", apuI);
      break;
    case T_FLOAT :
      apuF = WS_attr_table[aRef].examples[row].val.float_val;
      sprintf (value, "%f", apuF);
      break;
  }
    
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
        Get class attributes's value on given example in a string form
\* -------------------------------------------------------------------------- */

int WS_get_class_exval (int row, char *value) 
{
  int  apuI, aRef;
  float  apuF;
  
  if (WS_class_attr == -1)
    return (NON_FATAL_ERROR);
  aRef = WS_class_attr;
    
  if (row >= WS_exs_count)
    return (handle_err ("get_exval", 13));
    
  if (WS_attr_table[aRef].examples[row].ok != T_OK)
    return (NON_FATAL_ERROR);
    
  switch (WS_attr_table[aRef].a_type) {
    case T_USER : 
    case T_BINARY :
      apuI = WS_attr_table[aRef].examples[row].val.int_val;
      strcpy (value, WS_attr_table[aRef].disc_val_table[apuI].val);
      break;
    case T_INT :
      apuI = WS_attr_table[aRef].examples[row].val.int_val;
      sprintf (value, "%d", apuI);
      break;
    case T_FLOAT :
      apuF = WS_attr_table[aRef].examples[row].val.float_val;
      sprintf (value, "%f", apuF);
      break;
  }
    
  return (NO_ERROR);
}


/* -------------------------------------------------------------------------- *\
          Add the example database at the end of database list
\* -------------------------------------------------------------------------- */

void WS_add_db_list(t_exs_db **pp_list,t_exs_db *p_item)
{
  if (!p_item) return;
  if (*pp_list)
  {
    p_item->next = (**pp_list).next;
    (**pp_list).next = p_item;
  }
  else
  {
    p_item->next = p_item;   /* make it a ring */
  }
  *pp_list = p_item;
}

/* -------------------------------------------------------------------------- *\
                        Copy an example column
\* -------------------------------------------------------------------------- */

void WS_ex_copy(t_ex_node *p_source, t_ex_node **pp_dest, int count)
{
  /* allocate memory */
  *pp_dest = (t_ex_node *)malloc(sizeof(t_ex_node)*count);

  /* and copy */
  memcpy(*pp_dest,p_source,count*sizeof(t_ex_node));  
}

/* -------------------------------------------------------------------------- *\
                  Back up the example database
\* -------------------------------------------------------------------------- */

int WS_example_db_backup(char *name)
{
  t_attr_node *p_node;
  int i;
  t_exs_db *p_copy = (t_exs_db *)malloc(sizeof(t_exs_db));
  alloc_string(&p_copy->name, name);
  p_copy->atts_count = WS_attr_count ; 
  p_copy->exs_count = WS_exs_count;
  p_copy->first_read = WS_first_read_attr; 
  p_copy->class_attr = WS_class_attr;
  p_copy->attr_itername = WS_attr_iter_by_name;
  p_copy->attr_iter_number = WS_attr_iter_by_read_order;

  /* now, copy the database */

  /* allocate memory for attributes */
  p_copy->attr_table = 
		(t_attr_node *)malloc(sizeof(t_attr_node)*WS_attr_count);

  /* and copy */
  for (i=0;i<WS_attr_count;i++)
  {
    if (WS_attr_table[i].name == "") 
    {
      p_copy->atts_count--;
      continue;
    }
    WS_copy_attr(WS_attr_table+i,&p_node);
    alloc_string( &p_node->name, WS_attr_table[i].name);
    p_copy->attr_table[i] = *p_node;
    WS_ex_copy(WS_attr_table[i].examples,&(p_copy->attr_table[i].examples),
			WS_exs_count);  
  }  

  /* and add the new copy to the list of databases */

  WS_add_db_list(&WS_p_exs_db_list,p_copy);
  WS_exs_db_count++;  

  return 0;  

}



/* -------------------------------------------------------------------------- *\
          Remove an example database from the database list
\* -------------------------------------------------------------------------- */

t_exs_db *WS_exs_db_list_remove(char *name)
{
  t_exs_db *p_tmp1,*p_tmp2;
  int i;
  
   p_tmp2 = p_tmp1 = WS_p_exs_db_list;
  
  p_tmp1 = p_tmp1->next;
  while (strcmp(p_tmp1->name,name))
  {
    if (p_tmp1 == WS_p_exs_db_list) return 0; /* not found in list */
    p_tmp2 = p_tmp1;
    p_tmp1 = p_tmp1->next;
  }
  /* if we get here, we p_tmp1 points to the item we want to return */
  if (p_tmp1 == p_tmp2)
    WS_p_exs_db_list = 0;
  else
  {
    p_tmp2->next = p_tmp1->next;
    WS_p_exs_db_list = p_tmp2; /* the list is not ordered so we can rotate it */
  }

  return p_tmp1;
  
} 


/* -------------------------------------------------------------------------- *\
                      Get a database of a given name
\* -------------------------------------------------------------------------- */

t_exs_db *WS_exs_db_list_search(char *name)
{
  t_exs_db *p_tmp1,*p_tmp2;
  int i;
  
   p_tmp2 = p_tmp1 = WS_p_exs_db_list;
  
  p_tmp1 = p_tmp1->next;
  while (strcmp(p_tmp1->name,name))
  {
    if (p_tmp1 == WS_p_exs_db_list) return 0; /* not found in list */
    p_tmp2 = p_tmp1;
    p_tmp1 = p_tmp1->next;
  }
  /* if we get here, we p_tmp1 points to the item we want to return */

  return p_tmp1;

}


/* -------------------------------------------------------------------------- *\
                Restore an example database
\* -------------------------------------------------------------------------- */

int WS_example_db_restore(char *name)
{
  t_exs_db *p_copy;
  /* find the backup */
  if  (p_copy = WS_exs_db_list_remove(name))
  {
    WS_dispose_dbase();   
    WS_attr_count = p_copy->atts_count; 
    WS_exs_count  = p_copy->exs_count ;
    WS_first_read_attr  = p_copy->first_read; 
    WS_class_attr  = p_copy->class_attr;
    WS_attr_iter_by_name  = p_copy->attr_itername;
    WS_attr_iter_by_read_order  = p_copy->attr_iter_number;
    WS_attr_table = p_copy->attr_table;
    free(p_copy);
    return 0;  
  }
  return 1;
}


/* -------------------------------------------------------------------------- *\
                        Get example database
\* -------------------------------------------------------------------------- */

int WS_example_db_get(char *name)
{
  if (WS_example_db_restore(name))
    return 1;

  if (WS_example_db_backup(name))
    return 1;
}


/* -------------------------------------------------------------------------- *\
              Change the size of an example value column
\* -------------------------------------------------------------------------- */

int WS_change_column_size(t_attr_ref a_ref, int new_size)
{
  static char rutname[] = "WS_change_column_size";
  
  t_ex_node **pp_exnode = &WS_attr_table[a_ref].examples;

  if (new_size) 
    if (*pp_exnode = (t_ex_node*)realloc(*pp_exnode,new_size*sizeof(t_ex_node)))
      return NO_ERROR;
    else
      return (handle_err (rutname, 2));
  else
  {
    free(*pp_exnode);
    *pp_exnode = 0;
    return NO_ERROR;
  }
}


/* -------------------------------------------------------------------------- *\
          Is a given example active in activation mask
\* -------------------------------------------------------------------------- */

boolean WS_is_active_ex(int i, t_ex_mask *p_mask)
{
  /* if no mask present all are active */
  if (!p_mask) return 1;

  if (i >= p_mask->count) 
  {
    errmsg("Internal error: No such example");
    return 0;
  }  

  if (p_mask->indexes[i] == p_mask->active_subset)  
    if (!p_mask->all_except) return 1; else return 0;
  else      
    if (!p_mask->all_except) return 0; else return 1;
}


/* -------------------------------------------------------------------------- *\
      Remove examples that  have greater or equal indexes than the given 
\* -------------------------------------------------------------------------- */

int WS_exs_remove_to_end(int first_removed)
{
  int i,no_of_attrs = WS_get_attr_cnt();

  if (WS_get_ex_cnt() < first_removed) return EXECUTION_OK;

  for (i=0;i < no_of_attrs; i++)
    if (WS_change_column_size(i,first_removed) != NO_ERROR)
    {
      errmsg("Fatal error: couldn't set the database to a consistent state");
      return EXECUTION_FAILED;
    }
  WS_set_ex_cnt(first_removed);
  return EXECUTION_OK;
}


/* -------------------------------------------------------------------------- *\
                  Get an attributes value for an example
\* -------------------------------------------------------------------------- */

int WS_get_attr_val(int ex_index,t_attr_ref a_ref,
  t_val_type *p_val_type,t_val *p_val)
{
  int tmp_i;

  if (ex_index >= WS_get_ex_cnt())
  {
    errmsg("Internal error: No such example");
    return EXECUTION_FAILED;
  }

  if (a_ref >= WS_get_attr_cnt())
  {
    errmsg("Internal error: No such attribute");
    return EXECUTION_FAILED;
  }
    
  /* unknown or don't care ? */
  switch (WS_attr_table[a_ref].examples[ex_index].ok)
  {
    case T_DONT_KNOW: *p_val_type = l_dont_know; return EXECUTION_OK;
    case T_DONT_CARE: *p_val_type = l_dont_care; return EXECUTION_OK;   
  }    

  /* return actual value */
  switch (WS_attr_table[a_ref].a_type) {
    case T_USER : 
    case T_BINARY :
      tmp_i = WS_attr_table[a_ref].examples[ex_index].val.int_val;
      alloc_string(&p_val->discrete, 
        WS_attr_table[a_ref].disc_val_table[tmp_i].val);
      *p_val_type = l_discrete;
      break;
    case T_INT :
      p_val->i = WS_attr_table[a_ref].examples[ex_index].val.int_val;
      *p_val_type = l_int;
      break;
    case T_FLOAT :
      p_val->f = WS_attr_table[a_ref].examples[ex_index].val.float_val;
      *p_val_type = l_float;
      break;
  }
    
  return EXECUTION_OK;
}
