#ifndef H_TELA_WS_DATABASE
#define H_TELA_WS_DATABASE

/*                                                                            *\

	  TELA - Testing Environment for Learning Algorithms - version 1.995
    ==================================================================

		Module:       WS_Database

		Contents:			TELA Working Set Database

		Services:

	
		Author: Marko Nyman (1992), Ilkka Hyvarinen (1992,1993,1994),
						Juho Rousu (1995,1996)

\*                                                                            */ 
#include "tela.h" 

/* example database */
typedef struct s_exs_db
{
	char *name;
	struct s_exs_db *next;	
	int atts_count;
	int exs_count;
	int first_read;
	int class_attr;
	int attr_itername;
	int attr_iter_number;
	t_attr_node *attr_table;
	/* something else? */
} t_exs_db;

extern t_exs_db *p_exs_db_list;
extern int exs_db_count;

/* rule base */
typedef struct
{
	int cf_exists;	      	/* any active classifier ? */
	t_algorithms cf_type; 	/* Id of the active classifier */
	char cf_name[256];			/* File where classifier recides */
} t_rule_base;

extern t_rule_base rule_base; /* rulebase */

/* Osoitin attribuuttirakenteen alkuun */
extern t_attr_node *WS_attr_table;

/* Hakumoodin tyyppi */
typedef enum {A_JARJ, M_JARJ} /* Aakkos- ja lukujrjestys } */
  t_search_mode;
  
/* HaeEsimRivin kyttmn tyypin alkio. Periaatteessa yksinkertaisesti
   esimerkkialkio johon on liitetty ko. attribuutin tyyppi ja jossa
   diskreettien attribuuttien indeksit on korvattu osoitimilla todelli-
   siin merkkijonoihin. Niden osoittimien lpi kirjoittaminen ei ole
   suotavaa. Itse asiassa se on ehdottoman kielletty. */
typedef struct {
  t_attr_types aTyyppi;
  t_ex_ok ok;
  union {
    int  intArvo;
    float floatArvo;
    char *userArvo;
  }  arvo;
}  t_EsimRiviAlkio;


void WS_set_active_classifier(t_algorithms alg);
int WS_get_active_classifier(t_algorithms *alg);
void WS_remove_active_classifier();
int WS_rule_backup(t_algorithms alg,char *fname);
int WS_rule_restore(t_algorithms alg,char *fname);
void WS_get_internal_rulename(char *buf);
int WS_get_attr_cnt (void);
int WS_get_ex_cnt (void);
int WS_set_ex_cnt (int e_count);
void WS_dispose_dbase (void);
int WS_create_attr (char *a_name, t_attr_types a_type, t_attr_node **a_node);
int WS_free_attr (t_attr_node *a_node);
int WS_add_attr (t_attr_node *a);
int WS_create_and_add_attr (char *a_name, t_attr_types a_type, 
	t_attr_ref *a_ref);
int WS_remove_attr (t_attr_ref a_ref);
int WS_get_attr_ref (char *a_name, t_attr_ref *a_ref);
int WS_is_attr_name_used (char *name);
int WS_get_attr_name (t_attr_ref a_ref, char **a_name);
int WS_replace_attr (t_attr_ref old_attr, t_attr_node *new_attr);
int WS_get_first_attr (t_attr_ref *a_ref, t_search_mode search_mode);
int WS_get_next_attr (t_attr_ref *a_ref, t_search_mode search_mode);
int WS_get_last_attr (t_attr_ref *a_ref, t_search_mode search_mode);
int WS_get_class_attr_name(char *p_buf);
int WS_is_class_attr (void);
int WS_get_class_attr (t_attr_ref *a_ref);
int WS_set_class_attr (t_attr_ref a_ref);
int WS_get_attr_type (t_attr_ref a_ref, t_attr_types *a_type);
int WS_set_attr_type (t_attr_ref a_ref, t_attr_types a_type);
int WS_copy_attr (t_attr_node *source, t_attr_node **dest);
int WS_get_attr_domain_size (t_attr_ref a_ref, int *val_cnt);
int WS_get_val_index (t_attr_ref a_ref, char *mJono, int *index);
int WS_is_attr_val_name(t_attr_ref a_ref,char *val_name);
int WS_get_index_attr_val (t_attr_ref a_ref, int index, char **mJono);
int WS_join_val_low (t_attr_node *a, char *new_value);
int WS_add_attr_val (t_attr_ref a_ref, char *new_value);
int WS_get_exval_kind (t_attr_ref a_ref, int e_line, t_ex_ok *kind);
int WS_set_exval_kind (t_attr_ref a_ref, int e_line, t_ex_ok kind);
int WS_get_nominal_exval (t_attr_ref a_ref, int e_line, int *index);
int WS_set_nominal_exval (t_attr_ref a_ref, int e_line, int index);
int WS_get_integer_exval (t_attr_ref a_ref, int e_line, int *value);
int WS_set_integer_exval (t_attr_ref a_ref, int e_line, int value);
int WS_get_float_exval (t_attr_ref a_ref, int e_line, float *value);
int WS_set_float_exval (t_attr_ref a_ref, int e_line, float value);
int WS_create_ex_column (t_attr_types a_type, int e_count, t_ex_node **column);
int WS_get_ex_column (t_attr_ref a_ref, t_ex_node **column);
int WS_clone_ex_column (t_attr_ref a_ref, t_ex_node *kohde, int lkm);
int WS_replace_ex_column (t_attr_ref a_ref, t_ex_node *new_column);
int WS_move_ex_row (int source, int dest);
int WS_add_ex_row (void);
int WS_remove_examples (void);
int WS_map_domain (t_attr_ref aRef, int i, t_ex_node *newColPtr, 
           t_attr_node *a, t_map_node *mList, char *elseVal, 
      char *dKnowVal, char *dCareVal);
int WS_attr_table_cmp (char *backup_name);
int WS_get_exval (char *a_name, int row, char *value);
int WS_get_class_exval (int row, char *value);
void WS_add_db_list(t_exs_db **pp_list,t_exs_db *p_item);
void WS_ex_copy(t_ex_node *p_source, t_ex_node **pp_dest, int count);
int WS_example_db_backup(char *name);
t_exs_db *WS_exs_db_list_remove(char *name);
t_exs_db *WS_exs_db_list_search(char *name);
int WS_example_db_restore(char *name);
int WS_example_db_get(char *name);
int WS_change_column_size(t_attr_ref a_ref, int new_size);
boolean WS_is_active_ex(int i, t_ex_mask *p_mask);
int WS_exs_remove_to_end(int first_removed);
int WS_get_attr_val(int ex_index,t_attr_ref a_ref,
  t_val_type *p_val_type,t_val *p_val);

#endif
